/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * Ambisonics Amplitude Panning renderer class (implementation).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include <cassert>       // for assert()
#include <cmath>         // for cos()
#include "aaprenderer.h"
#include "controller.h"  // for set_active_sources(), ...
#include "ssr_global.h"
#include "maptools.h"    // for purge()
#include "mathtools.h"    // for purge()
#include "ptrtools.h"    // for is_null(), destroy()
#include "stringtools.h" // for A2S(), S2A()

// for debugging only:
#include <libxml/debugXML.h>

using ptrtools::is_null;
using ptrtools::destroy;
using stringtools::S2A;
using stringtools::A2S;
using namespace ssr; // for id_t, processing, ready, ...

/** ctor.
 * @param scene see create()
 * @param controller see create()
 * @param delayline_size see create()
 **/
AAPRenderer::AAPRenderer(
    const Scene&       scene,
          Controller&  controller,
    const unsigned int ambisonics_order,
    const bool in_phase_rendering,
    const std::string& client_name
    ) throw (jack_error, aap_error) :
  Renderer(scene, client_name, true), // base class ctor.
  // the second argument enables the jack_callback function
  _in_phase_rendering(in_phase_rendering),
  _controller(controller),
  _input_prefix("in"),
  _output_prefix("out")
{

  _ambisonics_order = ambisonics_order;

  if (_in_phase_rendering) VERBOSE("Using in-phase rendering.");

  if (!_make_outputs())
  {
    maptools::purge(_output_map); // delete any leftovers in _output_map
    throw(aap_error("Output objects could not be created!"));
  }

  // There shouldn't be any sources in the scene at this point.
  // If there are, they are ignored!

  if (this->get_buffer_size() == 0)
  {
    WARNING("Buffersize == 0!");
  }
  else
  {
    // fill silence buffer with zeroes
    _silence.resize(this->get_buffer_size(), 0);
  }

  // init memory for crossfade slopes
  _fade_in.resize (this->get_buffer_size(), static_cast<sample_t>(0.0f));
  _fade_out.resize(this->get_buffer_size(), static_cast<sample_t>(0.0f));

  // raised cosine fade
  for(unsigned int n = 0u; n < this->get_buffer_size(); n++)
  {
    _fade_in[n]  = static_cast<sample_t>(0.5f
        + 0.5f * cos(static_cast<float>(this->get_buffer_size()-n)
		     / static_cast<float>(this->get_buffer_size()) * mathtools::PI));
    _fade_out[n] = static_cast<sample_t>(0.5f
        + 0.5f * cos(static_cast<float>(n)
		     / static_cast<float>(this->get_buffer_size()) * mathtools::PI));
  } // for

  _controller.publish_sample_rate((unsigned int)this->get_sample_rate());
  _controller.set_processing_state(ready); // TODO: lock with mutex?

  this->activate(); // activate JACK client
}

AAPRenderer::~AAPRenderer()
{
  _controller.set_processing_state(exiting);
  // cond_signal to any waiting threads?
  maptools::purge(_output_map);
  VERBOSE2("AAPRenderer: Output map cleared.");
  maptools::purge(_input_map);
  VERBOSE2("AAPRenderer: Input map cleared.");
}

/** stop rendering.
 * This is only called by JACK. Do not dare to call it yourself!
 * - stops rendering
 * - calls its base class counterpart
 *
 * @return I don't know what
 */
int AAPRenderer::jack_shutdown()
{
  ERROR("Prematurely disconnected from JACK server! "
      "Trying to stop processing ...");
  _controller.set_processing_state(exiting);
  // cond_signal to any waiting threads?
  maptools::purge(_output_map); WARNING("Output map cleared.");
  maptools::purge(_input_map);  WARNING("Input map cleared.");

  return JackClient::jack_shutdown(); // call original base class function
}

/** _.
 * @param scene Reference to the Scene object of the calling class. This
 * is used to obtain the geometry data in each JACK callback.
 * @param controller Reference to the Controller object. This is used to send
 * back information about active/inactive loudspeakers.
 * @param delayline_size Size of the delay lines in samples (all of them have
 * the same length).
 * @param initial_delay Amount of negative delay that can be used when reading
 * from the delay line
 * @param client_name see JackClient::create()
 * @return auto_ptr to the new AAPRenderer object.
 * @todo information about active loudspeakers per source should be passed to
 * the Controller.
 **/
AAPRenderer::ptr_t AAPRenderer::create(const Scene& scene,
    Controller& controller,
    const unsigned int order,
    const bool in_phase_rendering,
    const std::string& client_name)
{
  ptr_t temp; // temp = NULL
  try
  {
    temp.reset(new AAPRenderer(scene, controller, order, 
			       in_phase_rendering, client_name));
  }
  catch(jack_error& e)
  {
    ERROR(e.what());
  }
  catch(aap_error& e)
  {
    ERROR(e.what());
  }
  return temp;
}

bool AAPRenderer::new_source(id_t id, const std::string& port_name)
{
  if (_input_map.find(id) != _input_map.end())
  {
    WARNING("Input '" << id << "' already exists!");
    return false;
  }
  Input::ptr_t temp(Input::create(*this,
          _scene,
          _controller,
          id,
          _input_prefix + "_" + A2S(id)));
  if (is_null(temp))
  {
    WARNING("Initialization of input '" << id << "' failed!");
    return false;
  }
  // connect portname with input port
  if (!this->connect_ports(port_name, temp->get_port_name()))
  {
    WARNING("Input " << id << " could not be connected!");
    return false;
  }
  _input_map[id] = temp.release();

  VERBOSE("AAPRenderer: Input " << id << " registered.");
  return true;
}

bool AAPRenderer::new_source(id_t id, const std::string& port_name, 
			     const std::string& brir_file)
{
  (void) id;
  (void) port_name;
  (void) brir_file;
  
  ERROR("This function may not be called.");
  
  return false;
}

/** _.
 * This will only be called once. Loudspeakers can @b not be added later.
 * %Loudspeaker-IDs always range from 1 to @a number.
 * @param number number of loudspeakers
 * @return @b true on success.
 **/
bool AAPRenderer::_make_outputs()
{
  if (!_output_map.empty())
  {
    WARNING("BUG: Output map is not empty!");
    maptools::purge(_output_map);
  }

  ///////// check which loudspeaker is farthest ///////

  // get loudspeakers
  Loudspeaker::container_t loudspeakers;
  _scene.get_loudspeakers(loudspeakers);

  unsigned int no_of_norm_ls = 0;
  float farthest_loudspeaker_distance = 0;

  Loudspeaker::container_t::const_iterator i;

  for (i = loudspeakers.begin(); i != loudspeakers.end(); ++i)
  {
    if (i->model != Loudspeaker::subwoofer) no_of_norm_ls++;

    farthest_loudspeaker_distance = 
      std::max(farthest_loudspeaker_distance, 
	       (i->position - _scene.get_reference().position).length());
  }

  Scene::loudspeakers_t::size_type id = 1;

  /////////// determine loudspeaker delays ////////////////
  for (i = loudspeakers.begin(); i != loudspeakers.end(); ++i)
  {
    float delay_sec = (farthest_loudspeaker_distance - 
       (i->position - _scene.get_reference().position).length()) / ssr::c;

    DelayLine<sample_t>::size_type delay_samples = 
         static_cast<DelayLine<sample_t>::size_type>
                  (delay_sec * this->get_sample_rate() + 0.5f);

    // make Output
    Output::ptr_t temp(Output::create(*this, 
				      delay_samples + this->get_buffer_size(),
				      delay_samples,
				      _output_prefix + "_" + A2S(id)));
    if (is_null(temp))
    {
      WARNING("Initialization of output " << id << " failed!");
      maptools::purge(_output_map);
      return false;
    }

    _output_map[id] = temp.release();

    // amplitude correction for misplaced loudspeakers
    _output_map[id]->weight = 
          (i->position - _scene.get_reference().position).length() 
              / farthest_loudspeaker_distance;
    
    ++id;
  }

  VERBOSE("AAPRenderer: " << _output_map.size() << " Output objects registered.");

  // if Ambisonics order is not specified by the user
  if (!_ambisonics_order)
  {
    // to avoid problems when no normal loudspeaker is specified
    no_of_norm_ls = std::max(no_of_norm_ls, 1u);

    _ambisonics_order = static_cast<unsigned int>((no_of_norm_ls-1)/2);
  }

  VERBOSE("Using Ambisonics order " << _ambisonics_order << ".");

  return true;
}

bool AAPRenderer::delete_source(const id_t id)
{
  return maptools::delete_item(_input_map, id);
}

void AAPRenderer::delete_all_sources()
{
  maptools::purge(_input_map);
  // TODO: anything else?
}

/** _.
 * @param id identifier of the input channel
 * @return Pointer to the corresponding @a id. If none was found, @b NULL is
 * returned.
 **/
const AAPRenderer::Input* AAPRenderer::get_input(id_t id) const
{
  return maptools::get_item(_input_map, id);
}

/** fill a buffer with zeroes.
 * @param buffer pointer to the memory to be cleared
 * @param buffer_size number of data elements to be cleared
 **/
void AAPRenderer::clear_buffer(sample_t* const buffer,
    const jack_nframes_t buffer_size) const
{
  if (buffer_size != _silence.size())
  {
    // this should never happen because JACK buffer size will not change!
    WARNING("BUG?: silence resized! (" << _silence.size() << " -> "
        << buffer_size << ")");
    _silence.resize(buffer_size, 0);
  }
  else
  {
    std::copy(_silence.begin(), _silence.end(), buffer);
  }
}

void AAPRenderer::get_output_port_names(port_list_t& port_list) const
{
  for (output_map_t::const_iterator i = _output_map.begin();
      i != _output_map.end(); ++i)
  {
    assert(i->second != NULL);
    port_list.push_back(i->second->get_port_name());
  }
}

/** JACK process callback function.
 * This function is called every time the JACK server has new audio data.
 * @param nframes number of frames
 * @return I don't know what
 */
int AAPRenderer::jack_process(const jack_nframes_t nframes)
{
  // send some information to the controller (and then to the subscribers)
  _controller.set_transport_state(this->get_transport_state());
  _controller.set_cpu_load(this->get_cpu_load());

  switch (_scene.get_processing_state())
  {
    case processing:
      {
        Loudspeaker::container_t all_loudspeakers;
        // necessary memory is reserved inside the function
        _scene.get_loudspeakers(all_loudspeakers);
        // the loudspeakers contain already absolute positions

        // TODO: remove this if clause?
        if (all_loudspeakers.size() != _output_map.size())
        {
          ERROR("Different number of loudspeakers in Scene and AAPRenderer!");
          // TODO: how to react on that?
        }

        for (Loudspeaker::container_t::size_type i = 0;
            i < all_loudspeakers.size(); ++i)
        {
          // WARNING: vector indices start with 0, loudspeaker IDs start with 1!
          if (_output_map.find(i + 1) == _output_map.end())
          {
            // This should never happen!
            ERROR("Loudspeaker #" << i + 1 << " not found in output map!");
          }
          else
          {
            *_output_map[i + 1] = all_loudspeakers[i]; // assign new position,...
          } // if

        } // for all loudspeakers

	// get sources
        SourceCopy::container_t sources;
        _scene.get_sources(sources);

        // all sources
        for (SourceCopy::container_t::iterator i = sources.begin();
             i != sources.end(); ++i)
        {
          if (_input_map.find(i->id) == _input_map.end())
          {
            // This should never happen!
            ERROR("Source #" << i->id << " not found in input map!");
          }
          else
          {
            *_input_map[i->id] = *i; // assign new position,...         
            // fetch audio data
            _input_map[i->id]->fetch_new_buffer(nframes);
          }
        }

        float master_level = 0.0f;

        // iterate through all output objects via output_map to render
        // sources
        for (output_map_t::const_iterator i = _output_map.begin();
             i != _output_map.end(); ++i)
        {
          assert(i->second != NULL);

          // render sources
          _render(sources, nframes, i->second);

          // get signal level
          master_level = std::max(master_level,
              i->second->get_signal_level(nframes));
        }

        _controller.set_master_signal_level(master_level);

        // return something?
        break;
      } // case

    case ready: // audio processing is OFF
      for (output_map_t::const_iterator i = _output_map.begin();
          i != _output_map.end(); ++i)
      {
        assert(i->second != NULL); // assert that object really exists
        i->second->write_empty_buffer(nframes);
      }
      break;

    case initializing:
      WARNING("JACK process callback called while in \"initializing\" state!");
      // return something bad?
      break;

    case exiting:
      WARNING("JACK process callback called while in \"exiting\" state!");
      // return something bad?
      break;

    default:
      ERROR("BUG: Unknown client state!");
      // return something bad?
  }
  // TODO: find out what the return value means
  return 0;
}

void AAPRenderer::_render(const SourceCopy::container_t& sources,
			  const jack_nframes_t nframes, Output* current_output)
{
  // Output buffer
  sample_t* out = current_output->get_output_buffer(nframes);

  VERBOSE3("\n======================================"
      "==========================================");
  VERBOSE3("== output position: x: " << current_output->position.x << ", y: "
      << current_output->position.y << ", azimuth: " << current_output->orientation.azimuth);
  VERBOSE3("== JACK port: " << current_output->get_port_name());

  // the first output block replaces the contents of "out", then "accumulate" is
  // set to "true".
  bool accumulate = false;

  Output::coefficients_map_t current_source_weights; // empty map

  for (SourceCopy::container_t::const_iterator i = sources.begin();
      i != sources.end(); ++i)
  {
    if (i->mute == true) continue; // next source

    // input object must not be deleted! It's deleted in ~WFSRenderer()
    const Input* current_input = this->get_input(i->id);

    if (current_input == NULL)
    { // this happens regularly when a source is added and the renderer is
      // already processing because it takes some time until the JACK ports are
      // ready. Therefore the following warning is only printed if verbose > 1.
      VERBOSE2("WFSRenderer::Output: Input for source " << i->id
          << " doesn't exist in input map!");
      continue; // continue with next source
    }

    float weighting_factor = 0.0f;


    if (current_output->model == Loudspeaker::subwoofer)
    {
      weighting_factor = 1.0f;
    }
    // normal loudspeaker
    else
    {
      // this is the calculation of the driving function

      float alpha_0  = mathtools::deg2rad((current_output->position - 
	   _scene.get_reference().position).orientation().azimuth);
      float theta_pw = mathtools::deg2rad((i->position - 
	   _scene.get_reference().position).orientation().azimuth);

      // in-phase rendering
      if (_in_phase_rendering)
      {
	weighting_factor = pow(cos((alpha_0-theta_pw)/2), static_cast<float>(2*_ambisonics_order));
      }
      // conventional rendering
      else
      {
        // check numerical stability
        if (std::fabs(sin((alpha_0-theta_pw)/2)) < 0.0001f)
        {
	  weighting_factor = 1;
	}
	else 
        {
	   weighting_factor = sin( (2*_ambisonics_order + 1) * (alpha_0-theta_pw)/2 ) / 
                                 ( (2*_ambisonics_order + 1) * sin((alpha_0-theta_pw)/2) );
	} // if

      } // if in_phase


    } // if subwoofer

    // apply loudspeaker weight
    weighting_factor *= current_output->weight;

    // no distance attenuation for plane waves 
    if (i->model == Source::plane)
    {
      weighting_factor *= 0.5f / _scene.get_amplitude_reference_distance(); // 1/r
      // weighting_factor *= 0.25f / sqrt(_scene.get_amplitude_reference_distance()); // 1/sqrt(r)
    }
    else
    {
      // consider distance attenuation
      float source_distance = (i->position - _scene.get_reference().position).length();
      
      // no volume increase for sources closer than 0.5m to reference position
      source_distance = std::max(source_distance, 0.5f);

      weighting_factor *= 0.5f / source_distance; // 1/r
      //weighting_factor *= 0.25f / sqrt(source_distance); // 1/sqrt(r)
    }

    // apply the gain factor of the current source
    weighting_factor *= i->gain;

    // apply master volume
    weighting_factor *= _scene.get_corrected_master_volume();

    float previous_weighting_factor = current_output->get_previous_source_weight(i->id);

    // read from input, weight and add to output buffer.
    // IMPORTANT: first source is not added but replaces the buffer content.
    sample_t* in = current_input->get_current_audio_frame();

    if (accumulate)
    {
      for (unsigned int sample = 0; sample < nframes; sample++)
      {
        //*(out+sample) += *(in+sample) * weighting_factor;

	// TODO: Make crossfade more elegant???
        *(out+sample) += *(in+sample) *
          (weighting_factor * _fade_in[sample] +
           previous_weighting_factor * _fade_out[sample]);
      }
    }
    else
    {
      for (unsigned int sample = 0; sample < nframes; sample++)
      {
	// TODO: Make crossfade more elegant???
        *(out+sample) = *(in+sample) *
          (weighting_factor * _fade_in[sample] +
           previous_weighting_factor * _fade_out[sample]);
      }
    }

    current_source_weights[i->id] = Output::Coefficient(weighting_factor);

    // after the first (successful) write operation everything is added to the
    // output.
    accumulate = true;
  } // for all sources

  // if there was no successful write operation (maybe even the source list was
  // empty), the old audio data is still in the buffer.
  // In this case, we have to clear it:
  if (!accumulate)
  {
    // fill buffer with zeros
    this->clear_buffer(out, nframes);
    VERBOSE3("No source active. Output buffer cleared.");
  }

  current_output->write_output_buffer(nframes);

  // store source weights for next audio cycle
  current_output->set_previous_source_weights(current_source_weights);

}

unsigned int AAPRenderer::get_number_of_output_channels() const
{
  return _output_map.size();
}

Renderer::renderer_t AAPRenderer::get_renderer_type() const
{
  return Renderer::AAP;
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class AAPRenderer::SourceCopy
////////////////////////////////////////////////////////////////////////////////

AAPRenderer::SourceCopy::SourceCopy(const std::pair<id_t, Source>& other) :
  DirectionalPoint(other.second), // copy ctor of base class
  id(other.first),
  model(other.second.model),
  gain(other.second.gain),
  mute(other.second.mute)
{}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class AAPRenderer::Output
////////////////////////////////////////////////////////////////////////////////

/** ctor.
 * @param renderer calling AAPRenderer object
 **/
AAPRenderer::Output::Output(const AAPRenderer& renderer, 
			    const DelayLine<sample_t>::size_type delay_line_size, 
			    const DelayLine<sample_t>::size_type delay) :
  DelayLine<sample_t>(delay_line_size, delay), 
  _renderer(renderer), 
  _delay_line_required(delay)
{}

AAPRenderer::Output::~Output()
{
  if (_port) _renderer.unregister_port(_port);
}

AAPRenderer::Output::ptr_t AAPRenderer::Output::create(
	 const AAPRenderer& renderer, 
	 const DelayLine<sample_t>::size_type delay_line_size, 
	 const DelayLine<sample_t>::size_type delay,
         const std::string& port_name)
{
  ptr_t temp(new Output(renderer, delay_line_size, delay));
  if (!temp->_init(port_name))
  {
    ERROR("Unable to initialize output port '" + port_name + "'!");
    destroy(temp);
  }
  return temp;
}

AAPRenderer::Output& AAPRenderer::Output::operator=(const Loudspeaker& ls)
{
  this->DirectionalPoint::operator=(ls);
  // we don't update the weight since we have already determined it 
  // automatically
  this->model  = ls.model;
  this->mute   = ls.mute;
  this->active = ls.active; // even if we are not interested in it
  return *this;
}

/** create a JACK output port and get its full name.
 * @param port_name the desired output port name (the part after the ":")
 *
 * The full port name including the client name can be obtained by calling
 * get_port_name() afterwards.
 **/
bool AAPRenderer::Output::_init(const std::string& port_name)
{
  _port = _renderer.register_out_port(port_name);
  if (_port == NULL)
  {
    WARNING("Output port '" + port_name
        + "' could not be registered.");
    return false;
  }
  // get actual port name and save it to member variable
  _port_name = jack_port_name(_port);

  _audio_buffer.resize(_renderer.get_buffer_size(), 0.0f);

  return true;
}

std::string AAPRenderer::Output::get_port_name() const
{
  return _port_name;
}

jack_port_t* AAPRenderer::Output::get_port() const
{
  return _port;
}

Renderer::sample_t* AAPRenderer::Output::get_output_buffer(const jack_nframes_t nframes)
{
  if (!_delay_line_required)
  {
    // return JACK output buffer
    return 
      static_cast<sample_t*>(jack_port_get_buffer(this->get_port(), nframes));
  }
  // return internal audio buffer
  else return &_audio_buffer[0]; 
}

void AAPRenderer::Output::write_output_buffer(const jack_nframes_t nframes)
{
  if (!_delay_line_required) return; // nothing to do

  // else: write to delay line, then read from it

  this->prepare_write();
  this->write(&_audio_buffer[0], nframes);

  // get JACK output buffer
  sample_t* out = static_cast<sample_t*>
       (jack_port_get_buffer(this->get_port(), nframes));

  if (!read(out, nframes))
  {
    ERROR("Can not write output buffer.");
  }
}

float AAPRenderer::Output::get_previous_source_weight(id_t source_id) const
{

  coefficients_map_t::const_iterator iter = 
                                 _previous_source_weights.find(source_id);

  if (iter != _previous_source_weights.end())
  {
    return iter->second.weight;
  }
  else return 0.0f;
}

void AAPRenderer::Output::set_previous_source_weights(coefficients_map_t weights)
{
  _previous_source_weights.clear();
  _previous_source_weights = weights;
}

void AAPRenderer::Output::write_empty_buffer(const jack_nframes_t nframes) const
{
  if (_port)
  {
    sample_t* out = static_cast<sample_t*>
                        (jack_port_get_buffer(_port, nframes));
    _renderer.clear_buffer(out, nframes);
  }
}

void AAPRenderer::Output::clear_active_source_ids()
{
  this->active_source_ids.clear(); // reset active sources to an empty set
}

float AAPRenderer::Output::get_signal_level(const jack_nframes_t nframes)
{
  // get JACK output buffer
  sample_t* out;
  out = static_cast<sample_t*>(jack_port_get_buffer(_port, nframes));

  return _renderer.get_max_signal_amplitude(out, nframes);
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class AAPRenderer::Input
////////////////////////////////////////////////////////////////////////////////

/** private ctor. Object can only be created with create().
 * @param renderer reference to the calling AAPRenderer object. Using this, we
 * can access the public members of the AAP renderer.
 **/
AAPRenderer::Input::Input(const AAPRenderer& renderer,
        const Scene& scene,
        Controller& controller,
        const id_t source_id) :
  SourceCopy(source_id),
  _renderer(renderer),
  _scene(scene),
  _controller(controller),
  _current_audio_frame(NULL),
  _port(NULL),
  _port_name("")
{}

AAPRenderer::Input::~Input()
{
  if (_port) _renderer.unregister_port(_port);
}

AAPRenderer::Input::ptr_t AAPRenderer::Input::create(
        const AAPRenderer& renderer,
        const Scene& scene,
        Controller& controller,
        const id_t source_id,
        const std::string& name)
{
  ptr_t temp(new Input(renderer, scene, controller, source_id));
  if (!temp->_init(name))
  {
    ERROR("Unable to initialize input '" + name + "'!");
    destroy(temp);
  }
  return temp;
}

bool AAPRenderer::Input::_init(const std::string& name)
{
  _port = _renderer.register_port(name, JackPortIsInput);
  if (_port == NULL)
  {
    WARNING("Input port '" + name + "' could not be registered.");
    return false;
  }
  // get actual port name and save it to member variable
  _port_name = jack_port_name(_port);
  return true;
}

AAPRenderer::Input& AAPRenderer::Input::operator=(const SourceCopy& source)
{
  this->DirectionalPoint::operator=(source);
  this->model  = source.model;
  this->mute   = source.mute;
  this->gain   = source.gain;

  return *this;
}

bool AAPRenderer::Input::fetch_new_buffer(jack_nframes_t nframes)
{
  _current_audio_frame = static_cast<sample_t*>(jack_port_get_buffer(_port, nframes));

  if (_current_audio_frame == NULL)
  {
    WARNING("Got no input buffer from JACK!");
    return false;
  }

  assert(_current_audio_frame != NULL);

  if (_scene.get_source_mute_state(this->id))
  {
    // update source signal level
    _controller.set_source_signal_level(this->id, 0.0f);
  }
  else
  {
    // update source signal level (max amplitude)
    _controller.set_source_signal_level(this->id,
        _renderer.get_max_signal_amplitude(_current_audio_frame, nframes) *
        _scene.get_source_gain(this->id));

    // update source signal level (RMS)
    //_controller.set_source_signal_level(_source_id, 
    //      _renderer.get_signal_rms(in, nframes) * 
    //      std::pow(_scene.get_source_gain(_source_id), int(nframes))
    //      , 0);
  }
  return true;
}

Renderer::sample_t* AAPRenderer::Input::get_current_audio_frame() const
{
  return _current_audio_frame;
}

std::string AAPRenderer::Input::get_port_name() const
{
  return _port_name;
}

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
