/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * Audio file I/O tools (implementation).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include <cstring>

#include "audiofileiotools.h"

/** @param info_container container to be filled with info data. Note: Any 
 * previously apparent data will be deleted.
 **/
bool audiofileiotools::get_audio_file_info(const std::string file_name, SF_INFO& info_container)
{
  // clear memory
  memset (&info_container, 0, sizeof (info_container));

  SNDFILE* audio_file = sf_open(file_name.c_str(), SFM_READ, &info_container);

  if (audio_file == NULL)
  {
    ERROR("Cannot open file '" << file_name << "' as audio file.");
    return false;
  }

  /*
    WARNING("Sample Rate   : " << info_container.samplerate) ;
    
    if (info_container.frames > 0x7FFFFFFF) WARNING("Frames        : unknown\n") ;
    else  WARNING("Frames        : " << (long) info_container.frames) ;
    
    WARNING("Channels      : " << info_container.channels) ;
    // WARNING("Format        : " << info_container.format) ;
    WARNING("Sections      : " << info_container.sections) ;
    WARNING("Seekable      : " << (info_container.seekable ? "TRUE" : "FALSE")) ;
    WARNING("Duration (s)  : " << info_container.frames / info_container.samplerate) ;
  */

  sf_close(audio_file);

  return true;
}

int audiofileiotools::get_no_of_channels(const std::string file_name)
{
  SF_INFO info_container;

  if(!get_audio_file_info(file_name, info_container)) return 0;
  else return info_container.channels;
}

int audiofileiotools::get_sample_rate(const std::string file_name)
{
  SF_INFO info_container;

  if(!get_audio_file_info(file_name, info_container)) return 0;
  else return info_container.samplerate;
}

/** @param container will be filled with (multichannel) audio data. If the audio
 * data is multichannel, the container contains all channels concatenated.  
 * The container is cleared before filling.
 **/
bool audiofileiotools::load_filter_t(const std::string file_name, Convolver::data_t& container)
{
  container.clear();

  SF_INFO info_container;
  memset (&info_container, 0, sizeof (info_container));

  SNDFILE* audio_file = sf_open(file_name.c_str(), SFM_READ, &info_container);

  if (audio_file == NULL) return false;// no file was opened

  VERBOSE("Loading " << info_container.channels
      << " channel(s) of audio data from file '" << file_name << "'.");

  // prepare buffer for audio data
  Convolver::data_t buffer;
  buffer.resize(info_container.frames * info_container.channels, 0.0f);

  // read audio data
  const sf_count_t no_of_frames_read =
    sf_readf_float(audio_file, buffer.data(), info_container.frames);

  // check if things are ok
  if (no_of_frames_read != info_container.frames)
  {
    ERROR("Did not find the announced number of audio samples in file '"
        << file_name << "' (" << no_of_frames_read << " vs. "
        << info_container.frames << ").");

    // return false;
  }

  // prepare container
  container.resize(buffer.size());

  float* audio_sample = buffer.data();

  // separate interleaved channels and concatenate them
  for (int n = 0; n < static_cast<int>(info_container.frames); n++)
  {
    for (int nn = 0; nn < info_container.channels; nn++)
    {
      container[n + nn*info_container.frames] = *audio_sample;
      audio_sample++;
    }
  }

  sf_close(audio_file);

  return true;
}

/*
PlaneWaveDecomposition::ptr_t audiofileiotools::load_directivity_f(const std::string file_name, const unsigned int partition_size)
{ 

  (void) file_name; (void)partition_size;
  PlaneWaveDecomposition::ptr_t temp(NULL);
  
  //  
  Convolver::filter_t exp_coeffs_t;
  
  if (!audiofileiotools::load_filter_t(file_name, exp_coeffs_t)) return temp;
  
  std::vector<Convolver::filter_t> exp_coeffs_f;
  exp_coeffs_f.resize(exp_coeffs_t.size());

  // convert expansion coefficients to frequency domain
  for (unsigned int n = 0u; n < exp_coeffs_t.size(); n++)
  {
    Convolver::prepare_impulse_response(exp_coeffs_f[n], exp_coeffs_t[n].data(), 
          exp_coeffs_t[n].size(), partition_size);
  }
  
  temp = PlaneWaveDecomposition::create(exp_coeffs_f,1);
  
  if (ptrtools::is_null(temp))
  {
    WARNING("Initialization of plane wave decomposition failed.");
    return temp;
  }
  //

  return temp;
}
*/

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
