/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * Binaural renderer class (implementation).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include <cassert>       // for assert()
#include <cmath>         // for cos()
#include "binauralrenderer.h"
#include "controller.h"  // for set_active_sources(), ...
#include "ssr_global.h"
#include "mathtools.h"    // for purge()
#include "ptrtools.h"    // for is_null(), destroy()
#include "stringtools.h" // for A2S(), S2A()
#include "audiofileiotools.h"

// for debugging only:
#include <libxml/debugXML.h>

using ptrtools::is_null;
using ptrtools::destroy;
using stringtools::S2A;
using stringtools::A2S;
using namespace ssr; // for id_t, processing, ready, ...

/** ctor.
 * @param scene see create()
 * @param controller see create()
 **/
BinauralRenderer::BinauralRenderer(const Scene& scene, Controller& controller,
    const unsigned int hrir_size, const std::string& hrir_file_name,
    const std::string& client_name) throw (jack_error, binaural_error) :
  Renderer(scene, client_name, true), // base class ctor
  _controller(controller),
  _input_prefix("in"),
  _output_prefix("out")
{
  // There shouldn't be any sources in the scene at this point.
  // If there are, they are ignored!

  if (this->get_buffer_size() == 0)
  {
    WARNING("Buffersize == 0!");
  }
  else
  {
    // fill silence buffer with zeroes
    _silence.resize(this->get_buffer_size(), 0);
  }

  // try to create output
  Output::ptr_t temp(Output::create(*this, _output_prefix));
  if (is_null(temp))
  {
    throw(binaural_error("Initialization of outputs failed!"));
  }
  _output = temp.release();

  // Instantiate a temporary hrtf container
  hrtfs_t hrtfs;

  // load hrirs
  if (!_load_hrtfs(hrtfs, hrir_size, hrir_file_name))
  {
    throw(binaural_error("Error loading HRIRs."));
  }

  // set hrirs
  _output->set_hrtfs(hrtfs);

  _controller.publish_sample_rate(static_cast<unsigned int>(this->get_sample_rate()));
  _controller.set_processing_state(ready); // TODO: lock with mutex?

  this->activate(); // activate JACK client
}

BinauralRenderer::~BinauralRenderer()
{
  _controller.set_processing_state(exiting);

  // cond_signal to any waiting threads?

  delete _output;
  VERBOSE2("BinauralRenderer: Output cleared.");

  maptools::purge(_input_map);
  VERBOSE2("BinauralRenderer: Input map cleared.");
}

/** stop rendering.
 * This is only called by JACK. Do not dare to call it yourself!
 * - stops rendering
 * - calls its base class counterpart
 *
 * @return I don't know what
 */
int BinauralRenderer::jack_shutdown()
{
  ERROR("Prematurely disconnected from JACK server! "
      "Trying to stop processing ...");
  _controller.set_processing_state(exiting);
  // cond_signal to any waiting threads?
  //maptools::purge(_output_map); WARNING("Output map cleared.");
  maptools::purge(_input_map);  WARNING("Input map cleared.");

  return JackClient::jack_shutdown(); // call original base class function
}

/** _.
 * @param scene Reference to the Scene object of the calling class. This
 * is used to obtain the geometry data in each JACK callback.
 * @param controller Reference to the Controller object. This is used to send
 * back information about active/inactive loudspeakers.
 * @param delayline_size Size of the delay lines in samples (all of them have
 * the same length).
 * @param initial_delay Amount of negative delay that can be used when reading
 * from the delay line
 * @param client_name see JackClient::create()
 * @return auto_ptr to the new BinauralRenderer object.
 * @todo information about active loudspeakers per source should be passed to
 * the Controller.
 **/
BinauralRenderer::ptr_t BinauralRenderer::create(const Scene& scene,
    Controller& controller, const unsigned int hrir_size,
    const std::string& hrir_file_name,
    const std::string& client_name)
{
  ptr_t temp; // temp = NULL
  try
  {
    temp.reset(new BinauralRenderer(scene, controller,
          hrir_size, hrir_file_name, client_name));
  }
  catch(jack_error& e)
  {
    ERROR(e.what());
  }
  catch(binaural_error& e)
  {
    ERROR(e.what());
  }
  return temp;
}

bool BinauralRenderer::_load_hrtfs(hrtfs_t& hrtfs,
           const unsigned int hrir_size,
           const std::string& hrir_file_name)
{
  std::cout << "Loading HRIRs. Please be patient.\n";
  
  // get sample rate of hrirs
  const int sr = audiofileiotools::get_sample_rate(hrir_file_name);
  if (!sr) return false;
  
  // if hrirs do not have the correct sample rate
  else if (sr != static_cast<int>(this->get_sample_rate()))
  {
    throw(binaural_error("The head-related impulse responses in file '"
          + hrir_file_name + "' are sampled at rate " + A2S(sr)
          + " Hz. JACK is currently running at "
          + A2S(this->get_sample_rate()) + " Hz."));
  }

  // check number of impulse responses
  const int no_of_ch = audiofileiotools::get_no_of_channels(hrir_file_name);

  // error message comes from function get_no_of_channels()
  if (!no_of_ch) return false;

  if (no_of_ch != 720)
  {
    ERROR("The file '" << hrir_file_name << "' does not contain "
        "the required number of channels (" << no_of_ch << " vs. 720).");
    return false;
  }

  Convolver::data_t container;

  if (!audiofileiotools::load_filter_t(hrir_file_name, container))
    return false;

  const unsigned int hrir_size_in_file = container.size()/720;
  unsigned int effective_hrir_size = hrir_size_in_file;

  // check if user has specified something
  if (hrir_size && hrir_size < hrir_size_in_file)
  {
    WARNING("I'm loading only " << hrir_size << " of " << hrir_size_in_file << 
            " HRIR coefficients from file '" << hrir_file_name << "'.");
    effective_hrir_size = hrir_size;
  }
  else if (hrir_size && hrir_size > hrir_size_in_file)
  {
    WARNING("File '" << hrir_file_name << "' contains only " << hrir_size_in_file << 
	    " HRIR coefficients. The user has specified to load " << hrir_size << ".");
  }

  // fill HRTF container
  for (int n = 0; n < 360; n++)
  {
    Convolver::data_t dummy;

    // initialize hrtfs.first[n] and hrtfs.second[n]
    hrtfs.first.push_back(dummy);
    hrtfs.second.push_back(dummy);

    // transfer hrirs to frequency domain and store them in vector hrtfs
    Convolver::prepare_impulse_response(hrtfs.first[n],
                            container.data() + 2*n*hrir_size_in_file,
                            effective_hrir_size, this->get_buffer_size());
    Convolver::prepare_impulse_response(hrtfs.second[n],
                            container.data() + (2*n+1)*hrir_size_in_file,
                            effective_hrir_size, this->get_buffer_size());
  }

  VERBOSE("HRIRs of size " << effective_hrir_size << " loaded.");

  return true;
}


bool BinauralRenderer::new_source(id_t id, const std::string& port_name)
{
  if (_input_map.find(id) != _input_map.end())
  {
    WARNING("Input '" << id << "' already exists!");
    return false;
  }
  Input::ptr_t temp(Input::create(*this,
          _scene,
          _controller,
          id,
          _input_prefix + "_" + A2S(id)));
  if (is_null(temp))
  {
    WARNING("Initialization of input '" << id << "' failed!");
    return false;
  }
  // connect portname with input port
  if (!this->connect_ports(port_name, temp->get_port_name()))
  {
    return false;
  }
  _input_map[id] = temp.release();

  VERBOSE("BinauralRenderer: Input " << id << " registered.");
  return true;
}

bool BinauralRenderer::new_source(id_t id, const std::string& port_name, 
				  const std::string& brir_file)
{
  (void) id;
  (void) port_name;
  (void) brir_file;
  
  ERROR("This function may not be called.");
  
  return false;
}

bool BinauralRenderer::delete_source(const id_t id)
{
  return maptools::delete_item(_input_map, id);
}


void BinauralRenderer::delete_all_sources()
{
  maptools::purge(_input_map);
  // TODO: anything else?
}

/** _.
 * @param id identifier of the input channel
 * @return Pointer to the corresponding @a id. If none was found, @b NULL is
 * returned.
 **/
BinauralRenderer::Input* BinauralRenderer::get_input(id_t id)
{
  return maptools::get_item(_input_map, id);
}

/** fill a buffer with zeroes.
 * @param buffer pointer to the memory to be cleared
 * @param buffer_size number of data elements to be cleared
 **/
void BinauralRenderer::clear_buffer(sample_t* const buffer,
    const jack_nframes_t buffer_size) const
{
  if (buffer_size != _silence.size())
  {
    // this should never happen because JACK buffer size will not change!
    WARNING("BUG?: silence resized! (" << _silence.size() << " -> "
        << buffer_size << ")");
    _silence.resize(buffer_size, 0);
  }
  else
  {
    std::copy(_silence.begin(), _silence.end(), buffer);
  }
}

void BinauralRenderer::get_output_port_names(port_list_t& port_list) const
{
  _output->get_port_names(port_list);
}

/** JACK process callback function.
 * This function is called every time the JACK server has new audio data.
 * @param nframes number of frames
 * @return I don't know what
 */
int BinauralRenderer::jack_process(const jack_nframes_t nframes)
{
// code must be suitable for real-time execution.
// That means following function are not allowed:
// malloc, free, printf, pthread_mutex_lock, sleep,
// wait, poll, select, pthread_join, pthread_cond_wait,
// etc, etc.

  // clear all information about active/inactive loudspeakers
  _controller.clear_all_active_sources();
  // send some information to the controller (and then to the subscribers)
  _controller.set_transport_state(this->get_transport_state());
  _controller.set_cpu_load(this->get_cpu_load());

  switch (_scene.get_processing_state())
  {
    case processing:
      {
        // TODO: is it more efficient to initialize the list as member variable
        // and clear it at each jack_process?
        SourceCopy::container_t sources;
        // reservation of memory is done in Scene::get_sources if a std::vector
        // is used (like in our case)
        _scene.get_sources(sources);

        const DirectionalPoint reference = _scene.get_reference();
        float master_volume = _scene.get_corrected_master_volume();

        (*_output) = reference;

        // TODO: Necessary??
        _output->write_empty_buffer(nframes);

        // render sources
        _output->render(sources, nframes, master_volume,
			_scene.get_amplitude_reference_distance());

        _controller.set_master_signal_level(_output->get_signal_level(nframes));

        break;
      }

    case ready: // audio processing is OFF
        _output->write_empty_buffer(nframes);
        break;

    case initializing:
            //WARNING("JACK process callback called while in \"initializing\" state!");
      // return something bad?
      break;

    case exiting:
      WARNING("JACK process callback called while in \"exiting\" state!");
      // return something bad?
      break;

    default:
      ERROR("BUG: Unknown client state!");
      // return something bad?
  }
  // TODO: find out what the return value means
  return 0;
}

unsigned int BinauralRenderer::get_number_of_output_channels() const
{
  return 2u;
}

Renderer::renderer_t BinauralRenderer::get_renderer_type() const
{
  return Renderer::binaural;
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class BinauralRenderer::SourceCopy
////////////////////////////////////////////////////////////////////////////////

BinauralRenderer::SourceCopy::SourceCopy(const std::pair<id_t, Source>& other) :
  DirectionalPoint(other.second), // copy ctor of base class
  id(other.first),
  model(other.second.model),
  gain(other.second.gain),
  mute(other.second.mute)
{}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class BinauralRenderer::Output
////////////////////////////////////////////////////////////////////////////////

/** ctor.
 * @param renderer calling BinauralRenderer object
 **/
BinauralRenderer::Output::Output(BinauralRenderer& renderer) :
  _renderer(renderer)
{}

BinauralRenderer::Output::~Output()
{
  if (_port_left) _renderer.unregister_port(_port_left);
  if (_port_right) _renderer.unregister_port(_port_right);
}

BinauralRenderer::Output::ptr_t BinauralRenderer::Output::create(
    BinauralRenderer& renderer, const std::string& port_name)
{
  ptr_t temp(new Output(renderer));
  if (!temp->_init(port_name))
  {
    ERROR("Unable to initialize output port '" + port_name + "'!");
    destroy(temp);
  }
  return temp;
}

/** create a JACK output port and get its full name.
 * @param port_name the desired output port name (the part after the ":")
 *
 * The full port name including the client name can be obtained by calling
 * get_port_name() afterwards.
 **/
bool BinauralRenderer::Output::_init(const std::string& port_name)
{
  _port_left  = _renderer.register_out_port(port_name + "_left");
  _port_right = _renderer.register_out_port(port_name + "_right");

  if (_port_left == NULL || _port_right == NULL)
  {
    WARNING("Output ports for Binaural-Renderer could not be registered.");
    return false;
  }

  // get actual port names and save it to member variables
  _port_name_left  = jack_port_name(_port_left);
  _port_name_right = jack_port_name(_port_right);

  return true;
}

void BinauralRenderer::Output::set_hrtfs(const hrtfs_t hrtfs)
{
  _hrtfs = hrtfs;
}

void BinauralRenderer::Output::get_port_names(port_list_t& port_list) const
{
  port_list.push_back(_port_name_left);
  port_list.push_back(_port_name_right);
}

float BinauralRenderer::Output::get_signal_level(const jack_nframes_t nframes)
{
  // get JACK output buffer
  sample_t* out_left;
  out_left = static_cast<sample_t*>(jack_port_get_buffer(_port_left, nframes));

  sample_t* out_right;
  out_right = static_cast<sample_t*>(jack_port_get_buffer(_port_right, nframes));

  // get maximum amplitude for right ear and return
  return std::max( _renderer.get_max_signal_amplitude(out_left, nframes),
       _renderer.get_max_signal_amplitude(out_right, nframes));
}

void BinauralRenderer::Output::write_empty_buffer(const jack_nframes_t nframes) const
{
  if (_port_left)
  {
    sample_t *out;
    out = static_cast<sample_t*>(jack_port_get_buffer(_port_left, nframes));
    _renderer.clear_buffer(out, nframes);
  }
  if (_port_right)
  {
    sample_t *out;
    out = static_cast<sample_t*>(jack_port_get_buffer(_port_right, nframes));
    _renderer.clear_buffer(out, nframes);
  }
}

void BinauralRenderer::Output::render(SourceCopy::container_t& sources,
              const jack_nframes_t nframes,
	      const float master_volume,
	      const float ampl_ref_dist)
{
  // get JACK output buffers
  sample_t* output_buffer_left  = static_cast<sample_t*>
    (jack_port_get_buffer(_port_left, nframes));
  sample_t* output_buffer_right = static_cast<sample_t*>
    (jack_port_get_buffer(_port_right, nframes));

  if (output_buffer_left == NULL || output_buffer_right == NULL)
  {
    ERROR("Cannot access JACK output ports.");
    return;
  }

  for (SourceCopy::container_t::iterator i = sources.begin();
       i != sources.end(); ++i)
  {
    float weighting_factor = 1.0f;

    // input object must not be deleted! It's deleted in ~BinauralRenderer()
    Input* current_input = _renderer.get_input(i->id);
    if (current_input == NULL)
    { 
      // this happens regularly when a source is added and the renderer is
      // already processing because it takes some time until the JACK ports are
      // ready. Therefore the following warning is only printed if verbose > 1.
      VERBOSE2("BinauralRenderer::Output: Input for source " << i->id
         << " doesn't exist in input map!");
      continue; // continue with next source
    }

    float interp_factor = 0.0f;

    // no distance attenuation for plane waves 
    if (i->model == Source::plane)
    {
      weighting_factor *= 0.5f / ampl_ref_dist; // 1/r
      // weighting_factor *= 0.25f / sqrt(ampl_ref_dist); // 1/sqrt(r)
    }
    else 
    {
      // consider distance attenuation
      float source_distance = (i->position - this->position).length();
      
      if (source_distance < 0.5f) interp_factor = 1.0f - 2*source_distance;

      // no volume increase for sources closer than 0.5m
      source_distance = std::max(source_distance, 0.5f);
      
      weighting_factor *= 0.5f / source_distance; // 1/r
      // weighting_factor *= 0.25f / sqrt(source_distance); // 1/sqrt(r)
    }

    // apply the gain factor of the current source
    weighting_factor *= i->gain;

    // apply master volume
    weighting_factor *= master_volume;

    // TODO: Save computation power!
    if (i->mute == true) weighting_factor = 0.0f;//continue; // next source

    assert(weighting_factor >= 0.0f);

    current_input->fetch_new_buffer(nframes);
    current_input->weight_audio_data(weighting_factor);

    // calculate relative position of sound source
    int hrtf_index = static_cast<int>(mathtools::rad2deg(
	    atan2(i->position.y - this->position.y, 
		  i->position.x - this->position.x))
				      - this->orientation.azimuth + 0.5f)%360;

    // make sure that index is in proper range
    if (hrtf_index < 0) hrtf_index += 360;

    sample_t* audio_sample_in_left = 
      current_input->apply_hrtf_to_left_ear(_hrtfs.first[hrtf_index], interp_factor, weighting_factor);
    sample_t* audio_sample_in_right = 
      current_input->apply_hrtf_to_right_ear(_hrtfs.second[hrtf_index], interp_factor, weighting_factor);

    sample_t* audio_sample_out_left  = output_buffer_left;
    sample_t* audio_sample_out_right = output_buffer_right;

    // add audio data to output buffers
    for(unsigned int n = 0; n<nframes; n++)
    {
      (*audio_sample_out_left)  += (*audio_sample_in_left);
      (*audio_sample_out_right) += (*audio_sample_in_right);

      audio_sample_out_left++;  audio_sample_in_left++;
      audio_sample_out_right++; audio_sample_in_right++;
    }
  } // for all sources

  // TODO: return true or something?
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class BinauralRenderer::Input
////////////////////////////////////////////////////////////////////////////////

/** private ctor. Object can only be created with create().
 * @param renderer reference to the calling BinauralRenderer object. Using this, we
 * can access the public members of the Binaural renderer.
 **/
BinauralRenderer::Input::Input(const BinauralRenderer& renderer,
        const Scene& scene,
        Controller& controller,
        const id_t source_id) :
  _renderer(renderer),
  _scene(scene),
  _controller(controller),
  _source_id(source_id),
  _port(NULL),
  _port_name("")
{}

BinauralRenderer::Input::~Input()
{
  if (_port) _renderer.unregister_port(_port);
}

BinauralRenderer::Input::ptr_t BinauralRenderer::Input::create(
        const BinauralRenderer& renderer,
        const Scene& scene,
        Controller& controller,
        const id_t source_id,
        const std::string& name)
{
  ptr_t temp(new Input(renderer, scene, controller, source_id));
  if (!temp->_init(name))
  {
    ERROR("Unable to initialize input '" + name + "'!");
    destroy(temp);
  }
  return temp;
}

bool BinauralRenderer::Input::_init(const std::string& name)
{
  _port = _renderer.register_port(name, JackPortIsInput);
  if (_port == NULL)
  {
    WARNING("Input port '" + name + "' could not be registered.");
    return false;
  }

  // get actual port name and save it to member variable
  _port_name = jack_port_name(_port);

  _convolver_left_ear = Convolver::create(_renderer.get_buffer_size());

  if (ptrtools::is_null(_convolver_left_ear))
  {
      WARNING("Initialization of convolver failed!");
      return false;
  }

  _convolver_right_ear = Convolver::create(_renderer.get_buffer_size());

  if (ptrtools::is_null(_convolver_right_ear))
  {
      WARNING("Initialization of convolver failed!");
      return false;
  }

  _convolver_dir_left = Convolver::create(_renderer.get_buffer_size());

  if (ptrtools::is_null(_convolver_dir_left))
  {
      WARNING("Initialization of convolver failed!");
      return false;
  }

  _convolver_dir_right = Convolver::create(_renderer.get_buffer_size());

  if (ptrtools::is_null(_convolver_dir_right))
  {
      WARNING("Initialization of convolver failed!");
      return false;
  }

  _audio_buffer.resize(_renderer.get_buffer_size(), 0.0f);

  // calculate transfer function of a dirac
  Convolver::data_t dirac_buffer(_renderer.get_buffer_size(), 0.0f);
  dirac_buffer[0] = 1.0f;
  //Convolver::data_t dirac_container;

  //Convolver::prepare_impulse_response(dirac_container, dirac_buffer.data(),
  Convolver::prepare_impulse_response(_dirac, dirac_buffer.data(),
      dirac_buffer.size(), _renderer.get_buffer_size());

  //_dirac = *dirac_container.begin();

  /*
  ////////////// prepare directivity //////////////////////////
  std::string directivity_file = "/home/j.ahrens/audio_files/impulse.wav";//"dipole_test.wav";

  // get sample rate of directivity coefficients
  const int sr = audiofileiotools::get_sample_rate(directivity_file);
  
  // if file could not be opened
  if (!sr) return false; 
  
  // if hrirs do not have the correct sample rate
  else if (sr != static_cast<int>(_renderer.get_sample_rate()))
  {
    ERROR("The coefficients of the directivity expansion in file '" << directivity_file
          << "' are sampled at rate " << sr 
          << " Hz. JACK is currently running at " 
          << _renderer.get_sample_rate() << " Hz.");
    
    // return false;
  }
  
  
  _directivity = (audiofileiotools::load_directivity_f(directivity_file, _renderer.get_buffer_size())).release();

  */

  //    if (ptrtools::is_null(*_directivity)) return false;

  return true;

}

bool BinauralRenderer::Input::fetch_new_buffer(jack_nframes_t nframes)
{
  sample_t* in = static_cast<sample_t*>(jack_port_get_buffer(_port, nframes));
  if (in == NULL)
  {
    WARNING("Got no input buffer from JACK!");
    return false;
  }
  assert(in != NULL);

  memcpy(&_audio_buffer[0], in, nframes*sizeof(sample_t));

  return true;
}

// TODO: rename this function:
void BinauralRenderer::Input::weight_audio_data(const float weight)
{
  //sample_t* audio_sample = &_audio_buffer[0];

  //// weight audio signal; C-style to be efficient
  //for (unsigned int n=0; n<_renderer.get_buffer_size(); n++)
  //{
  //  *audio_sample *= weight; audio_sample++;
  //}

  // update source signal level (max amplitude)
  _controller.set_source_signal_level(_source_id
      , _renderer.get_max_signal_amplitude(&_audio_buffer[0]
      , _renderer.get_buffer_size()) * _scene.get_source_gain(_source_id)
      * weight);
}

BinauralRenderer::sample_t* BinauralRenderer::Input::apply_hrtf_to_left_ear(
    Convolver::data_t& hrtf, const float interp_factor, float weighting_factor)
{
  // if hrtf has to be interpolated with the dirac
  if (interp_factor > 0.0f)
  {
    Convolver::data_t interp_filter = hrtf;

    const unsigned int partition_size = 2u*_renderer.get_buffer_size();
    const unsigned int overall_filter_size = interp_filter.size();

    float* filter_coefficient = interp_filter.data();
    float* dirac_coefficient = _dirac.data();

    // first partition, loop over all samples
    for (unsigned int sample = 0u; sample < partition_size; sample++)
    {
      // interpolate linearly between dirac and hrtfs
      (*filter_coefficient) = (*filter_coefficient) * (1.0f-interp_factor)
        + interp_factor * (*dirac_coefficient);

      filter_coefficient++; dirac_coefficient++;
    } // for

    // loop over all other samples
    for (unsigned int sample = partition_size; sample < overall_filter_size; sample++)
    {
      // interpolate linearly between dirac and hrtfs (Dirac is zero here???)
      (*filter_coefficient) *= (1.0f-interp_factor);

       filter_coefficient++;
    } // for

    _convolver_left_ear->set_filter_f(interp_filter);

  } // if
  else _convolver_left_ear->set_filter_f(hrtf);

  /*
  Convolver::filter_t tmp;
  float angle = 270.0f - ((*_scene.get_source_position(_source_id))+Position(-0.085f,0.0f)).orientation().azimuth - (*_scene.get_source_orientation(_source_id)).azimuth;
  _directivity->get_spectrum_of_plane_wave(tmp, angle);
  _convolver_dir_left->set_filter_f(tmp);
  */

  //  return _convolver_dir_left->convolve_signal(&_audio_buffer[0]); 
  // return _convolver_left_ear->convolve_signal(_convolver_dir_left->convolve_signal(&_audio_buffer[0])); 
  return _convolver_left_ear->convolve_signal(&_audio_buffer[0], weighting_factor);
}

BinauralRenderer::sample_t* BinauralRenderer::Input::apply_hrtf_to_right_ear(
    Convolver::data_t& hrtf, const float interp_factor, float weighting_factor)
{
   // if hrtf has to be interpolated with the dirac
  if (interp_factor > 0.0f)
  {
    Convolver::data_t interp_filter = hrtf;

    const unsigned int partition_size = 2u*_renderer.get_buffer_size();
    const unsigned int overall_filter_size = interp_filter.size();

    float* filter_coefficient = interp_filter.data();
    float* dirac_coefficient = _dirac.data();

    // first partition, loop over all samples
    for (unsigned int sample = 0u; sample < partition_size; sample++)
    {
      // interpolate linearly between dirac and hrtfs
      (*filter_coefficient) = (*filter_coefficient) * (1.0f-interp_factor)
        + interp_factor * (*dirac_coefficient);

      filter_coefficient++; dirac_coefficient++;
    } // for

    // loop over all other samples
    for (unsigned int sample = partition_size; sample < overall_filter_size; sample++)
    {
      // interpolate linearly between dirac and hrtfs (Dirac is zero here)
      (*filter_coefficient) *= (1.0f-interp_factor);

      filter_coefficient++;
    } // for

    _convolver_right_ear->set_filter_f(interp_filter);

  } // if
  else _convolver_right_ear->set_filter_f(hrtf);

  /*
  Convolver::filter_t tmp;
  float angle = 270.0f - ((*_scene.get_source_position(_source_id))+Position(-0.085f,0.0f)).orientation().azimuth - (*_scene.get_source_orientation(_source_id)).azimuth;
  _directivity->get_spectrum_of_plane_wave(tmp, angle);
  _convolver_dir_right->set_filter_f(tmp);
  */

  //  return _convolver_dir_right->convolve_signal(&_audio_buffer[0]);
  // return _convolver_right_ear->convolve_signal(_convolver_dir_right->convolve_signal(&_audio_buffer[0])); 
  return _convolver_right_ear->convolve_signal(&_audio_buffer[0], weighting_factor);
}

std::string BinauralRenderer::Input::get_port_name()
{
  return _port_name;
}

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
