/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * %BinauralRenderer class (definition).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef BINAURALRENDERER_H
#define BINAURALRENDERER_H

#include <map>
#include <set>
#include <vector>

#include "renderer.h"   // base class of BinauralRenderer
#include "directionalpoint.h" // Position and Orientation
#include "convolver.h"
//#include "fouriertools.h"

class Controller; // forward declaration

using ssr::id_t;

/** Binaural renderer class.
 * This class does the binaural rendering. Audio processing is done with the
 * inherited JACK functionality.
 **/
class BinauralRenderer : public Renderer
{
  private:
    class Output;          // nested class, defined later
    class Input;           // nested class, defined later
    /** Class to temporarily store selected source information.
     * By use of a class like this, the desired information for every source can
     * be obtained using the function Scene::get_source_list().
     **/
    struct SourceCopy : DirectionalPoint
    {
      typedef std::vector<SourceCopy> container_t; ///< list of SourceCopys
      /// type conversion constructor
      SourceCopy(const std::pair<id_t, Source>& other);

      id_t id;               ///< unique ID, see Scene::source_map_t
      Source::model_t model; ///< model, see Source
      float gain;            ///< gain, see Source
      bool mute;             ///< mute/unmute, see Source

      /// focused/non-focussed source, gets calculated at each audio block
      bool focused;
    };


  public:
    typedef std::auto_ptr<BinauralRenderer> ptr_t;

    /// a map which holds all input objects.
    typedef std::map<id_t, Input*> input_map_t;

    /// exception to be thrown by ctor.
    struct binaural_error : public std::runtime_error
    {
      binaural_error(const std::string& s): std::runtime_error(s) {}
    };

    BinauralRenderer(const Scene& scene, Controller& controller,
            const unsigned int hrir_size,
            const std::string& hrir_file_name,
            const std::string& client_name) throw (jack_error, binaural_error);

    /// "named constructor" for %BinauralRenderer objects
    static ptr_t create(const Scene& scene, Controller& controller,
            const unsigned int hrir_size,
            const std::string& hrir_file_name,
            const std::string& client_name = "Binaural-Renderer");

    /// delete all Input/Output objects and clear input and output map.
    virtual ~BinauralRenderer();

    /// HRTF set for one ear
    typedef std::vector<Convolver::data_t> hrtf_set_t;
    /// HRTFs for both ears 
      //typedef std::vector< hrtf_set_t > hrtfs_t;
    typedef std::pair<hrtf_set_t, hrtf_set_t> hrtfs_t;

    // from JackClient:
    virtual int jack_shutdown();
    // from Renderer:
    virtual bool new_source(id_t id, const std::string& port_name);
    virtual bool new_source(id_t id, const std::string& port_name,
			    const std::string& brir_file);
    virtual bool delete_source(id_t id);
    virtual void delete_all_sources();
    virtual void get_output_port_names(port_list_t& port_list) const;
    virtual unsigned int get_number_of_output_channels() const;
    virtual renderer_t get_renderer_type() const;

    /// Get pointer to input object corresponding to @a id.
    Input* get_input(id_t id);

    void clear_buffer(sample_t* const buffer, jack_nframes_t buffer_size) const;

  private:
    /// overwrites virtual member function declared in JackClient class
    virtual int jack_process(jack_nframes_t nframes);

    bool _load_hrtfs(hrtfs_t& hrtfs,
         const unsigned int hrir_size,
         const std::string& hrir_file_name);

    /// this will be filled with zeroes
    mutable std::vector<sample_t> _silence;

    input_map_t   _input_map;    ///< all input  objects are stored here
    Output* _output;             ///< we have only one output object

    // deprecated:
    /// relative angles of the array loudspeakers
    //std::map<float, Scene::loudspeakers_t::size_type> _loudspeaker_angles;

    Controller& _controller;          ///< Reference to the Controller
    const std::string  _input_prefix; ///< prefix for the input  ports
    const std::string _output_prefix; ///< prefix for the output ports

}; // end of BinauralRenderer class

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Output
////////////////////////////////////////////////////////////////////////////////

/** Manages an output buffer and the corresponding JACK calls.
 * @warning z coordinates and elevation are ignored
 **/
class BinauralRenderer::Output : public DirectionalPoint
{
  public:
    typedef std::auto_ptr<Output> ptr_t;
    /// "named constructor"
    static ptr_t create(BinauralRenderer& renderer,
        const std::string& port_name);

    ~Output(); ///< Unregister associated JACK output port

    /// Render audio data to the corresponding output object.
    void render(SourceCopy::container_t& source_list,
          const jack_nframes_t nframes,
	  const float master_volume = 1.0f,
          const float ampl_ref_dist = 3.0f);

    void get_port_names(port_list_t& port_list) const; ///< get JACK output port names

    /// clear output buffer
    void write_empty_buffer(jack_nframes_t nframes) const;

    /// use the assignment operator of the base class.
    using DirectionalPoint::operator=;

    void set_hrtfs(const hrtfs_t hrtfs);

    float get_signal_level(const jack_nframes_t nframes);
  private:
    Output(BinauralRenderer& renderer);

    hrtfs_t _hrtfs;

    bool _init(const std::string& port_name);

    BinauralRenderer& _renderer; /// ref. to the calling BinauralRenderer object

    jack_port_t* _port_left;     ///< JACK port associated with left ear output
    jack_port_t* _port_right;    ///< JACK port associated with right ear output
    std::string _port_name_left; ///< full name of the JACK port
    std::string _port_name_right; ///< full name of the JACK port

    Output(const Output&);            ///< deactivated
    Output& operator=(const Output&); ///< deactivated
};

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Input
////////////////////////////////////////////////////////////////////////////////

/** Manages an input buffer (including delay line) and the corresponding JACK
 * calls.
 **/
class BinauralRenderer::Input
{
  public:
    typedef std::auto_ptr<Input> ptr_t;
    /// "named constructor"
    static ptr_t create(const BinauralRenderer& renderer,
            const Scene& scene,
            Controller& controller,
            const id_t source_id,
            const std::string& name);

    /// Unregisters associated JACK input port
    ~Input();

    /// gets a buffer full of audio from JACK and writes it into the delay line
    bool fetch_new_buffer(jack_nframes_t nframes);

    void weight_audio_data(const float weight);

    BinauralRenderer::sample_t* apply_hrtf_to_left_ear(Convolver::data_t& hrtf,
                   const float interp_factor, float weighting_factor);
    BinauralRenderer::sample_t* apply_hrtf_to_right_ear(Convolver::data_t& hrtf,
                    const float interp_factor, float weighting_factor);

    std::string get_port_name(); ///< get JACK port name of the input object

  private:
    // if it were not private, I'd make it "explicit"
    Input(const BinauralRenderer& renderer,
          const Scene& scene,
          Controller& controller,
          const id_t source_id);

    /// creates a JACK port and maybe does some other things, too.
    bool _init(const std::string& port_name);

    /// holds a reference to the calling BinauralRenderer object
    const BinauralRenderer& _renderer;
    /// holds a reference to the scene

    const Scene& _scene;
    /// Reference to the Controller

    Controller& _controller;

    /// id of associated sound source
    id_t _source_id;

    /// the JACK port corresponding to this input object
    jack_port_t* _port;

    /// full name of the JACK port
    std::string _port_name;

    Convolver::ptr_t _convolver_left_ear;
    Convolver::ptr_t _convolver_right_ear;

    Convolver::ptr_t _convolver_dir_left;
    Convolver::ptr_t _convolver_dir_right;

    /// pointer to audio data of current frame
    std::vector<sample_t> _audio_buffer;
    Convolver::data_t _dirac; ///< in frequency domain

      //  fouriertools::PlaneWaveDecomposition* _directivity;

    Input(const Input&);            ///< deactivated
    Input& operator=(const Input&); ///< deactivated
};

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
