/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * Connection class (implementation). 
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include <boost/bind.hpp>

#include "connection.h"
#include "controller.h"

/// ctor
Connection::Connection(boost::asio::io_service &io_service
#ifdef ENABLE_SSL
    , boost::asio::ssl::context& context
#endif
    , Controller &controller)
  : _socket(io_service
#ifdef ENABLE_SSL
      , context
#endif
      )
  , _timer(io_service)
  , _controller(controller)
  , _subscriber(*this)
  , _commandparser(controller)
{}

/// dtor
Connection::~Connection()
{
  _controller.unsubscribe(&_subscriber);
  std::cout << "Connection destroyed" << std::endl;
}

/** Get an instance of Connection.
 * @param io_service 
 * @param controller used to (un)subscribe and get the actual Scene
 * @return ptr to Connection
 **/
Connection::pointer
Connection::create(boost::asio::io_service &io_service
#ifdef ENABLE_SSL
    , boost::asio::ssl::context& context
#endif
    , Controller &controller)
{
  return pointer(new Connection(io_service
#ifdef ENABLE_SSL
        , context
#endif
        , controller));
}

/** Start the connection.
 * - Subscribe this instance of Connection to the Controller.
 * - Send the actual scene over the network.
 * - Start reading incoming messages.
 * - Initialize the timer.
 **/
void
Connection::start()
{
#ifdef ENABLE_SSL
  boost::system::error_code ec;
  if (!!_socket.handshake(boost::asio::ssl::stream_base::server, ec))
  {
    // TODO: error message using ec.message()
    _timer.cancel();  // TODO: is this correct?
    return;
  }
#endif

  // ok... this Connection object is activated.
  // now we can connect the NetworkSubscriber.

  _controller.subscribe(&_subscriber);
  // this stuff should perhaps get refactored.
  // need to think about this. not sure if i like this mixed into
  // the Connection code. 
  std::string whole_scene = _controller.get_scene_as_XML();
  this->write(whole_scene);
  // And we can also start_read ing.
  start_read();

  // intialize the timer
  _timer.expires_from_now(boost::posix_time::milliseconds(100));
  _timer.async_wait(boost::bind(&Connection::timeout_handler, shared_from_this()
        , boost::asio::placeholders::error));
}

/** Send levels on timeout.
 * - Send level.
 * - Reset timer.
 * - Wait async.
 * @param e self explanatory
 **/
void
Connection::timeout_handler(const boost::system::error_code &e)
{
  if (e) return;

  _subscriber.send_levels();

  // Set timer again.
  _timer.expires_from_now(boost::posix_time::milliseconds(100));
  _timer.async_wait(boost::bind(&Connection::timeout_handler, shared_from_this()
        , boost::asio::placeholders::error));
}

/// Start reading from socket. 
void
Connection::start_read()
{
  async_read_until(_socket, _streambuf, '\0'
      , boost::bind(&Connection::read_handler, shared_from_this()
        , boost::asio::placeholders::error
        , boost::asio::placeholders::bytes_transferred));
}

/// Forward string from socket to CommandParser.
void
Connection::read_handler(const boost::system::error_code &error
    , size_t size)
{
  if (!error)
  {
    std::istream input_stream(&_streambuf);
    std::string  packet_string;
    getline(input_stream, packet_string, '\0');
    (void) size;
    //cout << "size= " << size << endl;
    //cout << "line: " << packet_string << endl;

    _commandparser.parse_cmd(packet_string);

    this->start_read();
  }
  else
  {
    _timer.cancel();
  }
}

/** Write to socket.
 * @param writestring: String to be send over the network. 
 **/
void
Connection::write(std::string &writestring)
{
  // Create a Copy of this string.
  // Put into shared_ptr bound to the callback
  // handler. This is sufficient to make it
  // be destroyed on exit.

  boost::shared_ptr<std::string> str_ptr(new std::string(writestring + '\0'));

  boost::asio::async_write(_socket, boost::asio::buffer(*str_ptr)
      , boost::bind(&Connection::write_handler, shared_from_this(), str_ptr
        , boost::asio::placeholders::error
        , boost::asio::placeholders::bytes_transferred));
}

/** Empty callback handler.  
 *
 * @param str_ptr String passed to Connection::write 
 * @param error error code
 * @param bytes_transferred self explanatory 
 *
 * @todo Check if we can delete this function.
 **/
void
Connection::write_handler(boost::shared_ptr<std::string> str_ptr
    , const boost::system::error_code &error, size_t bytes_transferred)
{
  (void) str_ptr;
  (void) error;
  (void) bytes_transferred;
  // Dont do nothing. we could check for error and stuff.
  // the shared_ptr was just used to keep the string alive and
  // destroy it now.
}

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
