/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * Binaural playback renderer class (implementation).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include <cassert>       // for assert()
#include <cmath>         // for cos()
#include "bpbrenderer.h"
#include "controller.h"  // for set_active_sources(), ...
#include "ssr_global.h"
#include "mathtools.h"    // for purge()
#include "ptrtools.h"    // for is_null(), destroy()
#include "stringtools.h" // for A2S(), S2A()
#include "audiofileiotools.h"

// for debugging only:
#include <libxml/debugXML.h>

using ptrtools::is_null;
using ptrtools::destroy;
using stringtools::S2A;
using stringtools::A2S;
using namespace ssr; // for id_t, processing, ready, ...

/** ctor.
 * @param scene see create()
 * @param controller see create()
 **/
BPBRenderer::BPBRenderer(const Scene& scene, Controller& controller,
    const std::string& client_name) 
  throw (jack_error, bpb_error) :
  Renderer(scene, client_name, true), // base class ctor
  _controller(controller),
  _output_prefix("out")
{

  // There shouldn't be any sources in the scene at this point.
  // If there are, they are ignored!

  if (this->get_buffer_size() == 0)
  {
    WARNING("Buffersize == 0!");
  }
  else
  {
    // fill silence buffer with zeroes
    _silence.resize(this->get_buffer_size(), 0);
  }

  // try to create output
  Output::ptr_t temp(Output::create(*this, _output_prefix));
  if (is_null(temp))
  {
    throw(bpb_error("Initialization of outputs failed!"));
  }
  _output = temp.release();

  _controller.publish_sample_rate(static_cast<unsigned int>(this->get_sample_rate()));
  _controller.set_processing_state(ready); // TODO: lock with mutex?

  this->activate(); // activate JACK client
}

BPBRenderer::~BPBRenderer()
{
  _controller.set_processing_state(exiting);

  // cond_signal to any waiting threads?

  delete _output;
  VERBOSE2("BPBRenderer: Output cleared.");

  maptools::purge(_input_map);
  VERBOSE2("BPBRenderer: Input map cleared.");
}

/** stop rendering.
 * This is only called by JACK. Do not dare to call it yourself!
 * - stops rendering
 * - calls its base class counterpart
 *
 * @return I don't know what
 */
int BPBRenderer::jack_shutdown()
{
  ERROR("Prematurely disconnected from JACK server! "
      "Trying to stop processing ...");
  _controller.set_processing_state(exiting);
  // cond_signal to any waiting threads?
  //maptools::purge(_output_map); WARNING("Output map cleared.");
  maptools::purge(_input_map);  WARNING("Input map cleared.");

  return JackClient::jack_shutdown(); // call original base class function
}

/** _.
 * @param scene Reference to the Scene object of the calling class. This
 * is used to obtain the geometry data in each JACK callback.
 * @param controller Reference to the Controller object. This is used to send
 * back information about active/inactive loudspeakers.
 * @param delayline_size Size of the delay lines in samples (all of them have
 * the same length).
 * @param initial_delay Amount of negative delay that can be used when reading
 * from the delay line
 * @param client_name see JackClient::create()
 * @return auto_ptr to the new BPBRenderer object.
 * @todo information about active loudspeakers per source should be passed to
 * the Controller.
 **/
BPBRenderer::ptr_t BPBRenderer::create(const Scene& scene,
    Controller& controller, const std::string& client_name)
{
  ptr_t temp; // temp = NULL
  try
  {
    temp.reset(new BPBRenderer(scene, controller, client_name));
  }
  catch(jack_error& e)
  {
    ERROR(e.what());
  }
  catch(bpb_error& e)
  {
    ERROR(e.what());
  }
  return temp;
}

bool BPBRenderer::new_source(id_t id, const std::string& port_name)
{
  (void) id;
  (void) port_name;
  
  ERROR("This function may not be called.");
  
  return false;
}

bool BPBRenderer::new_source(id_t id, const std::string& port_name,
			     const std::string& data_file_name)
{
  // input does not use a port 
  (void) port_name;

  if (_input_map.find(id) != _input_map.end())
  {
    ERROR("Input '" << id << "' already exists!");
    return false;
  }

  if (data_file_name.empty())
  {
    ERROR("Binaural data are not specified for source '" << port_name <<"'.");
    return false;
  }

  Input::ptr_t temp(Input::create(*this,
          _scene,
          _controller,
          id,
	  data_file_name));

  if (is_null(temp))
  {
    ERROR("Initialization of input '" << id << "' failed!");
    return false;
  }

  _input_map[id] = temp.release();

  VERBOSE("BPBRenderer: Input " << id << " registered.");
  return true;
}

bool BPBRenderer::delete_source(const id_t id)
{
  return maptools::delete_item(_input_map, id);
}

void BPBRenderer::delete_all_sources()
{
  maptools::purge(_input_map);
  // TODO: anything else?
}

/** _.
 * @param id identifier of the input channel
 * @return Pointer to the corresponding @a id. If none was found, @b NULL is
 * returned.
 **/
BPBRenderer::Input* BPBRenderer::_get_input(id_t id)
{
  return maptools::get_item(_input_map, id);
}

/** fill a buffer with zeroes.
 * @param buffer pointer to the memory to be cleared
 * @param buffer_size number of data elements to be cleared
 **/
void BPBRenderer::clear_buffer(sample_t* const buffer,
    const jack_nframes_t buffer_size) const
{
  if (buffer_size != _silence.size())
  {
    // this should never happen because JACK buffer size will not change!
    WARNING("BUG?: silence resized! (" << _silence.size() << " -> "
        << buffer_size << ")");
    _silence.resize(buffer_size, 0);
  }
  else
  {
    std::copy(_silence.begin(), _silence.end(), buffer);
  }
}

void BPBRenderer::get_output_port_names(port_list_t& port_list) const
{
  _output->get_port_names(port_list);
}

/** JACK process callback function.
 * This function is called every time the JACK server has new audio data.
 * @param nframes number of frames
 * @return I don't know what
 */
int BPBRenderer::jack_process(const jack_nframes_t nframes)
{
// code must be suitable for real-time execution.
// That means following function are not allowed:
// malloc, free, printf, pthread_mutex_lock, sleep,
// wait, poll, select, pthread_join, pthread_cond_wait,
// etc, etc.

  // clear all information about active/inactive loudspeakers
  _controller.clear_all_active_sources();
  // send some information to the controller (and then to the subscribers)
  _controller.set_transport_state(this->get_transport_state());
  _controller.set_cpu_load(this->get_cpu_load());

  switch (_scene.get_processing_state())
  {
    case processing:
      {

	// update reference position
        (*_output) = _scene.get_reference();

        // TODO: Necessary??
        _output->write_empty_buffer(nframes);

        // render sources
        _render(nframes);

        _controller.set_master_signal_level(_output->get_signal_level(nframes));

        break;
      }

    case ready: // audio processing is OFF
        _output->write_empty_buffer(nframes);
        break;

    case initializing:
      // WARNING("JACK process callback called while in \"initializing\" state!");
      // return something bad?

      break;

    case exiting:
      WARNING("JACK process callback called while in \"exiting\" state!");
      // return something bad?
      break;

    default:
      ERROR("BUG: Unknown client state!");
      // return something bad?
  }
  // TODO: find out what the return value means
  return 0;
}

void BPBRenderer::_render(const jack_nframes_t nframes)
{
  // get JACK output buffers
  sample_t* output_buffer_left  = static_cast<sample_t*>
    (jack_port_get_buffer(_output->get_port_left(), nframes));
  sample_t* output_buffer_right = static_cast<sample_t*>
    (jack_port_get_buffer(_output->get_port_right(), nframes));

  if (output_buffer_left == NULL || output_buffer_right == NULL)
  {
    ERROR("Cannot access JACK output ports.");
    return;
  }

  SourceCopy::container_t sources;
  _scene.get_sources(sources);

  for (SourceCopy::container_t::iterator i = sources.begin();
       i != sources.end(); ++i)
  {
    // input object must not be deleted! It's deleted in ~BPBRenderer()
    Input* current_input = _get_input(i->id);
    if (current_input == NULL)
    { // this happens regularly when a source is added and the renderer is
      // already processing because it takes some time until the JACK ports are
      // ready. Therefore the following warning is only printed if verbose > 1.
      VERBOSE2("BPBRenderer::Output: Input for source " << i->id
         << " doesn't exist in input map!");
      continue; // continue with next source
    }

    float weighting_factor = 1.0f;

    // if source is muted
    if (i->mute == true) 
    {
      // THE FOLLOWING CODE IS TO BE USED TO WHEN SOURCE IS SUPPOSED 
      // TO ALWAYS REPLAY FROM THE BEGINNING OF THE BUFFER WHEN UNMUTED
      // if playback pointer is not at beginning 
      // (this means that source has played at previous frame)
      if ( current_input->get_audio_position() != 0 )
      {
        current_input->fade_out(output_buffer_left,
                                output_buffer_right,
                                nframes);
      }
    
      // set playback pointer to the start
      current_input->reset_audio_position();
    
      continue; // next source

      // THE FOLLOWING CODE IS TO BE USED WHEN SOURCE DOES NOT
      // AUTOMATICALLY START REPLAYING FROM THE BEGINNING OF THE FILE
      // AFTER UNMUTING
      // weighting_factor = 0.0f;
    }

    // apply the gain factor of the current source
    weighting_factor *= i->gain;

    // apply master volume
    weighting_factor *= _scene.get_corrected_master_volume();

    assert(weighting_factor >= 0.0f);

    // get orientation of listener (relative source positions are NOT considered!)
    int head_orientation = static_cast<int>(_output->orientation.azimuth - 90 + 0.5f)%360;

    // finally, replay audio
    current_input->add_audio_frame_into_buffers(output_buffer_left,
                                                output_buffer_right,
                                                head_orientation,
                                                weighting_factor,
                                                nframes);

  } // for all sources

}

unsigned int BPBRenderer::get_number_of_output_channels() const
{
  return 2u;
}

Renderer::renderer_t BPBRenderer::get_renderer_type() const
{
  return Renderer::BPB;
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class BPBRenderer::SourceCopy
////////////////////////////////////////////////////////////////////////////////

BPBRenderer::SourceCopy::SourceCopy(const std::pair<id_t, Source>& other) :
  DirectionalPoint(other.second), // copy ctor of base class
  id(other.first),
  model(other.second.model),
  gain(other.second.gain),
  mute(other.second.mute)
{}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class BPBRenderer::Output
////////////////////////////////////////////////////////////////////////////////

/** ctor.
 * @param renderer calling BPBRenderer object
 **/
BPBRenderer::Output::Output(BPBRenderer& renderer) :
  _renderer(renderer)
{}

BPBRenderer::Output::~Output()
{
  if (_port_left) _renderer.unregister_port(_port_left);
  if (_port_right) _renderer.unregister_port(_port_right);
}

BPBRenderer::Output::ptr_t BPBRenderer::Output::create(
    BPBRenderer& renderer, const std::string& port_name)
{
  ptr_t temp(new Output(renderer));
  if (!temp->_init(port_name))
  {
    ERROR("Unable to initialize output port '" + port_name + "'!");
    destroy(temp);
  }
  return temp;
}

/** create a JACK output port and get its full name.
 * @param port_name the desired output port name (the part after the ":")
 *
 * The full port name including the client name can be obtained by calling
 * get_port_name() afterwards.
 **/
bool BPBRenderer::Output::_init(const std::string& port_name)
{
  _port_left  = _renderer.register_out_port(port_name + "_left");
  _port_right = _renderer.register_out_port(port_name + "_right");

  if (_port_left == NULL || _port_right == NULL)
  {
    WARNING("Output ports for BRS-Renderer could not be registered.");
    return false;
  }

  // get actual port names and save it to member variables
  _port_name_left  = jack_port_name(_port_left);
  _port_name_right = jack_port_name(_port_right);

  return true;
}

void BPBRenderer::Output::get_port_names(port_list_t& port_list) const
{
  port_list.push_back(_port_name_left);
  port_list.push_back(_port_name_right);
}

jack_port_t* BPBRenderer::Output::get_port_left() const
{
  return _port_left;
}

jack_port_t* BPBRenderer::Output::get_port_right() const
{
  return _port_right;
}

float BPBRenderer::Output::get_signal_level(const jack_nframes_t nframes)
{
  // get JACK output buffer
  sample_t* out_left;
  out_left = static_cast<sample_t*>(jack_port_get_buffer(_port_left, nframes));

  sample_t* out_right;
  out_right = static_cast<sample_t*>(jack_port_get_buffer(_port_right, nframes));

  // get maximum amplitude for right ear and return
  return std::max( _renderer.get_max_signal_amplitude(out_left, nframes),
       _renderer.get_max_signal_amplitude(out_right, nframes));
}

void BPBRenderer::Output::write_empty_buffer(const jack_nframes_t nframes) const
{
  if (_port_left)
  {
    sample_t *out;
    out = static_cast<sample_t*>(jack_port_get_buffer(_port_left, nframes));
    _renderer.clear_buffer(out, nframes);
  }
  if (_port_right)
  {
    sample_t *out;
    out = static_cast<sample_t*>(jack_port_get_buffer(_port_right, nframes));
    _renderer.clear_buffer(out, nframes);
  }
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class BPBRenderer::Input
////////////////////////////////////////////////////////////////////////////////

/** private ctor. Object can only be created with create().
 * @param renderer reference to the calling BPBRenderer object. Using this, we
 * can access the public members of the BRS renderer.
 **/
BPBRenderer::Input::Input(const BPBRenderer& renderer,
        const Scene& scene,
        Controller& controller,
        const id_t source_id) :
  _number_of_head_orientations(0),
  _audio_position(0),
  _data_size(0),
  _previous_head_orientation(0),
  _previous_weight(0),
  _renderer(renderer),
  _scene(scene),
  _controller(controller),
  _source_id(source_id)
{}

BPBRenderer::Input::~Input()
{}

BPBRenderer::Input::ptr_t BPBRenderer::Input::create(
        const BPBRenderer& renderer,
        const Scene& scene,
        Controller& controller,
        const id_t source_id,
	const std::string& data_file_name)
{
  ptr_t temp(new Input(renderer, scene, controller, source_id));
  if (!temp->_init(data_file_name))
  {
    ERROR("Unable to initialize source '" + data_file_name + "'!");
    destroy(temp);
  }
  return temp;
}

bool BPBRenderer::Input::_init(const std::string& data_file_name)
{
  // stop processing until data are loaded
  //_controller.set_processing_state(initializing, 0);
  
  _audio_buffer.resize(_renderer.get_buffer_size(), 0.0f);
  
  if (!_load_binaural_data(data_file_name)) return false;
      
  return true;

}

bool BPBRenderer::Input::_load_binaural_data(const std::string& file_name)
{

 // get sample rate of hrirs
  const int sr = audiofileiotools::get_sample_rate(file_name);
  if (!sr) return false;

  // if hrirs do not have the correct sample rate
  else if (sr != static_cast<int>(_scene.get_sample_rate()))
  {
    ERROR("The data in file '"
          + file_name + "' are sampled at rate " + A2S(sr)
          + " Hz. JACK is currently running at "
          + A2S(_scene.get_sample_rate()) + " Hz.");

    return false;
  }

  const int no_of_ch = audiofileiotools::get_no_of_channels(file_name);

  // error message comes from function get_no_of_channels()
  if ( no_of_ch == 0) return false;
  // check if no of channels makes sense
  else if ( no_of_ch%2 )
  {
    ERROR("File '" << file_name 
               << "' has to have an even number of channels.");
    return false;
  }
  else if ( (no_of_ch-2) % 4 )
  {
    ERROR("File '" << file_name 
       << "' does not represent a symmetric interval of head orientations.");
    return false;
  }

  _number_of_head_orientations = no_of_ch / 2;

  std::cout << "Loading " << _number_of_head_orientations << 
     " head orientations from file '" << file_name  << "'. Please be patient.\n";

  // load data
  std::vector<float> container;

  if (!audiofileiotools::load_filter_t(file_name, container))
  {
    return false;
  }

  ////////////////////////// audio_position_t ///////////////////////////////////////
  _data_size = container.size() / (2*_number_of_head_orientations);

  // clear data container
  _binaural_data.first.clear();
  _binaural_data.second.clear();

  // initialize data container
  for (unsigned int n = 0; n < _number_of_head_orientations; n++)
  {
    std::vector<float> dummy;

    dummy.clear();

    // initialize
    _binaural_data.first.push_back(dummy);
    _binaural_data.second.push_back(dummy);
  }

  // step through container and copy binaural data channels
  // iterate over channels
  for (unsigned int c = 0; c < _number_of_head_orientations; c++ ) 
  {  

    for (unsigned int n = 0; n < _data_size; ++n ) 
    {
      // left ear
      _binaural_data.first[c].push_back( container[(2*c) * _data_size  + n] ); 
      // right ear
      _binaural_data.second[c].push_back( container[(2*c+1) * _data_size  + n] );
    }   

  }

  VERBOSE("Data of size " << _data_size << " loaded.");

  return true;
}

void BPBRenderer::Input::add_audio_frame_into_buffers(
                                sample_t* audio_sample_out_left,
                                sample_t* audio_sample_out_right,
                                int head_orientation,
                                const float weight,
				const jack_nframes_t nframes)
{

    // make sure that orientation is in proper range
    if (head_orientation < 0) head_orientation += 360;

    // clamp head orentation to the most extreme available values
    if ( head_orientation > 
                         static_cast<int>( (_number_of_head_orientations-1)/2 )
         && head_orientation < 181 )
    {
      head_orientation = (_number_of_head_orientations-1)/2;
    }
    else if ( head_orientation > 180 
              && head_orientation < 
                 static_cast<int>( 360 - (_number_of_head_orientations-1)/2 ) )
    {
      head_orientation = 360 - (_number_of_head_orientations-1)/2;
    }

    // now we have to make sure that a given head_orientation is translated
    // correctly into an index
    if ( head_orientation > 
                       static_cast<int>( (_number_of_head_orientations-1)/2 ) )
    {
      head_orientation = 
        _number_of_head_orientations - (360 - head_orientation);
    }

    /////////// from here on the head_orientation is correct ////////////////
    // THE FOLLOWING CODE IS TO BE USED TO WHEN SOURCE IS SUPPOSED 
    // TO ALWAYS REPLAY FROM THE BEGINNING OF THE BUFFER WHEN UNMUTED
    // if we are at the beginning don't do crossfade
    if ( _audio_position == 0 ) 
    {
      _previous_head_orientation = head_orientation;
      _previous_weight = weight;
    }

    // add audio data to output buffers
    for(unsigned int n = 0; n < nframes; n++)
    {
      // left ear
      (*audio_sample_out_left)  += 
         // fade-in with raised cosine window
         static_cast<sample_t>(0.5f
         + 0.5f * cos(static_cast<float>(nframes-n)
         / static_cast<float>(nframes) * mathtools::PI)) * 
         weight * _binaural_data.first[head_orientation][_audio_position] + 
         
         // fade-out with raised cosine window
         static_cast<sample_t>(0.5f
         + 0.5f * cos(static_cast<float>(n)
         / static_cast<float>(nframes) * mathtools::PI)) * 
         _previous_weight * _binaural_data.first[_previous_head_orientation][_audio_position];

       // right ear
      (*audio_sample_out_right) +=                   
         // fade-in with raised cosine window
         static_cast<sample_t>(0.5f
         + 0.5f * cos(static_cast<float>(nframes-n)
         / static_cast<float>(nframes) * mathtools::PI)) * 
         weight * _binaural_data.second[head_orientation][_audio_position] + 
         
         // fade-out with raised consine window
         static_cast<sample_t>(0.5f
         + 0.5f * cos(static_cast<float>(n)
         / static_cast<float>(nframes) * mathtools::PI)) * 
         _previous_weight * _binaural_data.second[_previous_head_orientation][_audio_position];


      audio_sample_out_left++;  
      audio_sample_out_right++; 
      _audio_position++;

      // if end of data is reached
      if ( _audio_position >= _data_size )
      {
	// stop playback and start from beginning
        _audio_position = 0;
        return; 
      }
    }
   
    _previous_head_orientation = head_orientation;
    _previous_weight = weight;
}
    
void BPBRenderer::Input::reset_audio_position()
{
     _audio_position = 0;
}

BPBRenderer::Input::audio_position_t BPBRenderer::Input::get_audio_position()
{
  return _audio_position;
}

void BPBRenderer::Input::fade_out(sample_t* audio_sample_out_left,
                                sample_t* audio_sample_out_right,
				const jack_nframes_t nframes)

{
    // add audio data to output buffers
    for(unsigned int n = 0; n < nframes; n++)
    {
      // left ear
      (*audio_sample_out_left)  += 
         // fade-out with raised consine window
         static_cast<sample_t>(0.5f
         + 0.5f * cos(static_cast<float>(n)
         / static_cast<float>(nframes) * mathtools::PI)) * 
         _previous_weight * _binaural_data.first[_previous_head_orientation][_audio_position];

       // right ear
      (*audio_sample_out_right) +=                          
         // fade-out with raised consine window
         static_cast<sample_t>(0.5f
         + 0.5f * cos(static_cast<float>(n)
         / static_cast<float>(nframes) * mathtools::PI)) * 
         _previous_weight * _binaural_data.second[_previous_head_orientation][_audio_position];


      audio_sample_out_left++;  
      audio_sample_out_right++; 
      _audio_position++;

      // if end of data is reached
      if ( _audio_position >= _data_size )
      {
	// stop playback and start from beginning
        _audio_position = 0;
        return; 
      }
    } // for n

}

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
