/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * %Binaural playback renderer class (definition).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef BPBRENDERER_H
#define BPBRENDERER_H

#include <map>
#include <set>
#include <vector>
#include <sndfile.h> // for sf_count_t

#include "renderer.h"   // base class of BPBRenderer
#include "directionalpoint.h" // Position and Orientation
//#include "fouriertools.h"

class Controller; // forward declaration

using ssr::id_t;

/** Binaural Room Scanning Renderer class.
 * This class does BRS rendering. Audio processing is done with the
 * inherited JACK functionality.
 **/
class BPBRenderer : public Renderer
{
  private:
    class Output;          // nested class, defined later
    class Input;           // nested class, defined later
    /** Class to temporarily store selected source information.
     * By use of a class like this, the desired information for every source can
     * be obtained using the function Scene::get_source_list().
     **/
    struct SourceCopy : DirectionalPoint
    {
      typedef std::vector<SourceCopy> container_t; ///< list of SourceCopys
      /// type conversion constructor
      SourceCopy(const std::pair<id_t, Source>& other);

      id_t id;               ///< unique ID, see Scene::source_map_t
      Source::model_t model; ///< model, see Source
      float gain;            ///< gain, see Source
      bool mute;             ///< mute/unmute, see Source

      /// focused/non-focussed source, gets calculated at each audio block
      bool focused;
    };


  public:
    typedef std::auto_ptr<BPBRenderer> ptr_t;

    /// a map which holds all input objects.
    typedef std::map<id_t, Input*> input_map_t;

    /// exception to be thrown by ctor.
    struct bpb_error : public std::runtime_error
    {
      bpb_error(const std::string& s): std::runtime_error(s) {}
    };

    BPBRenderer(const Scene& scene, Controller& controller,
            const std::string& client_name) throw (jack_error, bpb_error);

    /// "named constructor" for %BPBRenderer objects
    static ptr_t create(const Scene& scene, Controller& controller,
            const std::string& client_name = "BPB-Renderer");

    /// delete all Input/Output objects and clear input and output map.
    virtual ~BPBRenderer();

    /// binaural data set for one ear
    typedef std::vector< std::vector<float> > binaural_data_set_t;
    /// binaural data set for both ears 
    typedef std::pair<binaural_data_set_t, binaural_data_set_t> binaural_data_t;

    // from JackClient:
    virtual int jack_shutdown();
    // from Renderer:
    virtual bool new_source(id_t id, const std::string& port_name);
    virtual bool new_source(id_t id, const std::string& port_name,
			    const std::string& data_file_name);
    virtual bool delete_source(id_t id);
    virtual void delete_all_sources();
    virtual void get_output_port_names(port_list_t& port_list) const;
    virtual unsigned int get_number_of_output_channels() const;
    virtual renderer_t get_renderer_type() const;

    void clear_buffer(sample_t* const buffer, jack_nframes_t buffer_size) const;

  private:
    /// overwrites virtual member function declared in JackClient class
    virtual int jack_process(jack_nframes_t nframes);

    /// this will be filled with zeroes
    mutable std::vector<sample_t> _silence;

    input_map_t   _input_map;    ///< all input  objects are stored here
    Output* _output;             ///< we have only one output object

    // deprecated:
    /// relative angles of the array loudspeakers
    //std::map<float, Scene::loudspeakers_t::size_type> _loudspeaker_angles;

    Controller& _controller;          ///< Reference to the Controller
    const std::string  _input_prefix; ///< prefix for the input  ports
    const std::string _output_prefix; ///< prefix for the output ports

    /// Get pointer to input object corresponding to @a id.
    Input* _get_input(id_t id);

    /// Render audio data to the corresponding outputs.
    void _render(const jack_nframes_t nframes);

}; // end of BPBRenderer class

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Output
////////////////////////////////////////////////////////////////////////////////

/** Manages an output buffer and the corresponding JACK calls.
 * @warning z coordinates and elevation are ignored
 **/
class BPBRenderer::Output : public DirectionalPoint
{
  public:
    typedef std::auto_ptr<Output> ptr_t;
    /// "named constructor"
    static ptr_t create(BPBRenderer& renderer,
        const std::string& port_name);

    ~Output(); ///< Unregister associated JACK output port

    void get_port_names(port_list_t& port_list) const; ///< get JACK output port names

    /// clear output buffer
    void write_empty_buffer(jack_nframes_t nframes) const;

    /// use the assignment operator of the base class.
    using DirectionalPoint::operator=;

    float get_signal_level(const jack_nframes_t nframes);

    jack_port_t* get_port_left() const;
    jack_port_t* get_port_right() const;

  private:
    Output(BPBRenderer& renderer);

    bool _init(const std::string& port_name);

    BPBRenderer& _renderer; /// ref. to the calling BPBRenderer object

    jack_port_t* _port_left;     ///< JACK port associated with left ear output
    jack_port_t* _port_right;    ///< JACK port associated with right ear output
    std::string _port_name_left; ///< full name of the JACK port
    std::string _port_name_right; ///< full name of the JACK port

    Output(const Output&);            ///< deactivated
    Output& operator=(const Output&); ///< deactivated
};

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Input
////////////////////////////////////////////////////////////////////////////////

/** Manages an input buffer (including delay line) and the corresponding JACK
 * calls.
 **/
class BPBRenderer::Input
{
  public:

    typedef std::auto_ptr<Input> ptr_t;
    typedef sf_count_t audio_position_t;

    /// "named constructor"
    static ptr_t create(const BPBRenderer& renderer,
            const Scene& scene,
            Controller& controller,
            const id_t source_id,
            const std::string& data_file_name);

    ~Input();

    void add_audio_frame_into_buffers(sample_t* audio_sample_left,
                                      sample_t* audio_sample_right,
                                      int head_orientation,
                                      const float weight,
                                      const jack_nframes_t nframes);
    
    void reset_audio_position();

    void fade_out(sample_t* audio_sample_out_left,
                  sample_t* audio_sample_out_right,
                  const jack_nframes_t nframes);
    
    audio_position_t get_audio_position();

  private:

    // if it were not private, I'd make it "explicit"
    Input(const BPBRenderer& renderer,
          const Scene& scene,
          Controller& controller,
          const id_t source_id);

    unsigned int _number_of_head_orientations;
    audio_position_t _audio_position;
    audio_position_t _data_size;
    unsigned int _previous_head_orientation;
    float _previous_weight;

    binaural_data_t _binaural_data;

    /// loads binaural data from file.
    bool _init(const std::string& data_file_name);

    bool _load_binaural_data(const std::string& data_file_name);

    /// holds a reference to the calling BPBRenderer object
    const BPBRenderer& _renderer;
    /// holds a reference to the scene

    const Scene& _scene;
    /// Reference to the Controller

    Controller& _controller;

    /// id of associated sound source
    id_t _source_id;

    /// pointer to audio data of current frame
    std::vector<sample_t> _audio_buffer;

    Input(const Input&);            ///< deactivated
    Input& operator=(const Input&); ///< deactivated
};

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
