/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * %Controller (implementation).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include <algorithm>
#include <errno.h>
#include <fstream>
#include <cassert> // for assert()
#include <cstdlib> // for rand()
#include <ctime>   // for time()
#include <cmath>   // for cos(), sin()

#include <libxml/xmlsave.h> // temporary hack!

#include "controller.h"
#include "scene.h"
#include "ptrtools.h"
#include "stringtools.h"
#include "posixpathtools.h"
#include "xmlparser.h"
#include "mathtools.h" // for dB2linear()
#include "audiofileiotools.h"

#include "rendersubscriber.h"

#ifdef ENABLE_MCP
#include "mcpconnector_static.h"
#endif

#ifdef ENABLE_INTERSENSE
#include "trackerintersense.h"
#endif
#ifdef ENABLE_POLHEMUS
#include "trackerpolhemus.h"
#endif
#ifdef ENABLE_RAZOR
#include "trackerrazor.h"
#endif

using ptrtools::is_null;
using stringtools::A2S;
using stringtools::S2A;
using stringtools::S2RV;
using posixpathtools::make_path_relative_to_current_dir;
using posixpathtools::make_path_relative_to_file;
using posixpathtools::get_file_extension;

typedef XMLParser::Node Node; ///< a node of the DOM tree

Controller::Controller(const std::string& schema_file_name
    , const std::string& input_port_prefix, const float ampl_ref_dist)
  : _scene(MasterScene::get()) // singleton!
  , _schema_file_name(schema_file_name)
  , _input_port_prefix(input_port_prefix)
  , _tracker(0)
  , _first_loudspeaker_channel(1)
  , _stand_ampl_ref_dist(ampl_ref_dist)
  , _loop(false)
{
  // _scene is always in the subscriber list. Additionally, there can be also
  // _scene_player, _ip_server and maybe more ...
  _subscribe(&_scene);
}

Controller::~Controller()
{
  _subscribers.clear();
}

/// unnamed namespace to encapsulate functions which are only used within
/// controller.cpp
namespace
{
  struct PositionPlusBool : Position
  {
    PositionPlusBool()
                                   :                fixed(false) {}
    explicit PositionPlusBool(Position pos, const bool fixed = false)
                                   : Position(pos), fixed(fixed) {}
    PositionPlusBool(const float x, const float y, const bool fixed = false)
                                   : Position(x,y), fixed(fixed) {}

    bool fixed;
  };

  /// find position in child nodes.
  /// Traverse through all child nodes of @a node and check for a @b position
  /// element.
  /// @param node parent node
  /// @return std::auto_ptr to the obtained position, empty auto_ptr if no
  /// position element was found.
  /// @warning If you want to extract e.g. position and orientation it would be
  /// more effective to do both (or even more) in one loop. But anyway, this
  /// seems easier to use.
  std::auto_ptr<PositionPlusBool>
  get_position(const Node& node)
  {
    std::auto_ptr<PositionPlusBool> temp; // temp = NULL
    if (!node) return temp; // return NULL

    for (Node i = node.child(); !!i; ++i)
    {
      if (i == "position")
      {
        float x, y;
        bool fixed;

        // if read operation successful
        if (S2A(i.get_attribute("x"), x) && S2A(i.get_attribute("y"), y))
        {
          // "fixed" indicated
          if (S2A(i.get_attribute("fixed"), fixed))
          {
            temp.reset(new PositionPlusBool(x, y, fixed));
          }
          else // "fixed" not indicated
          {
            temp.reset(new PositionPlusBool(x, y));
          }

          return temp; // return sucessfully
        }
        else
        {
          ERROR("Invalid position!");
          return temp; // return NULL
        } // if read operation successful

      } // if (i == "position")
    }
    return temp; // return NULL
  }

  /// find orientation in child nodes.
  /// @param node parent node
  /// @see get_position
  std::auto_ptr<Orientation>
  get_orientation(const Node& node)
  {
    std::auto_ptr<Orientation> temp; // temp = NULL
    if (!node) return temp;          // return NULL
    for (Node i = node.child(); !!i; ++i)
    {
      if (i == "orientation")
      {
        float azimuth;
        if (S2A(i.get_attribute("azimuth"), azimuth))
        {
          temp.reset(new Orientation(azimuth));
          return temp; // return sucessfully
        }
        else
        {
          ERROR("Invalid orientation!");
          return temp; // return NULL
        }
      }
    }
    return temp;       // return NULL
  }

  /// get an angle from the child nodes.
  /// This is nearly the same as get_orientation().
  /// @param node parent node
  /// @return std::auto_ptr to Orientation, which represents the found angle. If
  /// none was found, an empty auto_ptr is returned.
  std::auto_ptr<Orientation>
  get_angle(const Node& node)
  {
    std::auto_ptr<Orientation> temp; // temp = NULL
    if (!node) return temp;          // return NULL
    for (Node i = node.child(); !!i; ++i)
    {
      if (i == "angle")
      {
        float azimuth;
        if (S2A(i.get_attribute("azimuth"), azimuth))
        {
          temp.reset(new Orientation(azimuth));
          return temp; // return sucessfully
        }
        else
        {
          ERROR("Invalid angle!");
          return temp; // return NULL
        }
      }
    }
    return temp;       // return NULL
  }

  /** get attribute of a node.
   * @param node the node you want to have the attribute of.
   * @param attribute name of attribute
   * @param default_value default return value if something goes wrong
   * @return value default_value on error.
   **/
  template<typename T>
  T
  get_attribute_of_node(const Node& node, const std::string attribute
      , const T default_value)
  {
    if (!node) return default_value;
    return S2RV(node.get_attribute(attribute), default_value);
  }

  /** check for file/port
   * @param node parent node
   * @param file_or_port_name a string where the obtained file- or portname is
   * stored
   * @return channel number, 0 if port was given.
   * @note on error, file_or_port_name is set to the empty string "" and 0 is
   * returned.
   **/
  unsigned int
  get_file_or_port_name(const Node& node, std::string& file_or_port_name)
  {
    for (Node i = node.child(); !!i; ++i)
    {
      if (i == "file")
      {
        file_or_port_name = get_content(i);
        return S2RV(i.get_attribute("channel"), 1U);
      }
      else if (i == "port")
      {
        file_or_port_name = get_content(i);
        return 0;
      }
    }
    // nothing found:
    file_or_port_name = "";
    return 0;
  }

  /// helper function to load a circular array of loudspeakers.
  void
  load_circular_array(const Node& node, Loudspeaker::container_t& loudspeakers)
  {
    if (!node) return;

    int number;
    std::auto_ptr<PositionPlusBool> first_pos, center_pos;
    std::auto_ptr<Orientation>      first_dir, second_dir, last_dir;

    if (!S2A(node.get_attribute("number"), number) || (number < 2))
    {
      ERROR("Invalid number of loudspeakers! Array not loaded.");
      return;
    }

    // reserve memory for the loudspeakers
    loudspeakers.reserve(number + loudspeakers.size());

    center_pos = get_position   (node.child("center"));
    first_pos  = get_position   (node.child("first"));
    first_dir  = get_orientation(node.child("first"));
    second_dir = get_angle      (node.child("second"));
    last_dir   = get_angle      (node.child("last"));

    if (is_null(first_pos))
    {
      ERROR("No position found for first loudspeaker! Array not loaded!");
      return;
    }
    if (is_null(first_dir))
    {
      ERROR("No orientation found for first loudspeaker! Array not loaded!");
      return;
    }

    // if no center is given, it is set to the origin
    if (is_null(center_pos))
    {
      center_pos.reset(new PositionPlusBool(0,0));
    }

    // WARNING: The center is not supposed to have an orientation!
    DirectionalPoint radius(*first_pos - *center_pos, *first_dir);
    DirectionalPoint center(*center_pos, Orientation());

    float angle_increment = 360.0f / number; // full circle
    // division by zero not possible, as number >= 2 (see above)

    if (!is_null(second_dir))
    {
      angle_increment = second_dir->azimuth;
    }
    else if (!is_null(last_dir))
    {
      angle_increment = last_dir->azimuth / (number - 1);
      // division by zero not possible, as number >= 2 (see above)
    }

    for (int i = 0; i < number; ++i)
    {
      // Loudspeakers are ordered in mathematic positive direction, i.e.
      // counterclockwise (seen from above).

      // TODO: get loudspeaker type!

      loudspeakers.push_back(
          Loudspeaker(center + radius.rotate(i * angle_increment)));
    }
  }

  /// helper function to load a linear array of loudspeakers.
  void
  load_linear_array(const Node& node, Loudspeaker::container_t& loudspeakers)
  {
    if (!node) return;

    int number;
    if (!S2A(node.get_attribute("number"), number))
    {
      ERROR("Invalid number of loudspeakers! Array not loaded.");
      return;
    }

    if (number < 2)
    {
      // This is also checked in the Schema validation.
      ERROR("Number of loudspeakers in an array must be at least 2!");
      return;
    }
    // reserve memory for the loudspeakers
    loudspeakers.reserve(number + loudspeakers.size());

    std::auto_ptr<PositionPlusBool> first_pos, second_pos, last_pos;
    std::auto_ptr<Orientation>      first_dir, second_dir, last_dir;

    // we search the first occurence of each element, but
    // the XML Schema should assure that each element comes only once.
    first_pos  = get_position   (node.child("first"));
    first_dir  = get_orientation(node.child("first"));
    second_pos = get_position   (node.child("second"));
    second_dir = get_orientation(node.child("second"));
    last_pos   = get_position   (node.child("last"));
    last_dir   = get_orientation(node.child("last"));

    if (is_null(first_pos))
    {
      ERROR("No position found for first loudspeaker! Array not loaded!");
      return;
    }
    if (is_null(first_dir))
    {
      ERROR("No orientation found for first loudspeaker! Array not loaded!");
      return;
    }
    const DirectionalPoint first(*first_pos, *first_dir);
    DirectionalPoint increment; // default ctor. will be overwritten

    if (!is_null(second_pos) && is_null(last_pos))
    {
      // if no orientation was given for "second", it gets the same as "first"
      if (is_null(second_dir)) second_dir.reset(new Orientation(*first_dir));
      const DirectionalPoint second(*second_pos, *second_dir);
      increment = second - first;
    }
    else if (!is_null(last_pos) && is_null(second_pos))
    {
      // if no orientation was given for "last", it gets the same as "first"
      if (is_null(last_dir)) last_dir.reset(new Orientation(*first_dir));
      const DirectionalPoint last(*last_pos, *last_dir);
      increment = (last - first) / (number - 1);
    }
    else
    {
      ERROR("Either \"second\" or \"last\" has to be present! (But not both!) "
          "Array not loaded.");
      return;
    }

    DirectionalPoint current = first;
    for (int i = 0; i < number; ++i)
    {
      // TODO: get loudspeaker type!

      loudspeakers.push_back(Loudspeaker(current));
      current += increment;
    }
  }

  /// helper function to load a loudspeaker.
  void
  load_loudspeaker(const Node& node, Loudspeaker::container_t& loudspeakers)
  {
    if (!node) return;

    std::auto_ptr<PositionPlusBool> position;
    std::auto_ptr<Orientation>      orientation;

    if (ssr::verbose)
    {
      std::string name = node.get_attribute("name");
      if (name != "")
      {
        VERBOSE("Loading loudspeaker \"" << name << "\".");
      }
    }

    position    = get_position   (node);
    orientation = get_orientation(node);

    if (is_null(position))
    {
      ERROR("No position found for the loudspeaker! Not loaded!");
      return;
    }
    else if (is_null(orientation))
    {
      ERROR("No orientation found for the loudspeaker! Not loaded!");
      return;
    }

    std::string model_str = node.get_attribute("model");
    Loudspeaker::model_t model;
    if      (model_str == "normal")    model = Loudspeaker::normal;
    else if (model_str == "subwoofer") model = Loudspeaker::subwoofer;
    else                               model = Loudspeaker::unknown;

    float delay = get_attribute_of_node(node, "delay", 0.0f);

    // check validity
    if (delay < 0)
    {
      ERROR("Loudspeaker delay can not be negative. I'll ignore it");
      delay = 0.0f;
    }

    float weight = get_attribute_of_node(node, "weight", 1.0f);

    loudspeakers.push_back(Loudspeaker(DirectionalPoint(*position,*orientation)
          , model, weight, delay));
  }
} // end of anonymous namespace

// Public version of subscribe... private method will soon
// be phased out.
//
// FIXME: THREAD
void
Controller::subscribe(Subscriber* const subscriber)
{
  _subscribers.push_back(subscriber);
}

// FIXME: THREAD
void
Controller::unsubscribe(Subscriber* const subscriber)
{
  subscriber_list_t::iterator i;
  for (i=_subscribers.begin(); i!=_subscribers.end(); i++)
  {
    if (*i == subscriber)
    {
      _subscribers.erase(i);
      return;
    }
  }
  std::cerr << "Deletion of unsubscribed Subscriber requested" << std::endl;
}

void
Controller::_subscribe(Subscriber* const subscriber)
{
  _subscribers.push_back(subscriber);
}

bool
Controller::load_loudspeaker_setup(const std::string& setup_file_name)
{
  if (setup_file_name == "")
  {
    ERROR("No loudspeaker setup file given!");
    return false;
  }
  if (!is_null(_renderer))
  {
    ERROR("Renderer already running!");
    return false;
  }

  // read the setup from XML file
  XMLParser xp; // load XML parser
  XMLParser::doc_t setup_file = xp.load_file(setup_file_name);
  if (is_null(setup_file))
  {
    ERROR("Unable to load loudspeaker setup file '" << setup_file_name << "'!");
    return false;
  }
  if (_schema_file_name == "")
  {
    VERBOSE("No schema file specified!");
    // TODO: return true and continue anyway?
    return false;
  }
  else if (setup_file->validate(_schema_file_name))
  {
    VERBOSE("Valid loudspeaker setup (" << setup_file_name << ").");
  }
  else
  {
    ERROR("Error validating '" << setup_file_name << "' with schema '"
        << _schema_file_name << "'!");
    return false;
  }

  XMLParser::xpath_t xpath_result;
  xpath_result = setup_file->eval_xpath("//reproduction_setup/*");

  // one could use separate xpath expressions like
  // "//reproduction_setup/loudspeaker",
  // "//reproduction_setup/circular_array" and
  // "//reproduction_setup/linear_array",
  // but this would not preserve the order of the elements!

  if (is_null(xpath_result))
  {
    ERROR("No <reproduction_setup> section found in XML file!");
    return false;
  }

  // all loudspeakers specified in the XML file are written to the loudspeakers
  // container. This could be called a "Collecting Parameter" pattern.
  Loudspeaker::container_t loudspeakers;

  // it is not necessary to use xmlFreeNode on "node", it should be freed in the
  // destructor of xpath_result (hopefully?)
  for (Node node; (node = xpath_result->node()); ++(*xpath_result))
  {
    if (node == "loudspeaker")
    {
      load_loudspeaker(node, loudspeakers);
    }
    else if (node == "linear_array")
    {
      load_linear_array(node, loudspeakers);
    }
    else if (node == "circular_array")
    {
      load_circular_array(node, loudspeakers);
    }
    else if (node == "skip")
    {
      loudspeaker_id_t number;

      if (!S2A(node.get_attribute("number"), number) || (number < 1))
      {
        VERBOSE("Couldn't get \"number\" attribute of \"skip\" element."
            " Using default value 1.");
        number = 1;
      }
      for (loudspeaker_id_t i = 0; i < number; ++i)
      {
        _skipped_loudspeaker_channels.insert(_first_loudspeaker_channel
            + loudspeakers.size() + _skipped_loudspeaker_channels.size());
      }
    }
    else
    {
      WARNING("Ignored \"" << node << "\" element!");
    }
  }
  if (loudspeakers.empty())
  {
    ERROR("No loudspeakers found in the file '" << setup_file_name << "'!");
    return false;
  }

  _publish(&Subscriber::set_loudspeakers, loudspeakers);

  VERBOSE("Loaded " << loudspeakers.size() << " loudspeakers from '"
      << setup_file_name << "'.");
  return true;
}

bool
Controller::load_scene(const std::string& scene_file_name)
{
  if (scene_file_name == "")
  {
    ERROR("No scene setup file given!");
    return false;
  }

  this->stop_processing();
  // TODO: get state.

  // remove all existing sources (if any)
  this->delete_all_sources();

  std::string file_extension = get_file_extension(scene_file_name);

  if (file_extension == "")
  {
    ERROR("File name '" << scene_file_name << "' does not have an extension.");
    return false;
  }
  else if (file_extension == "asd")
  {
    XMLParser xp; // load XML parser
    XMLParser::doc_t scene_file = xp.load_file(scene_file_name);
    if (is_null(scene_file))
    {
      ERROR("Unable to load scene setup file '" << scene_file_name << "'!");
      return false;
    }

    if (_schema_file_name == "")
    {
      ERROR("No schema file specified!");
      // TODO: return true and continue anyway?
      return false;
    }
    else if (scene_file->validate(_schema_file_name))
    {
      VERBOSE("Valid scene setup (" << scene_file_name << ").");
    }
    else
    {
      ERROR("Error validating '" << scene_file_name << "' with schema '"
      << _schema_file_name << "'!");
      return false;
    }

    XMLParser::xpath_t xpath_result;

    // GET MASTER VOLUME
    float master_volume = 0.0f; // dB

    xpath_result = scene_file->eval_xpath("//scene_setup/volume");

    if ( !is_null(xpath_result) && 
         !S2A(get_content(xpath_result->node()), master_volume) )
    {
        WARNING("Invalid master volume specified in scene!");
        master_volume = 0.0f;
    }
    
    VERBOSE("Setting master volume to " << master_volume << " dB.");
    this->set_master_volume(mathtools::dB2linear(master_volume));


    // GET AMPLITUDE REFERENCE DISTANCE

    float ref_dist = _stand_ampl_ref_dist;

    xpath_result
      = scene_file->eval_xpath("//scene_setup/amplitude_reference_distance");
    if (!is_null(xpath_result))
    {
      if (!S2A(get_content(xpath_result->node()), ref_dist))
      {
        WARNING("Invalid amplitude reference distance!");
      }
    }

    // always use default value when nothing is specified
    VERBOSE("Setting amplitude reference distance to "
        << ref_dist << " meters.");
    _scene.set_amplitude_reference_distance(ref_dist);

    // LOAD REFERENCE

    std::auto_ptr<PositionPlusBool> pos_ptr;
    std::auto_ptr<Orientation>      dir_ptr;

    xpath_result = scene_file->eval_xpath("//scene_setup/reference");
    if (!is_null(xpath_result))
    {
      // there should be only one result:
      if (xpath_result->size() != 1)
      {
        ERROR("More than one reference found in scene setup! Aborting.");
        return false;
      }
      pos_ptr = get_position   (xpath_result->node());
      dir_ptr = get_orientation(xpath_result->node());
    }
    else
    {
      VERBOSE("No reference point given in XML file. "
          "Using standard (= origin).");
    }
    if (is_null(pos_ptr)) pos_ptr.reset(new PositionPlusBool());
    if (is_null(dir_ptr)) dir_ptr.reset(new Orientation(90));

    this->set_reference_position(*pos_ptr);
    this->set_reference_orientation(*dir_ptr);
    this->set_reference_offset_position(Position());
    this->set_reference_offset_orientation(Orientation());

    // LOAD SOURCES

    xpath_result = scene_file->eval_xpath("//scene_setup/source");
    if (!is_null(xpath_result))
    {
      for (Node node; (node = xpath_result->node()); ++(*xpath_result))
      {
        std::string name  = node.get_attribute("name");
        std::string id    = node.get_attribute("id");
        std::string properties_file = node.get_attribute("properties_file");

        properties_file = make_path_relative_to_current_dir(properties_file
            , scene_file_name);

        pos_ptr.reset(); dir_ptr.reset();
        pos_ptr = get_position   (node);
        dir_ptr = get_orientation(node);

        // just for the error message:
        std::string id_str;   if (id != "") id_str = " id: \"" + id + "\"";
        std::string name_str; if (name != "") name_str = " name: \""
          + name + "\"";

        Source::model_t model
          = get_attribute_of_node(node, "model", Source::unknown);

        if (model == Source::unknown)
        {
          VERBOSE("Source model not defined!" << id_str << name_str
              << " Using default (= point source).");
          model = Source::point;
        }

        if ((model == Source::point) && is_null(dir_ptr))
        {
          // orientation is optional for point sources, required for plane waves
          dir_ptr.reset(new Orientation);
        }

        if (is_null(pos_ptr) || is_null(dir_ptr))
        {
          ERROR("Both position and orientation have to be specified for source"
              << id_str << name_str << "! Not loaded");
          continue; // next source
        }

        std::string file_or_port_name;
        unsigned int channel = get_file_or_port_name(node, file_or_port_name);
        //if (file_or_port_name == "")
        //{
        //  ERROR("Either a file name or a port name/number have to be specified!"
        //      << id_str << name_str << "! Not loaded");
        //  continue; // next source
        //}
        if (channel == 0) // port specified
        {
          if (file_or_port_name != "")
          {
            file_or_port_name = _input_port_prefix + file_or_port_name;
          }
        }
        else // channel != 0 --> soundfile
        {
          file_or_port_name = make_path_relative_to_current_dir(
              file_or_port_name, scene_file_name);
        }

        float gain_dB = get_attribute_of_node(node, "volume", 0.0f);
        bool muted = get_attribute_of_node(node, "mute", false);
        // bool doppler = get_attribute_of_node(node, "doppler_effect", false);

        this->new_source(name, model, file_or_port_name, channel
            , *pos_ptr, pos_ptr->fixed, *dir_ptr, false
            , mathtools::dB2linear(gain_dB), muted, properties_file);
      }
    }
    else
    {
      WARNING("No sources found in \"" << scene_file_name << "\"!");
    }
  }
  else // file_extension != "asd" -> try to open file as audio file
  {
    WARNING("Trying to open specified file as audio file.");
    if (!_create_spontaneous_scene(scene_file_name))
    {
      ERROR("\"" << scene_file_name << "\" could not be loaded as audio file!");
      return false;
    }
  }

  this->transport_locate(0); // go to beginning of audio file
  // TODO: only start processing if it was on before
  this->start_processing();

  return true;
}

bool
Controller::_create_spontaneous_scene(const std::string& audio_file_name)
{

  const int no_of_audio_channels
    = audiofileiotools::get_no_of_channels(audio_file_name);

  if (no_of_audio_channels == 0)
  {
    WARNING("No audio channels found in file \"" << audio_file_name << "\"!");
    return false;
  }

  WARNING("Creating spontaneous scene from the audio file \""
      << audio_file_name << "\".");

  if (get_renderer_type() == Renderer::BRS)
  {
    WARNING("I don't have information on the BRIRs. I'll use default HRIRs. "
            "Everything will sound in front.");
  }

  // extract pure file name
  const std::string source_name
    = audio_file_name.substr(audio_file_name.rfind('/') + 1);

  // set master volume
  this->set_master_volume(mathtools::dB2linear(-6.0f));
  this->set_amplitude_reference_distance(_stand_ampl_ref_dist);
  // set reference
  this->set_reference_position(Position());
  this->set_reference_orientation(Orientation(90.0f));
  this->set_reference_offset_position(Position());
  this->set_reference_offset_orientation(Orientation());

  const float default_source_distance = 2.5f; // for mono and stereo files

  switch (no_of_audio_channels)
  {
    case 1: // mono file
      // create source
      this->new_source(source_name, Source::point, audio_file_name, 1
          , Position(0.0f, default_source_distance), false, Orientation()
          , false, mathtools::dB2linear(0.0f), false, "");

      VERBOSE("Creating point source at x = " << A2S(0.0f) << " mtrs, y = "
          << A2S(default_source_distance) << " mtrs.");

      break;

    case 2: // stereo file
      {
#undef PI
        const float PI = 3.14159265358979323846;

        const float pos_x = default_source_distance * cos(PI/3.0f);
        const float pos_y = default_source_distance * sin(PI/3.0f);

        // create source
        this->new_source(source_name + " left", Source::plane, audio_file_name
            , 1, Position(-pos_x, pos_y), false, Orientation(-60), false
            , mathtools::dB2linear(0.0f), false, "");

        VERBOSE("Creating point source at x = " << A2S(-pos_x)
            << " mtrs, y = " << A2S(pos_y) << " mtrs.");

        // create source
        this->new_source(source_name + " right", Source::plane, audio_file_name
            , 2u, Position(pos_x, pos_y), false, Orientation(-120), false
            , mathtools::dB2linear(0.0f), false, "");

        VERBOSE("Creating point source at x = " << A2S(pos_x) << " mtrs, y = "
            << A2S(pos_y) << " mtrs.");

        break;
      }

    default:
      // init random position generator
      srand((unsigned int)time(0));

      // create one source for each audio channel
      for (int n = 0; n < no_of_audio_channels; n++)
      {
        // random positions between -n mrts and n mtrs
        const float pos_x = (static_cast<float>(rand()%(10*(n+1)))
            - (5.0f*(n+1))) / 10.0f;

        const float pos_y = 2.0f + (static_cast<float>(rand()%(10*(n+1)))
            - (5.0f*(n+1))) / 10.0f;

        VERBOSE("Creating point source at x = " << A2S(pos_x) << " mtrs, y = "
            << A2S(pos_y) << " mtrs.");

        // create sources
        this->new_source(source_name + " " + A2S(n+1), Source::point
            , audio_file_name, n+1u, Position(pos_x, pos_y), false
            , Orientation(), false, mathtools::dB2linear(0.0f), false, "");
      } // for each audio channel
  } // switch
  return true;
}

#ifdef ENABLE_GUI
int
Controller::start_gui(int &argc, char *argv[]
        , const std::string& path_to_gui_images
        , const std::string& path_to_scene_menu)
{
  // Check whether the system supports OpenGL and set default OpenGL format
  // which will be applied to QGLWidget.
  // Should be equivalent to glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_MULTISAMPLE);
  QGLFormat gl_format = QGLFormat::defaultFormat();
  gl_format.setSampleBuffers(true);
  gl_format.setRgba(true);
  gl_format.setDoubleBuffer(true);
  gl_format.setDepth(true);
  QGLFormat::setDefaultFormat(gl_format);

  _gui.reset(new QGUI(*this, _scene, argc, argv, 
    path_to_gui_images, path_to_scene_menu));

  // check if anti-aliasing is possible
  if (!_gui->format().sampleBuffers())
  {
    WARNING("This system does not provide sample buffer support.\n"
        "I can not enable anti-aliasing for OpenGl stuff.");
  }

  return _gui->run();
}
#endif // ENABLE_GUI

#ifdef ENABLE_IP_INTERFACE
bool
Controller::start_ip_server(int port)
{
  _network_interface.reset(new Server(*this, port));
  _network_interface->start();
  return true;
}
#endif // ENABLE_IP_INTERFACE

#ifdef ENABLE_MCP
bool
Controller::start_mcp_connector()
{
  mcp_connector = NULL;
  mcp_connector = new MCPConnectorStatic(this);

  if (!mcp_connector) return false;

  mcp_connector->Run();
  _subscribe(mcp_connector->GetMsgMapper());
  return true;
}
#endif

#if !defined(ENABLE_NEWRENDERER)
bool
Controller::load_wfs_renderer(const std::string& wfs_prefilter_file_name
    , WFSRenderer::delay_size_t delayline_size
    , WFSRenderer::delay_size_t initial_delay)
{
  std::cout << "Loading WFS renderer.\n";
  _renderer = WFSRenderer::create(_scene, *this, wfs_prefilter_file_name
      , delayline_size, initial_delay);
  return !is_null(_renderer);
}

bool
Controller::load_binaural_renderer(const unsigned int hrir_size
    , const std::string& path_to_hrirs)
{
  std::cout << "Loading binaural renderer.\n";
  _renderer = BinauralRenderer::create(_scene, *this, hrir_size, path_to_hrirs);
  return !is_null(_renderer);
}

bool
Controller::load_brs_renderer(const unsigned int hrir_size)
{
  std::cout << "Loading BRS renderer.\n";
  _renderer = BRSRenderer::create(_scene, *this, hrir_size);
  return !is_null(_renderer);
}

bool
Controller::load_bpb_renderer()
{
  std::cout << "Loading BPB renderer.\n";
  _renderer = BPBRenderer::create(_scene, *this);
  return !is_null(_renderer);
}

bool
Controller::load_vbap_renderer()
{
  std::cout << "Loading VBAP renderer.\n";
  _renderer = VBAPRenderer::create(_scene, *this);
  return !is_null(_renderer);
}

bool
Controller::load_apa_renderer(unsigned int order, bool in_phase_rendering)
{
  std::cout << "Loading Ambisonics amplitude panning renderer.\n";
  _renderer = AAPRenderer::create(_scene, *this, order, in_phase_rendering);
  return !is_null(_renderer);
}

bool
Controller::load_generic_renderer()
{
  std::cout << "Loading generic renderer.\n";
  _renderer = GenericRenderer::create(_scene, *this);
  return !is_null(_renderer);
}
#endif // !ENABLE_NEWRENDERER

// TODO: this is temporary! just for the transition phase until all renderers
// are ported to the new renderer base class
namespace // anonymous
{
  std::string
  get_new_renderer_string(const Renderer::renderer_t& renderer)
  {
    switch (renderer)
    {
      case Renderer::VBAP:     return "vbap";
      case Renderer::WFS:      return "wfs";
      case Renderer::binaural: return "binaural";

      default:
        WARNING("Renderer type not supported! Using 'dumb' renderer.");
        return "dumb";
    }
  }

  Renderer::renderer_t
  get_old_renderer_type(const std::string& renderer)
  {
    if (renderer == "binaural")  return Renderer::binaural;
    else if (renderer == "vbap") return Renderer::VBAP;
    else if (renderer == "wfs")  return Renderer::WFS;

    ERROR("Renderer type '" << renderer << "' not supported!");
    return Renderer::WFS; // selected for no reason
  }
}

#ifdef ENABLE_NEWRENDERER
bool
Controller::load_newrenderer(const ssr::conf_struct &conf)
{
  VERBOSE("Loading new renderer.");
  parameter_map render_params;

  // TODO: remove once all renderers are ported to the new base class
  render_params["type"] = get_new_renderer_string(conf.renderer);

  render_params["delayline_size"] = A2S(conf.wfs_delayline_size);
  render_params["initial_delay"] = A2S(conf.wfs_initial_delay);
  render_params["prefilter_file_name"] = conf.wfs_prefilter_file_name;
  render_params["jackname"] = "SSR_" + render_params["type"];
  render_params["threads"] = A2S(conf.number_of_threads);
  render_params["hrir_size"] = A2S(conf.hrir_size);
  render_params["hrir_filename"] = conf.hrir_file_name;

  _renderer = std::auto_ptr<NewRenderer>(NewRenderer::create(_scene, *this
        , render_params));
  RenderSubscriber *subscriber = new RenderSubscriber(*_renderer);
  _subscribe(subscriber);

  return !is_null(_renderer);
}
#endif

void
Controller::start_tracker(const std::string& type, const std::string& ports)
{
  if (type == "")
  {
    return;
  }
  else if (type == "intersense")
  {
#if defined(ENABLE_INTERSENSE)
    _tracker = TrackerInterSense::create(*this, ports);
#else
    ERROR("The SSR was compiled without InterSense tracker support!");
    (void)ports;  // avoid "unused parameter" warning
    return;
#endif
  }
  else if (type == "polhemus")
  {
#if defined(ENABLE_POLHEMUS)
    _tracker = TrackerPolhemus::create(*this, ports);
#else
    ERROR("The SSR was compiled without Polhemus tracker support!");
    (void)ports;  // avoid "unused parameter" warning
    return;
#endif
  }
  else if (type == "razor")
  {
#if defined(ENABLE_RAZOR)
    _tracker = TrackerRazor::create(*this, ports);
#else
    ERROR("The SSR was compiled without Razor AHRS tracker support!");
    (void)ports;  // avoid "unused parameter" warning
    return;
#endif
  }
  else
  {
    ERROR("Unknown tracker type \"" << type << "\"!");
    return;
  }

  if (is_null(_tracker))
  {
    WARNING("Cannot find tracker. "
            "Make sure that you have the appropriate access rights "
            "to read from the port. I continue without tracker.");
  }
}

/// This is temporary!!!!
void
Controller::calibrate_client()
{
#if defined(ENABLE_INTERSENSE) || defined(ENABLE_POLHEMUS) || defined(ENABLE_RAZOR)
  if (!is_null(_tracker))
  {
    _tracker->calibrate();
  }
  else
  {
    WARNING("No tracker there to calibrate.");
  }
#endif
}

/**
 * @param output_port_prefix name of the desired output port but without the
 * final number.
 * @param skip if true, start with _first_loudspeaker_channel and skip channels
 * in _skipped_loudspeaker_channels, if false start with channel 1 and don't
 * skip any channels.
 * @attention The first output port number can be set with the member variable
 * _first_loudspeaker_channel.
 **/
bool
Controller::connect_renderer(const std::string& output_port_prefix, bool skip)
{
  if (is_null(_renderer))
  {
    ERROR("No renderer running!");
    return false;
  }
#ifdef ENABLE_NEWRENDERER
  NewRenderer::port_list_t port_list; // empty list
#else
  Renderer::port_list_t port_list; // empty list
#endif
  _renderer->get_output_port_names(port_list);
  if (port_list.empty())
  {
    WARNING("No output ports in renderer! Not connected to '"
        + output_port_prefix + "x'!");
    return false;
  }
  bool success = true;
  loudspeaker_id_t port_number = skip ? _first_loudspeaker_channel : 1;
  for (Renderer::port_list_t::const_iterator i = port_list.begin();
      i != port_list.end(); ++i, ++port_number)
  {
    if (skip)
    {
      // skip output ports from skip-list
      while (_skipped_loudspeaker_channels.find(port_number)
          != _skipped_loudspeaker_channels.end())
      {
        ++port_number;
      }
    }
    assert(!is_null(_renderer)); // existence of renderer was tested before
    if (!_renderer->connect_ports(*i, output_port_prefix + A2S(port_number)))
    {
      success = false;
    }
  }
  return success;
}

#ifdef ENABLE_ECASOUND
/**
 * The recorder is started as soon a JACK transport is running.
 * @param audio_file_name quite obviously, the the file which will be recorded
 * to. It will have the given @a sample_format, the same number of channels as
 * the renderer has output channels and it will have the same sample rate as the
 * JACK audio server.
 * @param sample_format for more information see the ecasound(1) manpage and
 * especially the documentation of the -f option.
 **/
bool
Controller::load_audio_recorder(const std::string audio_file_name
    , const std::string& sample_format)
{
  if (is_null(_renderer))
  {
    ERROR("No renderer running!");
    return false;
  }
#ifdef ENABLE_NEWRENDERER
  unsigned int channels = _renderer->get_output_list().size();
#else
  unsigned int channels = _renderer->get_number_of_output_channels();
#endif
  unsigned int sample_rate = _renderer->get_sample_rate();

  _audio_recorder = AudioRecorder::create(audio_file_name
      , sample_format + "," + A2S(channels) + "," + A2S(sample_rate));
  if (is_null(_audio_recorder))
  {
    ERROR("AudioRecorder (\"" + audio_file_name + "\") couldn't be created!");
    return false;
  }

  // This is already called in ctor:
  //_audio_recorder->enable();

  bool skip = false; // don't skip any output channels 
  if (!this->connect_renderer(_audio_recorder->client_name + ":"
            + _audio_recorder->input_prefix + "_", skip))
  {
    ERROR("Couldn't connect renderer to audio recorder!");
    return false;
  }
  return true;
}
#endif

/** start audio processing.
 * This sets the Scene's processing state to "processing". The
 * process callback function has to check for this variable and act accordingly.
 **/
void
Controller::start_processing()
{
  if (_scene.get_processing_state() == ssr::ready)
  {
    this->set_processing_state(ssr::processing);
  }
  else if (_scene.get_processing_state() == ssr::processing)
  {
    WARNING("Renderer is already processing.");
  }
  else ERROR("Couldn't start processing! Renderer is not ready.");
}

/** Stop audio processing.
 * This sets the Scene's processing state to "ready". The
 * process callback function has to check for this variable and act
 * accordingly.
 **/
void
Controller::stop_processing()
{
  if (_scene.get_processing_state() == ssr::processing)
  {
    this->set_processing_state(ssr::ready);
  }
  else if (_scene.get_processing_state() == ssr::ready)
  {
    WARNING("Renderer was already stopped.");
  }
  else ERROR("Couldn't stop processing!");
}

/** _.
 * @param state processing state.
 * @warning States are not checked for validity. To start and stop rendering,
 * use preferably start_processing() and stop_processing(). This function should
 * only be used by the respective renderer to set the state "ready" and by
 * anyone who wants to set the state "exiting".
 **/
void
Controller::set_processing_state(ssr::processing_state_t state)
{
  _publish(&Subscriber::set_processing_state, state);
}

bool
Controller::set_freewheeling()
{
  // TODO: save state and restore it on exit
  if (!is_null(_renderer))
  {
    if (_renderer->set_freewheel(1)) return true;
    else ERROR("Freewheeling doesn't work!");
  }
  return false;
}

// non-const because audioplayer could be started
void
Controller::transport_start()
{
  if (!is_null(_renderer)) _renderer->transport_start();
}

// non-const because audioplayer could be started
void
Controller::transport_stop()
{
  if (!is_null(_renderer)) _renderer->transport_stop();
}

/** Skips the scene to a specified instant of time
 * @ param frame instant of time in sec to locate 
 */
bool
Controller::transport_locate(float time)
{
  if (!is_null(_renderer))
  {
    // convert time to samples (cut decimal part)
    return _renderer->transport_locate(
        static_cast<jack_nframes_t>(time * _renderer->get_sample_rate()));
  }
  return false;
}

void
Controller::set_transport_state(const std::pair<bool, jack_nframes_t>& state)
{
  _publish(&Subscriber::set_transport_state, state);
}

/** Create a new source.
 * @param name Source name
 * @param model Source model
 * @param file_or_port_name File or port name
 * @param channel Channel of soundfile. If 0, a JACK portname is expected.
 * @param position initial position of the source.
 * @param orientation initial orientation of the source.
 * @param gain gain (=volume) of the source.
 * @return ID of the created source. If 0, no source was created.
 **/
void
Controller::new_source(const std::string& name, const Source::model_t model
    , const std::string& file_or_port_name, unsigned int channel
    , const Position& position, const bool pos_fixed
    , const Orientation& orientation, const bool or_fixed, const float gain
    , const bool muted, const std::string& properties_file)
{
  (void) or_fixed;

  id_t id = _scene.get_new_source_id();
  // if loading the source fails, this ID will not be used anymore
  std::string port_name;
  std::string file_name = "";
  long int file_length = 0;

  if (channel > 0) // we're dealing with a soundfile
  {
#ifdef ENABLE_ECASOUND
    // if not already running, start AudioPlayer
    if (is_null(_audio_player))
    {
      _audio_player = AudioPlayer::ptr_t(new AudioPlayer);
    }
    // the thing with _loop is a temporary hack, should be removed some time:
    port_name = _audio_player->get_port_name(file_or_port_name, channel, _loop);
    //port_name   = _audio_player->get_port_name  (file_or_port_name, channel);
    file_length = _audio_player->get_file_length(file_or_port_name);
    file_name   = file_or_port_name;
#else
    ERROR("Couldn't open audio file \"" << file_or_port_name
        << "\"! Ecasound was disabled at compile time.");
    return;
#endif
  }
  else // no audio file
  {
    port_name = file_or_port_name;
  }

  if (port_name == "" && get_renderer_type() != Renderer::BPB)
  {
    VERBOSE("No audio file or port specified for source '" << name << "'.");
  }

  if (is_null(_renderer))
  {
    WARNING("No Renderer running.");
  }
  else // Renderer is running
  {
#ifndef ENABLE_NEWRENDERER 
////////////////////////////////////////////////////////////////////////////////
///////////// DIRTY HACK FOR BRS AND BPB RENDERERS /////////////////////////////
////////////////////////////////////////////////////////////////////////////////
    if (get_renderer_type() == Renderer::BRS
        || get_renderer_type() == Renderer::generic
        || get_renderer_type() == Renderer::BPB)
    {
      if (!_renderer->new_source(id, port_name, properties_file))
      {
        ERROR("Source " << id << " could not be created in Renderer.");
        return;
      }
    }
    else
    {
      if (!_renderer->new_source(id, port_name))
      {
        ERROR("Source " << id << " could not be created in Renderer.");
        return;
      }
    }
////////////////////////////////////////////////////////////////////////////////
///////////// END OF DIRTY HACK FOR BRS AND BPB RENDERERS //////////////////////
////////////////////////////////////////////////////////////////////////////////
#else
    parameter_map input_params;
    input_params["connect_to"] = port_name;
    input_params["port_name"]  = "in_" + A2S(id);
    input_params["id"] = A2S(id);
    _renderer->addSourceInput(input_params);
#endif
  }

  _publish(&Subscriber::new_source, id);
  // mute while transmitting data
  _publish(&Subscriber::set_source_mute, id, true);
  _publish(&Subscriber::set_source_gain, id, gain);
  _publish(&Subscriber::set_source_position, id, position);
  _publish(&Subscriber::set_source_position_fixed, id, pos_fixed);
  _publish(&Subscriber::set_source_orientation, id, orientation);
  // _publish(&Subscriber::set_source_orientation_fix, id, or_fix);
  _publish(&Subscriber::set_source_name, id, name);
  _publish(&Subscriber::set_source_model, id, model);
  _publish(&Subscriber::set_source_port_name, id, port_name);
  if (file_name != "")
  {
    _publish(&Subscriber::set_source_file_name, id, file_name);
    _publish(&Subscriber::set_source_file_channel, id, channel);
  }
  _publish(&Subscriber::set_source_file_length, id, file_length);
  _publish(&Subscriber::set_source_properties_file, id, properties_file);
  // finally, unmute if requested
  _publish(&Subscriber::set_source_mute, id, muted);
}

void
Controller::delete_all_sources()
{
  _publish(&Subscriber::delete_all_sources);
#ifndef ENABLE_NEWRENDERER
  if (!is_null(_renderer)) _renderer->delete_all_sources();
#endif
#ifdef ENABLE_ECASOUND
  ptrtools::destroy(_audio_player); // shut down audio player
#endif
}

void
Controller::delete_source(id_t id)
{
  _publish(&Subscriber::delete_source, id);
  // TODO: stop AudioPlayer if not needed anymore?
}

void
Controller::set_source_position(const id_t id, const Position& position)
{
  // TODO: check if the client who sent the request is actually allowed to
  // change the position. (same TODO as above)

  // TODO: check if position is inside of room boundaries.
  // if not: change position (e.g. single dimensions) to an allowed position

  // check if source may be moved
  if (!_scene.get_source_position_fixed(id))
  {
    _publish(&Subscriber::set_source_position, id, position);
  }
  else
  {
    WARNING("Source \'" << _scene.get_source_name(id)
        << "\' can not be moved.");
  }
}

void
Controller::set_source_orientation(const id_t id
    , const Orientation& orientation)
{
  // TODO: validate orientation?

  _publish(&Subscriber::set_source_orientation, id, orientation);
}

void
Controller::set_source_gain(const id_t id, const float gain)
{
  _publish(&Subscriber::set_source_gain, id, gain);
}

void
Controller::set_source_signal_level(const id_t id, const float level)
{
  _publish(&Subscriber::set_source_signal_level, id, level);
}

void
Controller::set_source_mute(const id_t id, const bool mute)
{
  _publish(&Subscriber::set_source_mute, id, mute);
}

void
Controller::set_source_name(const id_t id, const std::string& name)
{
  _publish(&Subscriber::set_source_name, id, name);
}

void
Controller::set_source_properties_file(const id_t id, const std::string& name)
{
  _publish(&Subscriber::set_source_properties_file, id, name);
}

void
Controller::set_source_model(const id_t id, const Source::model_t model)
{
  _publish(&Subscriber::set_source_model, id, model);
}

void
Controller::set_source_port_name(const id_t id, const std::string& port_name)
{
  _publish(&Subscriber::set_source_port_name, id, port_name);
}

void
Controller::set_source_file_name(const id_t id, const std::string& file_name)
{
  _publish(&Subscriber::set_source_file_name, id, file_name);
}

void
Controller::set_source_file_channel(const id_t id, const unsigned int& channel)
{
  _publish(&Subscriber::set_source_file_channel, id, channel);
}

void
Controller::set_source_position_fixed(const id_t id, const bool fixed)
{
  _publish(&Subscriber::set_source_position_fixed, id, fixed);
}

void
Controller::set_reference_position(const Position& position)
{
  _publish(&Subscriber::set_reference_position, position);

  // TODO: update orientations of plane waves
}

void
Controller::set_reference_orientation(const Orientation& orientation)
{
  _publish(&Subscriber::set_reference_orientation, orientation);
}

void
Controller::set_reference_offset_position(const Position& position)
{
  _publish(&Subscriber::set_reference_offset_position, position);
}

void
Controller::set_reference_offset_orientation(const Orientation& orientation)
{
  _publish(&Subscriber::set_reference_offset_orientation, orientation);
}

// linear volume!
void
Controller::set_master_volume(const float volume)
{
  // TODO: validate volume?
  _publish(&Subscriber::set_master_volume, volume);
}

void
Controller::set_master_volume_correction(float corr_factor)
{
  // TODO: validate volume?
  _publish(&Subscriber::set_master_volume_correction, corr_factor);
}

void
Controller::set_amplitude_reference_distance(const float dist)
{
  if (dist > 1.0f)
  {
    _publish(&Subscriber::set_amplitude_reference_distance, dist);
  }
  else
  {
    ERROR("Amplitude reference distance can not be smaller than 1.");
  }
}

// linear scale
void
Controller::set_master_signal_level(float level)
{
  _publish(&Subscriber::set_master_signal_level, level);
}

void
Controller::set_cpu_load(const float load)
{
  _publish(&Subscriber::set_cpu_load, load);
}

void
Controller::publish_sample_rate(const unsigned int sample_rate)
{
  _publish(&Subscriber::set_sample_rate, sample_rate);
}

Renderer::renderer_t
Controller::get_renderer_type() const
{
#ifdef ENABLE_NEWRENDERER
  return get_old_renderer_type(_renderer->params["type"]);
#else
  return _renderer->get_renderer_type(); 
#endif
}

void
Controller::clear_all_active_sources()
{
  _publish(&Subscriber::clear_all_active_sources);
}

void
Controller::set_active_sources(
    Loudspeaker::container_t::size_type loudspeaker_id
    , const std::set<id_t>& active_source_ids)
{
  _publish(&Subscriber::set_active_sources, loudspeaker_id, active_source_ids);
}

// just a temporary hack, should be implemented properly later!
/*
void Controller::publish_everything()
{
  // TODO: implement properly!
  DirectionalPoint reference = _scene.get_reference();
  this->set_reference_position(reference.position);
  this->set_reference_orientation(reference.orientation);

  Loudspeaker::container_t loudspeakers;
  // second argument disables output of absolute position
  _scene.get_loudspeakers(loudspeakers, false);
  _publish(&Subscriber::set_loudspeakers, loudspeakers);
  
  SourceCopy::container_t sources;
  _scene.get_sources(sources);
  for (SourceCopy::container_t::const_iterator i = sources.begin();
      i != sources.end(); ++i)
  {
    _publish(&Subscriber::new_source, i->id);
    // mute while transmitting data
    _publish(&Subscriber::set_source_mute, i->id, true);
    _publish(&Subscriber::set_source_gain, i->id, i->gain);
    _publish(&Subscriber::set_source_position, i->id, i->position);
    _publish(&Subscriber::set_source_orientation, i->id, i->orientation);
    _publish(&Subscriber::set_source_name, i->id, i->name);

    // TODO: there are some functions missing here!

    _publish(&Subscriber::set_source_model, i->id, i->model);
    _publish(&Subscriber::set_source_port_name, i->id, i->port_name);
    _publish(&Subscriber::set_source_file_length, i->id, i->file_length);
    // finally, set original mute state
    _publish(&Subscriber::set_source_mute, i->id, i->mute);
  }
}
*/

std::string
Controller::get_scene_as_XML() const
{
  XMLParser xp; // load XML parser
  Node node = xp.new_node("update");

  // add master volume
  _add_master_volume(node);
  // quick hack: add transport state (play/stop)
  _add_transport_state(node);
  // TODO: add other scene information?
  _add_reference(node);
  _add_loudspeakers(node);
  _add_sources(node);

  // TODO: memory of node is never freed!

  return node.to_string();
}

bool
Controller::save_scene_as_XML(const std::string& filename) const
{
  // ATTENTION: this is an ugly work-around/quick-hack!
  // TODO: the following should be included into the XMLParser wrapper!

  XMLParser xp; // load XML parser
  Node root = xp.new_node("asdf");

  //std::cout << root.to_string() << std::endl;

  Node node = root.new_child("scene_setup");

  //std::cout << root.to_string() << std::endl;

  // add master volume
  _add_master_volume(node);

  // TODO: add other scene information?
  _add_reference(node);

  _add_sources(node, filename); // ugly quick hack!

  // TODO: memory of node is never freed!

  xmlDocPtr doc = xmlNewDoc(NULL);
  //xmlDocPtr doc = xmlNewDoc(BAD_CAST "1.0");
  if (!doc) return false;

  //std::cout << root.to_string() << std::endl;

  xmlDocSetRootElement(doc, root.get());

  //xmlNewNs(my_node, BAD_CAST "http://www.w3.org/1999/xhtml", NULL);

  //xmlSaveCtxtPtr ctxt = xmlSaveToFilename(filename.c_str(), NULL, 0);
  xmlSaveCtxtPtr ctxt = xmlSaveToFilename(filename.c_str(), NULL
      , XML_SAVE_FORMAT);
  if (!ctxt) return false;

  xmlSaveDoc(ctxt, doc);
  int ret = xmlSaveClose(ctxt);

  if (ret < 1) return false;
  else return true;

  // Memory for "doc" is never freed!
}

void
Controller::_add_master_volume(Node& node) const
{
  float volume = _scene.get_master_volume();
  node.new_child("volume", A2S(mathtools::linear2dB(volume)));
}

void
Controller::_add_transport_state(Node& node) const
{
  node.new_child("transport"
      , _renderer->get_transport_state().first ? "start" : "stop");
}

void
Controller::_add_reference(Node& node) const
{
  DirectionalPoint reference = _scene.get_reference();
  Node reference_node = node.new_child("reference");
  _add_position(reference_node, reference.position);
  _add_orientation(reference_node, reference.orientation);
}

void
Controller::_add_position(Node& node
    , const Position& position, const bool fixed) const
{
  Node position_node = node.new_child("position");
  position_node.new_attribute("x", A2S(position.x));
  position_node.new_attribute("y", A2S(position.y));
  if (fixed)
  {
    position_node.new_attribute("fixed", A2S(fixed));
  }
}

void
Controller::_add_orientation(Node& node, const Orientation& orientation) const
{
  Node orientation_node = node.new_child("orientation");
  orientation_node.new_attribute("azimuth", A2S(orientation.azimuth));
}

void
Controller::_add_loudspeakers(Node& node) const
{
  Loudspeaker::container_t loudspeakers;
  _scene.get_loudspeakers(loudspeakers, false); // get relative positions
  for (Loudspeaker::container_t::const_iterator i = loudspeakers.begin();
      i != loudspeakers.end(); ++i)
  {
    Node loudspeaker_node = node.new_child("loudspeaker");
    loudspeaker_node.new_attribute("model", A2S(i->model));
    _add_position(loudspeaker_node, i->position);
    _add_orientation(loudspeaker_node, i->orientation);
  }
}

void
Controller::_add_sources(Node& node, const std::string& scene_file_name) const
{
  SourceCopy::container_t sources;
  _scene.get_sources(sources);
  for (SourceCopy::container_t::const_iterator i = sources.begin();
      i != sources.end(); ++i)
  {
    Node source_node = node.new_child("source");
    if (scene_file_name != "")
    {
      // ignore "id" -> ugly quick hack!
    }
    else
    {
      source_node.new_attribute("id", A2S(i->id));
    }
    source_node.new_attribute("name", A2S(i->name));
    source_node.new_attribute("model", A2S(i->model));
    if (i->audio_file_name != "") // ugly work-around
    {
      _add_audio_file_name(source_node
          , make_path_relative_to_file(i->audio_file_name, scene_file_name)
          , i->audio_file_channel);
    }

    if (scene_file_name != "") // ugly quick hack
    {
      // don't add port name!
    }
    else
    {
      _add_port_name(source_node, i->port_name);
    }

    _add_position(source_node, i->position, i->fixed_position);
    _add_orientation(source_node, i->orientation);
    if (scene_file_name != "") // ugly quick hack
    {
      // don't add port name!
    }
    else
    {
      source_node.new_attribute("length", A2S(i->file_length));
    }
    source_node.new_attribute("mute", A2S(i->mute));
    // save volume in dB!
    source_node.new_attribute("volume", A2S(mathtools::linear2dB(i->gain)));
    // TODO: information about mirror sources

    if (i->properties_file != "")
    {
      source_node.new_attribute("properties_file", i->properties_file);
    }

    // TODO: save doppler effect setting (i->doppler_effect)
  }
}

void
Controller::_add_audio_file_name(Node& node, const std::string& name
    , unsigned int channel) const
{
  Node file_node = node.new_child("file", name);
  if (channel != 1)
  {
    file_node.new_attribute("channel", A2S(channel));
  }
}

void
Controller::_add_port_name(Node& node, const std::string& name) const
{
  Node port_node = node.new_child("port", name);
}

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
