/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * %Controller (definition).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef CONTROLLER_H
#define CONTROLLER_H

#ifdef HAVE_CONFIG_H
#include <config.h> // for ENABLE_*, HAVE_*, WITH_*
#endif

#include "ssr_global.h"
#include "publisher.h"
#include "masterscene.h"
#ifdef ENABLE_ECASOUND
#include "audioplayer.h"
#include "audiorecorder.h"
#endif
#include "sceneplayer.h"
#include "renderer.h"
#include "xmlparser.h"
#include "configuration.h"

#ifdef ENABLE_GUI
#include "qgui.h"
#endif

#ifdef ENABLE_IP_INTERFACE
#include "server.h"
#endif

#if !defined(ENABLE_NEWRENDERER)
#include "wfsrenderer.h"
#include "binauralrenderer.h"
#include "brsrenderer.h"
#include "bpbrenderer.h"
#include "vbaprenderer.h"
#include "aaprenderer.h"
#include "genericrenderer.h"
#endif

#include "tracker.h"

#ifdef ENABLE_MCP
// forward declaration:
class MCPConnectorStatic;
#endif

typedef XMLParser::Node Node; ///< a node of the DOM tree

using ssr::id_t;
//class Renderer;

/** %Controller class.
 * Has a list of objects which receive messages on events like position change
 * etc. With this list a kind of Publisher/Subscriber pattern is realized.
 **/
class Controller : public Publisher
{
  public:
    typedef Loudspeaker::container_t::size_type loudspeaker_id_t;
    /// ctor
    explicit Controller(const std::string& schema_file_name
        , const std::string& input_port_prefix, const float ampl_ref_dist);
    ~Controller(); ///< dtor

    bool load_loudspeaker_setup(const std::string& setup_file_name);
#ifdef ENABLE_NEWRENDERER
    bool load_newrenderer(const ssr::conf_struct &conf);
#else
    bool load_wfs_renderer(const std::string& wfs_prefilter_file_name
        , WFSRenderer::delay_size_t delayline_size
        , WFSRenderer::delay_size_t initial_delay);
    bool load_binaural_renderer(const unsigned int hrir_size
        , const std::string& path_to_hrirs);
    bool load_brs_renderer(const unsigned int hrir_size);
    bool load_bpb_renderer();
    bool load_vbap_renderer();
    bool load_apa_renderer(unsigned int order, bool in_phase_rendering);
    bool load_generic_renderer();
#endif // ENABLE_NEWRENDERER
    /// connect renderer outputs to the given JACK ports.
    bool connect_renderer(const std::string& output_port_prefix
        , bool skip = true);
#ifdef ENABLE_ECASOUND
    /// load the audio recorder and set it to "record enable" mode.
    bool load_audio_recorder(const std::string audio_file_name
        , const std::string& sample_format = "16");
#endif
#ifdef ENABLE_GUI
    int start_gui(int &argc, char *argv[]
        , const std::string& path_to_gui_images
        , const std::string& path_to_scene_menu);
#endif
#ifdef ENABLE_IP_INTERFACE
    bool start_ip_server(int port);
#endif

#ifdef ENABLE_MCP
    bool start_mcp_connector();
#endif

    void clear_all_active_sources(); ///< clear all Loudspeaker::active_sources
    /// transmit IDs of active sources for one Loudspeaker
    void set_active_sources(loudspeaker_id_t loudspeaker_id
        , const std::set<id_t>& active_source_ids);

    virtual bool load_scene(const std::string& scene_file_name);
    virtual bool save_scene_as_XML(const std::string& filename) const;

    virtual void start_processing();
    virtual void stop_processing();

    virtual void new_source(const std::string& name, Source::model_t model
        , const std::string& file_or_port_name, unsigned int channel = 0
        , const Position& position = Position(), const bool pos_fix = false
        , const Orientation& orientation = Orientation()
        , const bool or_fix = false
        , const float gain = 1.0f, const bool muted = false
        , const std::string& properties_file = "");

    virtual void delete_source(id_t id);
    /// delete all sources in all subscribers
    virtual void delete_all_sources();

    virtual void set_source_position(id_t id, const Position& position);
    virtual void set_source_orientation(id_t id
        , const Orientation& orientation);
    virtual void set_source_gain(id_t id, float gain);
    virtual void set_source_signal_level(const id_t id
        , const float level);
    virtual void set_source_mute(id_t id, bool mute);
    virtual void set_source_name(id_t id, const std::string& name);
    virtual void set_source_properties_file(id_t id, const std::string& name);
    virtual void set_source_model(id_t id, Source::model_t model);
    virtual void set_source_port_name(id_t id, const std::string& port_name);
    virtual void set_source_file_name(id_t id, const std::string& file_name);
    virtual void set_source_file_channel(id_t id, const unsigned int& channel);
    virtual void set_source_position_fixed(id_t id, const bool fixed);

    virtual void set_reference_position(const Position& position);
    virtual void set_reference_orientation(const Orientation& orientation);

    virtual void set_reference_offset_position(const Position& position);
    virtual void set_reference_offset_orientation(const Orientation& orientation);

    virtual void set_master_volume(float volume);

    virtual void set_master_volume_correction(float corr_factor);

    virtual void set_amplitude_reference_distance(const float dist);

    virtual void set_master_signal_level(float level);

    virtual void set_cpu_load(const float load);

    virtual void publish_sample_rate(const unsigned int sample_rate);

    virtual Renderer::renderer_t get_renderer_type() const;

    virtual void transport_start();
    virtual void transport_stop();
    virtual bool transport_locate(float time);

    virtual void calibrate_client();

    /// update JACK transport state
    void set_transport_state(const std::pair<bool, jack_nframes_t>& state);
    /// switch JACK to freewheeling mode.
    bool set_freewheeling();
    /// send processing state of the renderer to all subscribers.
    void set_processing_state(ssr::processing_state_t state);

    std::string get_scene_as_XML() const;

    void start_tracker(const std::string& type, const std::string& ports = "");

    void subscribe(Subscriber* const subscriber); ///< add a new subscriber
    void unsubscribe(Subscriber* const subscriber); ///< add a new subscriber

    void set_loop_mode(bool loop) { _loop = loop; } ///< temporary solution!

  private:
    MasterScene& _scene; ///< The one and only master scene
    std::auto_ptr<ScenePlayer> _scene_player; ///< scene player
    /// a list of subscribers
    typedef std::vector<Subscriber*> subscriber_list_t;
    /// list of objects that will be notified on all events
    subscriber_list_t _subscribers;
#ifdef ENABLE_GUI
    std::auto_ptr<QGUI> _gui;
#endif

#ifdef ENABLE_NEWRENDERER
    std::auto_ptr<NewRenderer> _renderer;       ///< pointer to renderer
#else
    std::auto_ptr<Renderer> _renderer;       ///< pointer to renderer
#endif
#ifdef ENABLE_ECASOUND
    AudioRecorder::ptr_t    _audio_recorder; ///< pointer to audio recorder
    AudioPlayer::ptr_t      _audio_player;   ///< pointer to audio player
#endif
    std::string _schema_file_name;           ///< XML Schema
    std::string _input_port_prefix;          ///< e.g. alsa_pcm:capture
#ifdef ENABLE_IP_INTERFACE
    std::auto_ptr<Server> _network_interface;
#endif
    std::auto_ptr<Tracker> _tracker;
#ifdef ENABLE_MCP
    MCPConnectorStatic *mcp_connector;
#endif

    loudspeaker_id_t             _first_loudspeaker_channel;
    std::set<loudspeaker_id_t> _skipped_loudspeaker_channels;
    float _stand_ampl_ref_dist;

    /// check if audio player is running and start it if necessary
    bool _audio_player_is_running();

    void _subscribe(Subscriber* const subscriber); ///< add a new subscriber

    // TODO: write _unsubscribe() ?

    /// @name publishing functions
    /// Several overloaded versions for different combinations of arguments.
    /// The first argument is a pointer to a member function of the Subscriber
    /// class.
    //@{
    /// non-const, 0 arguments, return type void
    //
    // FIXME: THREAD
    inline void _publish(void (Subscriber::*f)())
    {
      for (subscriber_list_t::iterator i = _subscribers.begin()
          ; i != _subscribers.end()
          ; ++i)
      {
        ((*i)->*f)();
      }
    }
    /// non-const, 1 argument (by value), return type void
    template<typename A> inline void _publish(void (Subscriber::*f)(A), A a)
    {
      for (subscriber_list_t::iterator i = _subscribers.begin()
          ; i != _subscribers.end()
          ; ++i)
      {
        ((*i)->*f)(a);
      }
    }
    /// non-const, 1 argument (by const reference), return type void
    template<typename A> inline void _publish(void (Subscriber::*f)(const A&)
        , const A& a)
    {
      for (subscriber_list_t::iterator i = _subscribers.begin()
          ; i != _subscribers.end()
          ; ++i)
      {
        ((*i)->*f)(a);
      }
    }
    /// non-const, 2 arguments (1st by value, 2nd by const reference),
    /// any return type (ignored)
    template<typename R, typename A1, typename A2> inline void _publish(
        R (Subscriber::*f)(A1, const A2&), A1 a1, const A2& a2)
    {
      for (subscriber_list_t::iterator i = _subscribers.begin()
          ; i != _subscribers.end()
          ; ++i)
      {
        ((*i)->*f)(a1, a2);
      }
    }
    //@}

    /// helper struct for a source including its source id
    struct SourceCopy : public Source
    {
      typedef std::vector<SourceCopy> container_t; ///< list of SourceCopys
      /// type conversion constructor
      SourceCopy(const std::pair<id_t, Source>& other)
        : Source(other.second) // copy ctor of base class
        , id(other.first)
      {}

      id_t id; ///< unique ID, see Scene::source_map_t
    };

    void _add_master_volume(Node& node) const;
    void _add_transport_state(Node& node) const;
    void _add_reference(Node& node) const;
    void _add_position(Node& node
        , const Position& position, const bool fixed = false) const;
    void _add_orientation(Node& node, const Orientation& orientation) const;
    void _add_loudspeakers(Node& node) const;
    void _add_sources(Node& node, const std::string& filename = "") const;
    void _add_audio_file_name(Node& node, const std::string& name
        , unsigned int channel) const;
    void _add_port_name(Node& node, const std::string& name) const;
    bool _create_spontaneous_scene(const std::string& audio_file_name);

    bool _loop; ///< part of a quick-hack. should be removed some time.
};

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
