/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * Generic renderer class (implementation).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include <cassert>
#include <cmath>
#include "genericrenderer.h"
#include "controller.h"
#include "ssr_global.h"
#include "mathtools.h"
#include "ptrtools.h"
#include "stringtools.h"
#include "audiofileiotools.h"

// for debugging only:
#include <libxml/debugXML.h>

using ptrtools::is_null;
using ptrtools::destroy;
using stringtools::S2A;
using stringtools::A2S;
using namespace ssr;

/** ctor.
 * @param scene see create()
 * @param controller see create()
 **/
GenericRenderer::GenericRenderer(const Scene& scene, Controller& controller,
				 const std::string& client_name) 
  throw (jack_error, gen_error) :
  Renderer(scene, client_name, true), // base class ctor
  _controller(controller),
  _input_prefix("in"),
  _output_prefix("out")
{
  // There shouldn't be any sources in the scene at this point.
  // If there are, they are ignored!

  if (!_make_outputs())
  {
    maptools::purge(_output_map); // delete any leftovers in _output_map
    throw(gen_error("Output objects could not be created!"));
  }

  if (this->get_buffer_size() == 0)
  {
    WARNING("Buffersize == 0!");
  }
  else
  {
    // fill silence buffer with zeroes
    _silence.resize(this->get_buffer_size(), 0);
  }

  // init memory for crossfade slopes
  _fade_in.resize (this->get_buffer_size(), static_cast<sample_t>(0.0f));
  _fade_out.resize(this->get_buffer_size(), static_cast<sample_t>(0.0f));

  // raised cosine fade
  for(unsigned int n = 0u; n < this->get_buffer_size(); n++)
  {
    _fade_in[n]  = static_cast<sample_t>(0.5f
        + 0.5f * cos(static_cast<float>(this->get_buffer_size()-n)
		     / static_cast<float>(this->get_buffer_size()) * mathtools::PI));
    _fade_out[n] = static_cast<sample_t>(0.5f
        + 0.5f * cos(static_cast<float>(n)
		     / static_cast<float>(this->get_buffer_size()) * mathtools::PI));
  } // for

  _controller.publish_sample_rate(
      static_cast<unsigned int>(this->get_sample_rate()));
  _controller.set_processing_state(ready); // TODO: lock with mutex?

  this->activate(); // activate JACK client
}

GenericRenderer::~GenericRenderer()
{
  _controller.set_processing_state(exiting);

  // cond_signal to any waiting threads?

  maptools::purge(_output_map);
  VERBOSE2("GenericRenderer: Output map cleared.");

  maptools::purge(_input_map);
  VERBOSE2("GenericRenderer: Input map cleared.");
}

/** stop rendering.
 * This is only called by JACK. Do not dare to call it yourself!
 * - stops rendering
 * - calls its base class counterpart
 *
 * @return I don't know what
 */
int GenericRenderer::jack_shutdown()
{
  ERROR("Prematurely disconnected from JACK server! "
      "Trying to stop processing ...");
  _controller.set_processing_state(exiting);
  // cond_signal to any waiting threads?
  maptools::purge(_output_map); WARNING("Output map cleared.");
  maptools::purge(_input_map);  WARNING("Input map cleared.");

  return JackClient::jack_shutdown(); // call original base class function
}

GenericRenderer::ptr_t GenericRenderer::create(const Scene& scene,
    Controller& controller, const std::string& client_name)
{
  ptr_t temp; // temp = NULL
  try
  {
    temp.reset(new GenericRenderer(scene, controller, client_name));
  }
  catch(jack_error& e)
  {
    ERROR(e.what());
  }
  catch(gen_error& e)
  {
    ERROR(e.what());
  }
  return temp;
}

bool GenericRenderer::new_source(id_t id, const std::string& port_name)
{
  (void) id;
  (void) port_name;
  
  ERROR("This function may not be called.");
  
  return false;
}

bool GenericRenderer::new_source(id_t id, const std::string& port_name,
			     const std::string& imp_resp_file_name)
{

  if (_input_map.find(id) != _input_map.end())
  {
    WARNING("Input '" << id << "' already exists!");
    return false;
  }
  Input::ptr_t temp(Input::create(*this,
          _scene,
          _controller,
          id,
	  imp_resp_file_name,
          _fade_in,
	  _fade_out,
	  _output_map.size(),
          _input_prefix + "_" + A2S(id)));
  if (is_null(temp))
  {
    WARNING("Initialization of input '" << id << "' failed!");
    return false;
  }

  // connect portname with input port
  if (!this->connect_ports(port_name, temp->get_port_name()))
  {
    return false;
  }
  _input_map[id] = temp.release();


  VERBOSE("GenericRenderer: Input " << id << " registered.");
  return true;
}

/** _.
 * This will only be called once. Loudspeakers can @b not be added later.
 * %Loudspeaker-IDs always range from 1 to @a number.
 * @param number number of loudspeakers
 * @return @b true on success.
 **/
bool GenericRenderer::_make_outputs()
{
  if (!_output_map.empty())
  {
    WARNING("BUG: Output map is not empty!");
    maptools::purge(_output_map);
  }

  // get loudspeakers
  Loudspeaker::container_t loudspeakers;
  _scene.get_loudspeakers(loudspeakers);

  Loudspeaker::container_t::const_iterator i;
  Scene::loudspeakers_t::size_type id = 1;

  // make outputs
  for (i = loudspeakers.begin(); i != loudspeakers.end(); ++i)
  {
    Output::ptr_t temp(Output::create(*this, _output_prefix + "_" + A2S(id)));
    
    if (is_null(temp))
    {
      WARNING("Initialization of output " << id << " failed!");
      maptools::purge(_output_map);
      return false;
    }
    _output_map[id] = temp.release();

    ++id;
  }

  VERBOSE("GenericRenderer: " << _output_map.size() << " Output objects registered.");
  return true;
}

bool GenericRenderer::delete_source(const id_t id)
{
  return maptools::delete_item(_input_map, id);
}

void GenericRenderer::delete_all_sources()
{
  maptools::purge(_input_map);
  // TODO: anything else?
}

/** _.
 * @param id identifier of the input channel
 * @return Pointer to the corresponding @a id. If none was found, @b NULL is
 * returned.
 **/
GenericRenderer::Input* GenericRenderer::_get_input(id_t id)
{
  return maptools::get_item(_input_map, id);
}

/** fill a buffer with zeroes.
 * @param buffer pointer to the memory to be cleared
 * @param buffer_size number of data elements to be cleared
 **/
void GenericRenderer::clear_buffer(sample_t* const buffer,
    const jack_nframes_t buffer_size) const
{
  if (buffer_size != _silence.size())
  {
    // this should never happen because JACK buffer size will not change!
    WARNING("BUG?: silence resized! (" << _silence.size() << " -> "
        << buffer_size << ")");
    _silence.resize(buffer_size, 0);
  }
  else
  {
    std::copy(_silence.begin(), _silence.end(), buffer);
  }
}

void GenericRenderer::get_output_port_names(port_list_t& port_list) const
{
  for (output_map_t::const_iterator i = _output_map.begin();
      i != _output_map.end(); ++i)
  {
    assert(i->second != NULL);
    port_list.push_back(i->second->get_port_name());
  }
}

/** JACK process callback function.
 * This function is called every time the JACK server has new audio data.
 * @param nframes number of frames
 * @return I don't know what
 */
int GenericRenderer::jack_process(const jack_nframes_t nframes)
{
// code must be suitable for real-time execution.
// That means following function are not allowed:
// malloc, free, printf, pthread_mutex_lock, sleep,
// wait, poll, select, pthread_join, pthread_cond_wait,
// etc, etc.

  // clear all information about active/inactive loudspeakers
  _controller.clear_all_active_sources();
  // send some information to the controller (and then to the subscribers)
  _controller.set_transport_state(this->get_transport_state());
  _controller.set_cpu_load(this->get_cpu_load());

  switch (_scene.get_processing_state())
  {
    case processing:
      {
        
	output_buffer_list_t output_buffer_list;

	for (output_map_t::const_iterator i = _output_map.begin();
	     i != _output_map.end(); ++i)
        {
          assert(i->second != NULL);
          i->second->write_empty_buffer(nframes);

	  output_buffer_list.push_back(i->second->get_output_buffer(nframes));
        }

        SourceCopy::container_t sources;
        _scene.get_sources(sources);

        for (SourceCopy::container_t::iterator i = sources.begin();
             i != sources.end(); ++i)
        {
          if (_input_map.find(i->id) == _input_map.end())
          {
            // This should never happen!
            ERROR("Source #" << i->id << " not found in input map!");
          }
          else
          {
            *_input_map[i->id] = *i; // assign new gain, mute, ...
	    
            _input_map[i->id]->apply_filters(output_buffer_list, 
					     _scene.get_corrected_master_volume(),
					     nframes);
          }
        }

	float master_level = 0.0f;

	for (output_map_t::const_iterator i = _output_map.begin();
	     i != _output_map.end(); ++i)
        {
          assert(i->second != NULL);

	  master_level = std::max(master_level, i->second->get_signal_level(nframes));
        }
	_controller.set_master_signal_level(master_level);

        break;
      }

    case ready: // audio processing is OFF
      
      // clear all output buffers
      for (output_map_t::const_iterator i = _output_map.begin();
	   i != _output_map.end(); ++i)
      {
	assert(i->second != NULL); // assert that object really exists
        i->second->write_empty_buffer(nframes);
      };

      break;

    case initializing:
      // WARNING("JACK process callback called while in \"initializing\" state!");
      // return something bad?

      break;

    case exiting:
      WARNING("JACK process callback called while in \"exiting\" state!");
      // return something bad?
      break;

    default:
      ERROR("BUG: Unknown client state!");
      // return something bad?
  }
  // TODO: find out what the return value means
  return 0;
}

unsigned int GenericRenderer::get_number_of_output_channels() const
{
  return _output_map.size();
}

Renderer::renderer_t GenericRenderer::get_renderer_type() const
{
  return Renderer::generic;
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class GenericRenderer::SourceCopy
////////////////////////////////////////////////////////////////////////////////

GenericRenderer::SourceCopy::SourceCopy(const std::pair<id_t, Source>& other) :
  id(other.first),
  gain(other.second.gain),
  mute(other.second.mute)
{}

GenericRenderer::Output::Output(const GenericRenderer& renderer,
    const std::string& port_name_) throw (output_error) :
  _renderer(renderer),
  _port(_renderer.register_out_port(port_name_)),
  // get actual port name and save it to member variable
  _port_name(_port ? jack_port_name(_port) : "")
{
  if (_port == NULL)
  {
    throw output_error("Output port '" + port_name_
        + "' could not be registered.");
  }
}

GenericRenderer::Output::~Output()
{
  if (_port) _renderer.unregister_port(_port);
}

GenericRenderer::Output::ptr_t GenericRenderer::Output::create(
			  const GenericRenderer& renderer,
			  const std::string& port_name)
{
  ptr_t temp; // temp = NULL
  try
  {
    temp.reset(new Output(renderer, port_name));
  }
  catch(output_error& e)
  {
    ERROR(e.what());
  }
  return temp;
}

float GenericRenderer::Output::get_signal_level(const jack_nframes_t nframes)
{
  // get JACK output buffer
  sample_t* out = static_cast<sample_t*>
                    (jack_port_get_buffer(_port, nframes));

  return _renderer.get_max_signal_amplitude(out, nframes);
}

void GenericRenderer::Output::write_empty_buffer(const jack_nframes_t nframes) const
{
  if (_port)
  {
    sample_t* out = static_cast<sample_t*>
                      (jack_port_get_buffer(_port, nframes));
    _renderer.clear_buffer(out, nframes);
  }
}

Renderer::sample_t* GenericRenderer::Output::get_output_buffer(
				      const jack_nframes_t nframes)
{
  return static_cast<sample_t*>(jack_port_get_buffer(this->_get_port(), nframes));
}

jack_port_t* GenericRenderer::Output::_get_port() const
{
  return _port;
}

std::string GenericRenderer::Output::get_port_name() const
{
  return _port_name;
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class GenericRenderer::Input
////////////////////////////////////////////////////////////////////////////////

/** private ctor. Object can only be created with create().
 * @param renderer reference to the calling GenericRenderer object. Using this, we
 * can access the public members of the BRS renderer.
 **/
GenericRenderer::Input::Input(const GenericRenderer& renderer,
        const Scene& scene,
        Controller& controller,
        const id_t source_id,
	const std::vector<sample_t>& fade_in,
	const std::vector<sample_t>& fade_out) :
  SourceCopy(source_id),
  _renderer(renderer),
  _scene(scene),
  _controller(controller),
  _source_id(source_id),
  //_previous_gain(0.0f),
  _port(NULL),
  _port_name(""),
  _fade_in(fade_in),
  _fade_out(fade_out),
  _filter_cycles(0u)
{}

GenericRenderer::Input::~Input()
{
  if (_port) _renderer.unregister_port(_port);

  // delete convolvers
  for (convolver_list_t::iterator convolver_iterator = _convolver_list.begin();
       convolver_iterator != _convolver_list.end(); ++convolver_iterator)
  {
    delete *convolver_iterator;
  }

  _convolver_list.clear();
}

GenericRenderer::Input::ptr_t GenericRenderer::Input::create(
        const GenericRenderer& renderer,
        const Scene& scene,
        Controller& controller,
        const id_t source_id,
	const std::string& imp_resp_file_name,
        const std::vector<sample_t>& fade_in,
	const std::vector<sample_t>& fade_out,
	const unsigned int no_of_output_channels,
        const std::string& name)
{
  ptr_t temp(new Input(renderer, scene, controller, 
		       source_id, fade_in, fade_out));
  if (!temp->_init(name, imp_resp_file_name, no_of_output_channels))
  {
    ERROR("Unable to initialize input '" + name + "'!");
    destroy(temp);
  }
  return temp;
}

GenericRenderer::Input& GenericRenderer::Input::operator=(const SourceCopy& source)
{
  this->id      = source.id;
  this->gain    = source.gain;
  this->mute    = source.mute;
  return *this;
}

bool GenericRenderer::Input::_init(const std::string& port_name, 
				   const std::string& imp_resp_file_name,
				   const unsigned int no_of_output_channels)
{
  _port = _renderer.register_port(port_name, JackPortIsInput);

  if (_port == NULL)
  {
    WARNING("Input port '" + port_name + "' could not be registered.");
    return false;
  }

  // get actual port name and save it to member variable
  _port_name = jack_port_name(_port);

  _audio_buffer.resize(_renderer.get_buffer_size(), 0.0f);
  
  if (!_load_impulse_responses(imp_resp_file_name, no_of_output_channels)) 
  {
    return false;
  }
      
  return true;

}

bool GenericRenderer::Input::_load_impulse_responses(
				     const std::string& file_name,
				     const unsigned int no_of_output_channels)
{
  /////////////////// load impulse responses //////////////////////

  // get sample rate of hrirs
  const int sr = audiofileiotools::get_sample_rate(file_name);
  
  if (!sr) return false;

  // if hrirs do not have the correct sample rate
  else if (sr != static_cast<int>(_scene.get_sample_rate()))
  {
    ERROR("The impulse responses in file '"
          + file_name + "' are sampled at rate " + A2S(sr)
          + " Hz. JACK is currently running at "
          + A2S(_scene.get_sample_rate()) + " Hz.");

    return false;
  }

  std::cout << "Loading impulse responses from file '" << 
                           file_name  << "'. Please be patient.\n";

  // check number of impulse responses
  const int no_of_ch = audiofileiotools::get_no_of_channels(file_name);

  // error message comes from function get_no_of_channels()
  if (no_of_ch == 0) return false;

  if (static_cast<unsigned int>(no_of_ch) != no_of_output_channels)
  {
    ERROR("Number of channels in file '" << file_name 
	  << "' does not correspond to number of outputs (" 
	  << no_of_ch << " vs. " << no_of_output_channels << ").");

    return false;
  }

  Convolver::data_t container;

  if (!audiofileiotools::load_filter_t(file_name, container))
  {
    return false;
  }

  ///////////////////// determine filter length /////////////////////

  const unsigned int imp_resp_size = container.size()/no_of_ch;    

  const unsigned int power_of_2 =
                mathtools::next_power_of_2(imp_resp_size);

  const unsigned int convolver_buffer_size = 
                std::min(power_of_2, _renderer.get_buffer_size());

  _filter_cycles = _renderer.get_buffer_size()/convolver_buffer_size;

  if (_filter_cycles != 1u)
  {
    VERBOSE("I create " << _filter_cycles << " subcycles of size " 
	    << convolver_buffer_size << " for convolution.");
  }

  //////////////////// create convolvers /////////////////////////

  _convolver_list.clear();
  
  // create convolvers and distribute impulse responses
  for (unsigned int current_channel = 0; 
       current_channel < no_of_output_channels; ++current_channel)
  {
    // create convolver
    Convolver::ptr_t temp(Convolver::create(convolver_buffer_size, 
					    Convolver::none));
    if (is_null(temp))
    {
      WARNING("Initialization of convolver failed.");
      return false;
    }

    // prepare filter
    Convolver::data_t dummy;
    dummy.clear();

    // extract individual channel
    for (unsigned int n = 0; n < imp_resp_size; ++n)
    {
      dummy.push_back(*(container.data() + 
			current_channel*imp_resp_size + n));
    }
    
    // set the filter
    temp->set_filter_t(dummy);

    // next convolver means next channel of impulse response file
    _convolver_list.push_back(temp.release());

  }

  return true;
}

bool GenericRenderer::Input::_fetch_new_buffer(const jack_nframes_t nframes)
{
  sample_t* in = static_cast<sample_t*>(jack_port_get_buffer(_port, nframes));
  if (in == NULL)
  {
    WARNING("Got no input buffer from JACK!");
    return false;
  }
  assert(in != NULL);

  memcpy(&_audio_buffer[0], in, nframes*sizeof(sample_t));

  return true;
}

void GenericRenderer::Input::_weight_audio_data(const float gain)
{
  //sample_t* audio_sample = &_audio_buffer[0];

  //// weight audio signal; C-style to be efficient
  //for (unsigned int n=0
  //    ; n < _renderer.get_buffer_size()
  //    ; ++n, ++audio_sample)
  //{
  //  // crossfade
  //  *audio_sample *= gain * _fade_in[n] + _previous_gain * _fade_out[n];
  //}

  //_previous_gain = gain;

  // update source signal level (max amplitude)
  _controller.set_source_signal_level(_source_id
      , _renderer.get_max_signal_amplitude(&_audio_buffer[0]
      , _renderer.get_buffer_size()) * _scene.get_source_gain(_source_id)
      * gain);
}

void GenericRenderer::Input::apply_filters(
    output_buffer_list_t& output_buffer_list
    , const float master_volume
    , const jack_nframes_t nframes)
{
  _fetch_new_buffer(nframes);

  // apply gain
  if (this->mute) _weight_audio_data(0.0f);
  else _weight_audio_data(this->gain*master_volume);
  // WARNING: this is not actually used for weighting, see above!

  const jack_nframes_t length_of_subframe = nframes/_filter_cycles;

  output_buffer_list_t::const_iterator output_iterator = output_buffer_list.begin();

  // convolver 1 -> output_buffer 1
  // convolver 2 -> output_buffer 2
  //    ...      ->      ...
  for (convolver_list_t::iterator convolver_iterator = _convolver_list.begin();
       convolver_iterator != _convolver_list.end(); ++convolver_iterator)
  {
    if (output_iterator == output_buffer_list.end())
    {
      ERROR("Number of output buffers does not correspond to number of convolvers.");
      break;
    }
    
    // do convolution in subcycles
    for (unsigned int cycle = 0; cycle < _filter_cycles; cycle++)
    {
      const unsigned int sample_pos = nframes/_filter_cycles * cycle;

      sample_t* sig = (*convolver_iterator)->
	convolve_signal(&_audio_buffer[0] + sample_pos, this->gain*master_volume);

      // write to buffer
      for (jack_nframes_t n = 0; n < length_of_subframe; ++n)
      {
	*(*output_iterator+n+sample_pos) += *(sig+n);
      }
    }

    // next output
    ++output_iterator;
  }
}

std::string GenericRenderer::Input::get_port_name()
{
  return _port_name;
}

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
