/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * Generic renderer class (definition).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef GENERICRENDERER_H
#define GENERICRENDERER_H

#include <map>
#include <set>
#include <vector>

#include "renderer.h"   // base class of GenericRenderer
#include "directionalpoint.h" // Position and Orientation
#include "convolver.h"

class Controller; // forward declaration

using ssr::id_t;

/** Binaural Room Scanning Renderer class.
 * This class does any rendering. Audio processing is done with the
 * inherited JACK functionality.
 **/
class GenericRenderer : public Renderer
{
  private:
    class Output;          // nested class, defined later
    class Input;           // nested class, defined later
    /** Class to temporarily store selected source information.
     * By use of a class like this, the desired information for every source can
     * be obtained using the function Scene::get_source_list().
     **/
    struct SourceCopy
    {
      SourceCopy(id_t id) : id(id) {};
      
      typedef std::vector<SourceCopy> container_t; ///< list of SourceCopys
      /// type conversion constructor
      SourceCopy(const std::pair<id_t, Source>& other);

      id_t id;               ///< unique ID, see Scene::source_map_t
      float gain;            ///< gain, see Source
      bool mute;             ///< mute/unmute, see Source
    };


  public:
    typedef std::auto_ptr<GenericRenderer> ptr_t;

    /// a map which holds all output objects.
    typedef std::map<Loudspeaker::container_t::size_type, Output*> output_map_t;
    /// a map which holds all input objects.
    typedef std::map<id_t, Input*> input_map_t;

    typedef std::list<sample_t*> output_buffer_list_t;

    /// exception to be thrown by ctor.
    struct gen_error : public std::runtime_error
    {
      gen_error(const std::string& s): std::runtime_error(s) {}
    };

    GenericRenderer(const Scene& scene, Controller& controller,
            const std::string& client_name) throw (jack_error, gen_error);

    /// "named constructor" for %GenericRenderer objects
    static ptr_t create(const Scene& scene, Controller& controller,
            const std::string& client_name = "Generic-Renderer");

    /// delete all Input/Output objects and clear input and output map.
    virtual ~GenericRenderer();

    /// set of impulse responses
    typedef std::vector<Convolver::data_t> impulse_response_container_t;

    // from JackClient:
    virtual int jack_shutdown();
    // from Renderer:
    virtual bool new_source(id_t id, const std::string& port_name);
    virtual bool new_source(id_t id, const std::string& port_name,
			    const std::string& imp_resp_file);
    virtual bool delete_source(id_t id);
    virtual void delete_all_sources();
    virtual void get_output_port_names(port_list_t& port_list) const;
    virtual unsigned int get_number_of_output_channels() const;
    virtual renderer_t get_renderer_type() const;

    void clear_buffer(sample_t* const buffer, jack_nframes_t buffer_size) const;

  private:
    /// overwrites virtual member function declared in JackClient class
    virtual int jack_process(jack_nframes_t nframes);

    /// this will be filled with zeroes
    mutable std::vector<sample_t> _silence;

    input_map_t   _input_map;    ///< all input objects are stored here
    output_map_t  _output_map;   ///< all output objects are stored here

    std::vector<sample_t> _fade_in;
    std::vector<sample_t> _fade_out;

    Controller& _controller;          ///< Reference to the Controller
    const std::string  _input_prefix; ///< prefix for the input  ports
    const std::string _output_prefix; ///< prefix for the output ports

    /// Get pointer to input object corresponding to @a id.
    Input* _get_input(id_t id);

    bool _make_outputs();

}; // end of GenericRenderer class

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Output
////////////////////////////////////////////////////////////////////////////////

class GenericRenderer::Output : public Loudspeaker
{
  public:
    typedef std::auto_ptr<Output> ptr_t;

    /// exception to be thrown by ctor.
    struct output_error : public std::runtime_error
    {
      output_error(const std::string& s): std::runtime_error(s) {}
    };

    Output(const GenericRenderer& renderer, const std::string& port_name)
      throw (output_error);

    /// "named constructor"
    static ptr_t create(const GenericRenderer& renderer,
        const std::string& port_name);

    ~Output(); ///< Unregister associated JACK output port

    Renderer::sample_t* get_output_buffer(const jack_nframes_t nframes);

    float get_signal_level(const jack_nframes_t nframes);

    /// clear output buffer
    void write_empty_buffer(jack_nframes_t nframes) const;

    std::string get_port_name() const;

  private:
    const GenericRenderer& _renderer; /// ref. to the calling Renderer object
    jack_port_t* _port;     ///< JACK port associated with this output
    const std::string _port_name; ///< full name of the JACK port

    jack_port_t* _get_port() const;

};

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Input
////////////////////////////////////////////////////////////////////////////////

/** Manages an input buffer (including delay line) and the corresponding JACK
 * calls.
 **/
class GenericRenderer::Input : public SourceCopy
{
  public:

    typedef std::auto_ptr<Input> ptr_t;
    /// "named constructor"
    static ptr_t create(const GenericRenderer& renderer,
            const Scene& scene,
            Controller& controller,
            const id_t source_id,
	    const std::string& imp_resp_file_name,
            const std::vector<sample_t>& fade_in,
	    const std::vector<sample_t>& fade_out,
            const unsigned int no_of_output_channels,
            const std::string& name);

    /// Unregisters associated JACK input port
    ~Input();

   Input& operator=(const SourceCopy&);

    std::string get_port_name(); ///< get JACK port name of the input object

    void apply_filters(output_buffer_list_t& output_buffer_list,
		       const float master_volume,
		       const jack_nframes_t nframes);

  private:

    // if it were not private, I'd make it "explicit"
    Input(const GenericRenderer& renderer,
          const Scene& scene,
          Controller& controller,
          const id_t source_id,
	  const std::vector<sample_t>& fade_in,
	  const std::vector<sample_t>& fade_out);

    /// creates a JACK port and maybe does some other things, too.
    bool _init(const std::string& port_name, 
	       const std::string& imp_resp_file_name,
	       const unsigned int no_of_output_channels);

    bool _load_impulse_responses(const std::string& file_name,
				 const unsigned int no_of_output_channels);

    bool _fetch_new_buffer(const jack_nframes_t nframes);

    void _weight_audio_data(const float gain);

    /// holds a reference to the calling GenericRenderer object
    const GenericRenderer& _renderer;
    /// holds a reference to the scene

    const Scene& _scene;
    /// Reference to the Controller

    Controller& _controller;

    /// id of associated sound source
    id_t _source_id;

    //float _previous_gain;

    /// the JACK port corresponding to this input object
    jack_port_t* _port;

    /// full name of the JACK port
    std::string _port_name;

    /// pointer to audio data of current frame
    std::vector<sample_t> _audio_buffer;

    const std::vector<sample_t>& _fade_in;
    const std::vector<sample_t>& _fade_out;

    typedef std::list<Convolver*>  convolver_list_t;

    convolver_list_t _convolver_list;
    
    unsigned int _filter_cycles;
};

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
