/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * JACK client (implementation).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include <cassert> // for assert()
#include <errno.h> // for EEXIST
#include <cmath>   // for fabs()
#include <cstring>
#include <jack/statistics.h> // for jack_get_xrun_delayed_usecs()
#include "jackclient.h"
#include "ssr_global.h" // for WARNING(), ...

/** JACK error callback.
 * Outputs a error string for JACK errors.
 * @param msg the error message
 **/
static void jack_error_function(const char* msg)
{
  ERROR("JACK: " + std::string(msg));
}

JackClient::JackClient(const std::string& client_name,
    const bool use_jack_process_callback) throw (jack_error) :
  _client_name(client_name),
  _client(NULL),   // will be set after connecting to JACK
  _sample_rate(0), //             -- " -- 
  _buffer_size(0)  //             -- " -- 
{
  // check if client_name is too long
  // jack_client_name_size() returns the size including the terminating \0
  // character, hence the ">=".
  if (static_cast<int>(client_name.size()) >= jack_client_name_size())
  {
    throw jack_error("JACK client name is too long! ('" + client_name + "')");
  }

  _client = jack_client_new(client_name.c_str());
  if (_client == NULL)
  {
    throw jack_error("\"" + client_name
        + "\": JACK server not running? (or duplicate client name?)");
  }
  assert(_client_name == jack_get_client_name(_client));

  jack_set_error_function(jack_error_function);

  if (!jack_is_realtime(_client))
  {
    VERBOSE("JACK server does not have realtime priority.");
  }

  if (use_jack_process_callback)
  {
    VERBOSE("Setting process callback ('" + _client_name + "').");
    if (jack_set_process_callback(_client, _jack_process, this))
    {
      throw jack_error("Could not set process callback function for '"
          + _client_name + "'!");
    }
  }

  jack_on_shutdown(_client, _jack_shutdown, this);

  //  if (jack_set_port_registration_callback(_client, _port_registration_callback, this))
  //  {
  //    ERROR("Could not set port registration callback function for '"
  //    + _client_name + "'!");
  //  }

  if (jack_set_xrun_callback(_client, _xrun_callback, this))
  {
    throw jack_error("Could not set xrun callback function for '"
        + _client_name + "'!");
  }


  // sometimes, jack_activate() returns successful although an error occured and
  // the thing "zombified". if the shutdown handler is called, _client is reset
  // to zero which we can check now:
  if (!_client)
  {
    throw jack_error("\"" + _client_name + "\": JACK was killed somehow!");
  }

  _sample_rate = jack_get_sample_rate(_client);
  _buffer_size = jack_get_buffer_size(_client);

  //_connections_map.clear();
}

JackClient::~JackClient()
{
  if (_client)
  {
    //jack_deactivate(_client);
    VERBOSE_NOLF("JackClient: Closing JACK client '" + _client_name + "' ... ");
    if (jack_client_close(_client))
    {
      VERBOSE("failed!");
      // nothing else is done, because we're shutting down anyway ...
    }
    else
    {
      VERBOSE("done.");
    }
  }
}

void JackClient::activate()
{
  if( _client )
  {
    VERBOSE_NOLF("\"" + _client_name + "\": Trying to activate ... ");
    if (jack_activate(_client))
    {
      VERBOSE("failed!");
    }
    else
    {
      VERBOSE("done.");
    }
  }
}

void JackClient::deactivate()
{
  if( _client )
    jack_deactivate(_client);
}

/** _.
 * @param client_name client name of the JACK client to be created.
 * @param use_jack_process_callback if @a true, the member function
 * jack_process() is called by JACK in each audio cycle.
 * @return std::auto_ptr to the new JackClient object.
 * @warning @a client_name should not include a colon. This doesn't cause an
 * error directly, but it messes up the JACK client- and portnames.
 **/
JackClient::ptr_t JackClient::create(const std::string& client_name,
    const bool use_jack_process_callback)
{
  ptr_t temp; // temp = NULL
  try
  {
    temp.reset(new JackClient(client_name, use_jack_process_callback));
  }
  catch(jack_error& e)
  {
    ERROR(e.what());
  }
  return temp;
}

std::string JackClient::get_client_name() const
{
  return _client_name;
}

jack_nframes_t JackClient::get_sample_rate() const
{
  return _sample_rate;
}

jack_nframes_t JackClient::get_buffer_size() const
{
  return _buffer_size;
}

jack_port_t* JackClient::register_port(const std::string& port_name,
    const unsigned int flags) const
{
  return jack_port_register(_client, port_name.c_str(), JACK_DEFAULT_AUDIO_TYPE,
      flags, 0);
}

jack_port_t* JackClient::register_in_port(const std::string& port_name) const
{
  return this->register_port(port_name, JackPortIsInput);
}

jack_port_t* JackClient::register_out_port(const std::string& port_name) const
{
  return this->register_port(port_name, JackPortIsOutput);
}

bool JackClient::unregister_port(jack_port_t* port) const
{
  VERBOSE2_NOLF("Unregistering '" + std::string(jack_port_name(port))
      + "' ... ");
  if (!jack_port_unregister(_client, port))
  {
    VERBOSE2("done.");
    port = NULL; // port still points somewhere. Set to NULL.
    return true;
  }
  else
  {
    VERBOSE2("failed!");
    return false;
  }
}

bool JackClient::connect_ports(const std::string& source_port,
    const std::string& destination_port) const
{
  VERBOSE2_NOLF("Connecting '" + source_port + "' and '" + destination_port
      + "' ... ");
  if (_client == NULL)
  {
    VERBOSE2("failed!");
    ERROR("Client not found! Not connected.");
    return false;
  }
  int success = jack_connect(_client, source_port.c_str(),
      destination_port.c_str());
  switch (success)
  {
    case 0:
      VERBOSE2("done.");
      break;
    case EEXIST:
      VERBOSE2("already connected!");
      WARNING("Connection already exists! (" + source_port + " -> "
          + destination_port + ")");
      break;
    default:
      VERBOSE2("failed!");
      ERROR("Couldn't connect '" + source_port + "' with '"
          + destination_port + "'!");
      return false;
      break;
  }
  return true;
}


bool JackClient::disconnect_ports(const std::string& source_port,
    const std::string& destination_port) const
{
  VERBOSE2("Disconnecting '" + source_port + "' and '"
      + destination_port + "'.");
  int success = jack_disconnect(_client, source_port.c_str(),
      destination_port.c_str());
  if (!success)
  {
    ERROR("Couldn't disconnect '" + source_port + "' and '"
          + destination_port + "'!");
    return false;
  }
  return true;
}

void JackClient::transport_start() const
{
  if (_client) jack_transport_start(_client);
}

void JackClient::transport_stop() const
{
  if (_client) jack_transport_stop(_client);
}

bool JackClient::transport_locate(jack_nframes_t frame) const
{
  if (_client) return !jack_transport_locate(_client, frame);
  else return false;
}

//JackClient::State JackClient::get_transport_state() const
std::pair<bool, jack_nframes_t> JackClient::get_transport_state() const
{
  std::pair<bool, jack_nframes_t> result(false, 0);

  if (_client)
  {
    jack_position_t jp;
    jack_transport_state_t jts = jack_transport_query(_client, &jp);
    result.first  = (jts == JackTransportRolling);
    result.second =  jp.frame;
  }
  return result;
}

bool JackClient::set_freewheel(int onoff) const
{
  if (_client) return !jack_set_freewheel(_client, onoff);
  else return false;
}

/** JACK process callback redirection.
 * Our own JACK callback is launched from here.
 * @param nframes number of frames available from JACK server
 * @param *arg pointer to the calling class instance.
 * @return ?
 **/
int JackClient::_jack_process(jack_nframes_t nframes, void *arg)
{
  return static_cast<JackClient*>(arg)->jack_process(nframes);
}

/** _.
 * This function is empty in the JackClient base class. Derived classes should
 * reimplement it if needed.
 * @param nframes Number of frames (~samples) in the current block. This value
 * is delivered by the JACK server.
 **/
int JackClient::jack_process(const jack_nframes_t)
{
  WARNING("JACK process callback not implemented!");
  // TODO: find out what return value means
  // answer: 0 means call me again, 1 dont call me anymore.
  return 0;
}

/** JACK shutdown callback redirection.
 * This function is called when the JACK server disconnects the client.
 * It gets redirected to our own JACK shutdown callback.
 * @param *arg pointer to the calling class instance.
 * @return ?
 **/
void JackClient::_jack_shutdown(void *arg)
{
  static_cast<JackClient*>(arg)->jack_shutdown();
}

/** JACK shutdown callback.
 * @return I don't know what
 */
int JackClient::jack_shutdown()
{
  _client = NULL;
  ERROR("JACK shutdown!");
  return 1; // TODO: what to return?
}

/** JACK sample rate callback redirection.
 * Call jack_sample_rate_callback() with a pointer to the current object.
 * @param sample_rate new sample rate delivered by JACK
 * @param arg Pointer to current object (this)
 * @warning This function will/should not be used, because sample rate changes
 * will most probably not work anyway.
 **/
int JackClient::_jack_sample_rate_callback(const jack_nframes_t sample_rate,
    void* arg)
{
  return static_cast<JackClient*>(arg)->jack_sample_rate_callback(sample_rate);
}

int JackClient::jack_sample_rate_callback(const jack_nframes_t sample_rate)
{
  WARNING("Sample rate changes are not supported!");
  _sample_rate = sample_rate;
  return 0;
}

/** JACK buffer size callback redirection
 * Call jack_buffer_size_callback() with a pointer to the current object.
 * @param buffer_size new sample rate delivered by JACK
 * @param arg Pointer to current object (this)
 * @warning This function will/should not be used, because buffer size changes
 * will most probably not work anyway.
 **/
int JackClient::_jack_buffer_size_callback(jack_nframes_t buffer_size,
    void *arg)
{
  return static_cast<JackClient*>(arg)->jack_buffer_size_callback(buffer_size);
}

int JackClient::jack_buffer_size_callback(jack_nframes_t buffer_size)
{
  WARNING("Buffer size changes are not supported!");
  _buffer_size = buffer_size;
  // TODO: find out what return value means
  return 0;
}

/*
bool JackClient::_find_port(const std::string& port_name)
{
  ERROR("Do not use this function. It is erroneous.");

  return false;

  // this is necessary to avoid errors in case port_name 
  // contains a [ in the beginning, then then the port is 
  // not found...

  // get port list
  jack_port_t* jack_port = 
    jack_port_by_name(_client, port_name.substr(1).c_str());
  
  // check if it contains something
  if (jack_port == NULL) 
  {
    WARNING("Don't find port.");;
    return false;
  }
  else 
  {
    free(jack_port);
    WARNING("Port found.");
    return true;
  }

}
*/
/*
void JackClient::_port_registration_callback(jack_port_id_t port_id, 
               int reg, void *arg)
{
  static_cast<JackClient*>(arg)->port_registration_callback(port_id, reg);
}
*/
/*
void JackClient::port_registration_callback(jack_port_id_t port_id, int reg)
{

  ERROR("Do not use this function. It is erroneous.");

  return;

  if (reg==0) return; // port is closed. We don't care.

  // get the port's name
  const std::string registered_port = 
    jack_port_name(jack_port_by_id(_client, port_id));

  // check if we want to use the port
  connections_map_t::iterator i;

  for (i = _connections_map.begin(); i != _connections_map.end(); i++)
  {
    if (i->first == registered_port)
    {
      if (_find_port(i->second))
      {
	if (!connect_ports(i->first, i->second))
	{
	  ERROR("Can not connect ports '" << i->first 
		<< "' and '" << i->second << "'."); 
	}
      }
      else break;
    } // if (i->first == registered_port)

    if (i->second == registered_port)
    {
      if (_find_port(i->first))
      {
	if (!connect_ports(i->first, i->second))
	{
	  ERROR("Can not connect ports '" << i->first 
		<< "' and '" << i->second << "'."); 
	}
      }
      else break;
    } // if (i->second == registered_port)
  } // for
  
}
 */
int JackClient::_xrun_callback(void *arg)
{
  return static_cast<JackClient*>(arg)->xrun_callback();
}

int JackClient::xrun_callback()
{
  VERBOSE2("JACK server reports xrun of "
    << jack_get_xrun_delayed_usecs(_client)/1000.0f << " msecs.");

  return 0;
}

float JackClient::get_cpu_load()
{
  return jack_cpu_load(_client);
}

float JackClient::get_max_signal_amplitude(sample_t *in, jack_nframes_t nframes) const
{
  float maximum = 0.0f;

  // find maximum signal level 
  for (jack_nframes_t n = 0; n < nframes; n++ )
  {
    // C-style to be efficient
    maximum = std::max(maximum, std::fabs(*in));
    in++;
  }
  return maximum;
}

float JackClient::get_signal_rms(sample_t *in, jack_nframes_t nframes) const
{
  // calculate signal RMS level
  float rms = 0.0f;
  const float norm = 1.0f / static_cast<float>(nframes);

  for (jack_nframes_t  n = 0; n < nframes; n++ )
  {
    // C-style to be efficient
    rms += (*in) * (*in) * norm;
    in++;
  }

  return rms;
}

/*
void JackClient::new_connection(const std::string source_port,
        const std::string destination_port)
{
  // connect ports if they exist
  if (_find_port(source_port) 
      && _find_port(destination_port))
  {
    if (!connect_ports(source_port, destination_port))
    {
      ERROR("Can not connect ports '" << source_port 
	    << "' and '" << destination_port << "'."); 
      return;
    }
  }
  
  // if connection could not have been established add it to map
  _connections_map[source_port] = destination_port;
}
*/

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
