/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * JACK client (definition).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef JACKCLIENT_H
#define JACKCLIENT_H

#ifdef HAVE_CONFIG_H
#include <config.h> // for ENABLE_*, HAVE_*, WITH_*
#endif

#include <jack/jack.h>
#include <jack/thread.h>
#include <string>
#include <map>
#include <utility>   // for std::pair<>
#include <stdexcept> // for std::runtime_error
#include <memory>

#include "classtools.h"

/** C++ wrapper for a JACK client.
 * @warning When several JACK clients are running and one of them is closed,
 * this can lead to a segmentation fault in the callback function
 * (jack_port_get_buffer() delivers bad data). I couldn't really track down the
 * error, but to be on the sure side, delete your JackClients only on the very
 * end (or just let the std::auto_ptr do its work).
 * \par
 * A related issue may be that after calling jack_client_close() the
 * corresponding thread is not closed, in the end it becomes a "zombie thread".
 * But maybe this is a different story ...
 * \par
 * Any comments on this topic are very welcome!
 **/
class JackClient : classtools::NonCopyable
{
  public:
    /// this_is_actually_a_too_long_type_name_to_write_under_any_circumstances
    typedef jack_default_audio_sample_t sample_t;
    typedef std::auto_ptr<JackClient>   ptr_t; ///< auto_ptr to JackClient

    /// exception to be thrown by ctor.
    struct jack_error : public std::runtime_error
    {
      jack_error(const std::string& s): std::runtime_error(s) {}
    };

    /// "named constructor" for JackClient objects
    static ptr_t create(const std::string& client_name = "JackClient",
        bool use_jack_process_callback = false);

    /// ctor.
    /// @warning Constructor may throw an exception! Use JackClient::create(...)
    /// if you don't want to care about that.
    /// @throw jack_error
    explicit JackClient(const std::string& client_name,
        bool use_jack_process_callback = false) throw (jack_error);

    /// Deactivate and close the client.
    virtual ~JackClient();

    std::string get_client_name() const;    ///< get name of JACK client
    jack_nframes_t get_sample_rate() const; ///< get JACK sample rate
    jack_nframes_t get_buffer_size() const; ///< get JACK buffer size

    void activate();
    void deactivate();

    /// register JACK port
    jack_port_t* register_port(const std::string& port_name,
        const unsigned int flags) const;
    /// register JACK input port
    jack_port_t* register_in_port (const std::string& port_name) const;
    /// register JACK output port
    jack_port_t* register_out_port(const std::string& port_name) const;
    /// unregister JACK port
    bool unregister_port(jack_port_t* port) const;

    /// connect two JACK ports
    bool connect_ports(const std::string& source_port,
        const std::string& destination_port) const;
    /// disconnect two JACK ports
    bool disconnect_ports(const std::string& source_port,
        const std::string& destination_port) const;

    void transport_start() const; ///< start JACK transport
    void transport_stop() const;  ///< stop  JACK transport
    /// get JACK transport state
    std::pair<bool, jack_nframes_t> get_transport_state() const;
    /// set JACK transport location
    bool transport_locate(jack_nframes_t frame) const;

    bool set_freewheel(int onoff) const; ///< switch JACK to freewheeling mode.

  protected:
    /// JACK process callback function.
    virtual int jack_process(const jack_nframes_t nframes);
    virtual int jack_shutdown();
    /// callback function which is called when JACK sample rate changes
    virtual int jack_sample_rate_callback(jack_nframes_t sample_rate);
    /// callback function which is called when JACK buffer size changes
    virtual int jack_buffer_size_callback(jack_nframes_t buffer_size);
    /// callback function which is called when a new JACK port is registered.
      //    virtual void port_registration_callback(jack_port_id_t port_id, int reg);

    virtual int xrun_callback();

    float get_max_signal_amplitude(sample_t *in, jack_nframes_t nframes) const;

    float get_signal_rms(sample_t *in, jack_nframes_t nframes) const;

    float get_cpu_load();


    // still under development:
    //    void new_connection(const std::string source_port,
    //			const std::string destination_port);

  private:
    std::string    _client_name; ///< Name of JACK client
    jack_client_t* _client;      ///< Pointer to JACK client.
    jack_nframes_t _sample_rate; ///< sample rate of JACK server
    jack_nframes_t _buffer_size; ///< buffer size of JACK server

    /// map holding the jack connections to be established
    //    typedef std::map<std::string, std::string> connections_map_t;
    //    connections_map_t _connections_map;

    static int _jack_process(jack_nframes_t nframes, void *arg);
    static void _jack_shutdown(void* arg);
    static int _jack_sample_rate_callback(jack_nframes_t sample_rate,
        void* arg);
    static int _jack_buffer_size_callback(jack_nframes_t buffer_size,
        void* arg);
    //    static void _port_registration_callback(jack_port_id_t port_id,   
    //              int reg, void *arg);
    static int _xrun_callback(void* arg);

    bool _find_port(const std::string& port_name);

  protected:
#ifdef ENABLE_NEWRENDERER
    int get_rt_prio() const { return jack_client_real_time_priority(_client); }
#endif
};

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
