/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * Lock-free first-in-first-out queue.
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef LOCKFREEFIFO_H
#define LOCKFREEFIFO_H

#include "mathtools.h"  // for next_power_of_2()
#include "classtools.h" // for NonCopyable

// most general case is undefined
template<typename T> class LockFreeFIFO;

/** Lock-free first-in-first-out (FIFO) queue. 
 * It is thread-safe for single reader/single writer access.
 * One thread may use push() to en-queue items and another thread may use pop()
 * to de-queue items.
 * @note This FIFO queue is implemented as a ring buffer.
 * @note The idea for this class is stolen from the JACK ringbuffer
 *   implementation: http://jackaudio.org/files/docs/html/ringbuffer_8h.html
 **/
template<typename T>
class LockFreeFIFO<T*> : classtools::NonCopyable
{
  public:
    explicit LockFreeFIFO(size_t size = 128);
    ~LockFreeFIFO();

    bool push(T* item);
    T* pop();

  private:
    volatile size_t _write_index; ///< Write pointer
    volatile size_t _read_index;  ///< Read pointer
    const size_t _size;           ///< Size of the ringbuffer
    const size_t _size_mask;      ///< Bit mask used in modulo operation
    T** const _data;              ///< Actual ringbuffer data
};

/** ctor.
 * @param size desired ring buffer size, gets rounded up to the next power of 2.
 **/
template<typename T>
LockFreeFIFO<T*>::LockFreeFIFO(size_t size)
  : _write_index(0)
  ,  _read_index(0)
  , _size(mathtools::next_power_of_2(size))
  , _size_mask(_size - 1)
  , _data(new T*[_size])
{}

/// dtor.
template<typename T>
LockFreeFIFO<T*>::~LockFreeFIFO() { delete _data; }

/** Add an item to the queue.
 * @param item pointer to an item to be added. 
 * @return @b true on success, @b false if queue is full.
 * @attention You have to check the return value to be sure the item has
 *   actually been added.
 **/
template<typename T>
bool
LockFreeFIFO<T*>::push(T* item)
{
  if (item == 0) return false;

  // Concurrent reading and writing is safe for one reader and writer. Once
  // the _read_index is read the _write_index won't change before reading 
  // it, because it is modified only in this function. This won't work for
  // multiple readers/writers.  
  size_t r = _read_index;
  size_t w = _write_index;

  // Move write pointer by FIFO-size in order to compute the distance to 
  // the read pointer in next step. 
  if (w < r) w += _size;

  // Check if FIFO is full and return false instead of waiting until space is
  // freed. (Prevent read pointer to overtake write pointer.) 
  if (w-r > _size-2) return false;

  _data[w & _size_mask] = item;

  // Set _write_index to next memory location (applying modulo operation)
  _write_index = ++w & _size_mask;
  return true;
}

/** Get an item and remove it from the queue.
 * @return Pointer to the item, @b NULL if queue is empty. 
 **/
template<typename T>
T*
LockFreeFIFO<T*>::pop()
{
  size_t r = _read_index;
  size_t w = _write_index;
  // If queue is empty return 0.
  if (r == w) return NULL;

  T* retval = _data[r];

  // Set _read_index to next memory location (applying modulo operation)
  _read_index = ++r & _size_mask;
  return retval;
}

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
