/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * Main file (implementation).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

/** @mainpage SoundScape Renderer
 *
 * Please send bug reports to SoundScapeRenderer@telekom.de
 *
 * @see http://tu-berlin.de/?id=ssr
 *
 * @date 2006 - 2012
 **/

#ifdef HAVE_CONFIG_H
#include <config.h> // for ENABLE_*, HAVE_*, WITH_*
#endif

#include "configuration.h"
#include "controller.h"
#include "mathtools.h"
#include "xmlparser.h"

/** Unnamed namespace to encapsulate print_about_message() which is only used in
 * main.cpp.
 **/
namespace
{
  /// self-explanatory
  void print_about_message()
  {
    const std::string about_string =
      "       ___     \n"
      "      /  ___   \n"
      "  ___/  /  ___ \n"
      "    ___/  /    " PACKAGE_NAME "\n"
      "         /     \n"
      "               \n"
      " Copyright (c) 2006-2012 Quality & Usability Lab\n"
      "                         Deutsche Telekom Laboratories, TU Berlin\n"
      "                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany\n"
      "\n"
      " Website: "PACKAGE_URL"\n"
      " Contact: "PACKAGE_BUGREPORT"\n"
      "\n"
      " This program comes with ABSOLUTELY NO WARRANTY;"
      " this is free software,\n"
      " and you are welcome to redistribute it under certain conditions;\n"
      " for details, see the enclosed file COPYING.\n";

    std::cout << about_string << std::endl;
  }
}

/** Main function.
 * @param argc number of input arguments
 * @param argv array of actual input arguments
 * @return 0 on success, else some error code (mostly 1)
 **/
int main(int argc, char *argv[])
{
  print_about_message();

  XMLParser::Init();
  ssr::conf_struct conf;

  if (!ssr::parse_options(argc, argv, conf))
  {
    // stop program. --help or --version or something has been called.
    return EXIT_SUCCESS;
    // TODO: check if return value always makes sense
  }

//////// check if specified arguments are actually available ///////////////////

#ifndef ENABLE_IP_INTERFACE
  if (conf.ip_server)
  {
    ERROR(conf.exec_name + " was compiled without IP-server support!\n"
        "Use the --no-ip-server option to disable the IP-server.\n"
        "Type '" + conf.exec_name + " --help' for more information.");
    return EXIT_FAILURE;
  }
#endif

#ifndef ENABLE_GUI
  if (conf.gui)
  {
    ERROR(conf.exec_name + " was compiled without GUI support!\n"
        "Use the --no-gui option to disable GUI.\n"
        "Type '" + conf.exec_name + " --help' for more information.");
    return EXIT_FAILURE;
  }
#endif

//////////////// check for illegal combination of options: /////////////////////

  if (conf.ip_server && conf.freewheeling)
  {
    WARNING("Freewheel mode cannot be used together with IP-server. Ignored.\n"
        "Type '" + conf.exec_name + " --help' for more information.");
    conf.freewheeling = false;
  }

  if (conf.freewheeling && conf.gui)
  {
    WARNING("In 'freewheeling' mode the GUI cannot be used! Disabled.\n"
        "Type '" + conf.exec_name + " --help' for more information.");
    conf.gui = false;
  }

//////// now all options have to be valid! /////////////////////////////////////

  Controller controller(conf.xml_schema, conf.input_port_prefix
      , conf.stand_ampl_ref_dist);

  controller.set_master_volume_correction(
      mathtools::dB2linear(conf.master_volume_correction));

  // temporary solution:
  controller.set_loop_mode(conf.loop);

  if (conf.playback_setup_file_name != "")
  {
    VERBOSE("Trying to load '" << conf.playback_setup_file_name << "' ...");
    if (conf.renderer == Renderer::WFS || conf.renderer == Renderer::AAP
        || conf.renderer == Renderer::VBAP
        || conf.renderer == Renderer::generic)
    {
      if (!controller.load_loudspeaker_setup(conf.playback_setup_file_name))
      {
        ERROR("Couldn't load reproduction setup!");
        return EXIT_FAILURE;
      }
    }
  }
  else
  {
    ERROR("No configuration file for the reproduction setup. "
        "No renderer created!");
    return EXIT_FAILURE;
  }

#ifdef ENABLE_NEWRENDERER
  if (!controller.load_newrenderer(conf))
  {
    ERROR("Couldn't load new renderer!");
    return EXIT_FAILURE;
  }

  controller.set_master_signal_level(0.0f);
#else
  if (conf.renderer == Renderer::WFS)
  {
    if (!controller.load_wfs_renderer(conf.wfs_prefilter_file_name,
          conf.wfs_delayline_size, conf.wfs_initial_delay))
    {
      ERROR("Couldn't load WFS renderer!");
      return EXIT_FAILURE;
    }
  }
  else if (conf.renderer == Renderer::binaural)
  {
    if (!controller.load_binaural_renderer(conf.hrir_size, conf.hrir_file_name))
    {
      ERROR("Couldn't load binaural renderer!");
      return EXIT_FAILURE;
    }
  }
  else if (conf.renderer == Renderer::VBAP)
  {
    if (!controller.load_vbap_renderer())
    {
      ERROR("Couldn't load stereophonic (VBAP) renderer!");
      return EXIT_FAILURE;
    }
  }
  else if (conf.renderer == Renderer::BRS)
  {
    if (!controller.load_brs_renderer(conf.hrir_size))
    {
      ERROR("Couldn't load BRS renderer!");
      return EXIT_FAILURE;
    }
  }
  else if (conf.renderer == Renderer::BPB)
  {
    if (!controller.load_bpb_renderer())
    {
      ERROR("Couldn't load BPB renderer!");
      return EXIT_FAILURE;
    }
  }
  else if (conf.renderer == Renderer::AAP)
  {
    if (!controller.load_apa_renderer(conf.ambisonics_order
          , conf.in_phase_rendering))
    {
      ERROR("Couldn't load Ambisonics (AAP) renderer!");
      return EXIT_FAILURE;
    }
  }
  else if (conf.renderer == Renderer::generic)
  {
    if (!controller.load_generic_renderer())
    {
      ERROR("Couldn't load generic renderer!");
      return EXIT_FAILURE;
    }
  }
  else
  {
    ERROR("Unknown renderer type!");
    return EXIT_FAILURE;
  }
#endif // ENABLE_NEWRENDERER

#ifdef ENABLE_ECASOUND
  if (conf.audio_recorder_file_name != "")
  {
    if (!controller.load_audio_recorder(conf.audio_recorder_file_name))
    {
      ERROR("Unable to start audio recorder!");
      return EXIT_FAILURE;
    }
  }
#endif

  if (conf.scene_file_name != "")
  {
    if (!controller.load_scene(conf.scene_file_name)) return EXIT_FAILURE;
  }
  else WARNING("No scene file specified. Opening empty scene ...");

  if (conf.freewheeling)
  {
    if (!controller.set_freewheeling())
    {
      ERROR("Unable to switch to freewheeling mode!");
      return EXIT_FAILURE;
    }
  }
  else // if not freewheeling, output ports can be connected
  {
    if (!controller.connect_renderer(conf.output_port_prefix))
    {
      ERROR("Could not connect renderer to all required JACK output channels!");
    }
  }
#ifdef ENABLE_IP_INTERFACE
  if (conf.ip_server)
  {
    std::cout << "Starting IP Server "
#ifdef ENABLE_SSL
      "(SSL) "
#endif
      "with port " << conf.server_port << std::endl;
    controller.start_ip_server(conf.server_port);
  }
#endif // ENABLE_IP_INTERFACE

  controller.start_tracker(conf.tracker, conf.tracker_ports);

  if (conf.gui)
  {
#ifdef ENABLE_GUI

    // TEMPORARY!!!
    controller.start_processing();
    controller.transport_locate(0.0f);
    //controller.transport_start();

    // only the remaining arguments (and argv[0]) are forwarded to the GUI, the
    // rest was deleted by parse_options()
    controller.start_gui(argc, argv, 
      conf.path_to_gui_images, conf.path_to_scene_menu);
#else
    // this condition has already been checked above!
    assert(false);
#endif // #ifdef ENABLE_GUI
  }
  else // without GUI
  {
    //int arg = 1;
    //if (arg < argc)
    //{
    //  std::cout << "ignored ARGV-elements:";
    //  while (arg < argc)
    //  {
    //    std::cout << " " << argv[arg++];
    //  }
    //  std::cout << std::endl;
    //}

    // TODO: check if IP-server is running
    // TODO: wait for shutdown command (run forever)

    // TODO: if no IP-server: start scene player

    // TEMPORARY!!!
    controller.start_processing();
    controller.transport_locate(0.0f);
    controller.transport_start();
    bool keep_running = true;
    while (keep_running)
    {
      switch(fgetc(stdin))
      {
        case 'c':
          std::cout << "Calibrating client.\n";
          controller.calibrate_client();
          break;
        case 'p':
          controller.transport_start();
          break;
        case 'q':
          keep_running = false;
          break;
        case 'r':
          controller.transport_locate(0.0f);
          break;
        case 's':
          controller.transport_stop();
          break;
      }
      //sleep(10); // run forever
    }
  }

  // when GUI is closed, it returns HERE!

  if (!controller.save_scene_as_XML("ssr_scene_autosave.asd"))
  {
    ERROR("Couldn't write XML scene! (It's an ugly hack anyway ...");
  }

  // TODO: do some cleanup and exit

  //sleep(1);

  //recorder->disable();
  controller.stop_processing();

  // TODO: check if transport needs to be stopped
  controller.transport_stop();
}

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
