/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * new binaural renderer class (implementation).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include <cassert>       // for assert()
#include <cmath>         // for cos()
#include "newbinauralrenderer.h"
#include "controller.h"  // for set_active_sources(), ...
#include "ssr_global.h"
#include "maptools.h"    // for purge()
#include "mathtools.h"   // for PI
#include "ptrtools.h"    // for is_null()
#include "stringtools.h" // for A2S()
#include "audiofileiotools.h"

using stringtools::A2S;
using namespace ssr; // for id_t, processing, ready, ...

/** ctor.
 * -# get number of loudspeakers from the scene and create outputs.
 * -# load WFS pre-filter
 * -# resize the silence buffer
 * -# and, finally, set the Renderer::_client_state to "ready"
 * @param scene see create()
 * @param controller see create()
 * @param delayline_size see create()
 * @param initial_delay see create()
 **/
NewBinauralRenderer::NewBinauralRenderer(const Scene& scene
    , Controller& controller, const parameter_map &params)
  : NewRenderer(scene, controller, params) // base class ctor.
  , _hrir_size(params.get<int>("hrir_size"))
  , _hrir_filename(params["hrir_filename"])
{
  if (!_make_outputs())
  {
    throw(renderer_error("Output objects could not be created!"));
  }

  if (!_load_hrtfs(_hrtfs, _hrir_size, _hrir_filename))
  {
    throw(renderer_error("Failure loading HRIRs!"));
  }
}

bool
NewBinauralRenderer::_load_hrtfs(hrtfs_t& hrtfs, const unsigned int hrir_size
    , const std::string& hrir_file_name)
{
  std::cout << "Loading HRIRs. Please be patient.\n";

  // get sample rate of hrirs
  const int sr = audiofileiotools::get_sample_rate(hrir_file_name);
  if (!sr)
  {
    return false;
  }
  // if hrirs do not have the correct sample rate
  else if (sr != static_cast<int>(this->get_sample_rate()))
  {
    throw(renderer_error("The head-related impulse responses in file '"
          + hrir_file_name + "' are sampled at rate " + A2S(sr)
          + " Hz. JACK is currently running at "
          + A2S(this->get_sample_rate()) + " Hz."));
  }

  // check number of impulse responses
  const int no_of_ch = audiofileiotools::get_no_of_channels(hrir_file_name);

  // error message comes from function get_no_of_channels()
  if (!no_of_ch) return false;

  if (no_of_ch != 720)
  {
    ERROR("The file '" << hrir_file_name << "' does not contain "
        "the required number of channels (" << no_of_ch << " vs. 720).");
    return false;
  }

  Convolver::data_t container;

  if (!audiofileiotools::load_filter_t(hrir_file_name, container))
  {
    return false;
  }

  const unsigned int hrir_size_in_file = container.size()/720;
  unsigned int effective_hrir_size = hrir_size_in_file;

  // check if user has specified something
  if (hrir_size && hrir_size < hrir_size_in_file)
  {
    WARNING("I'm loading only " << hrir_size << " of " << hrir_size_in_file
        << " HRIR coefficients from file '" << hrir_file_name << "'.");
    effective_hrir_size = hrir_size;
  }
  else if (hrir_size && hrir_size > hrir_size_in_file)
  {
    WARNING("File '" << hrir_file_name << "' contains only "
        << hrir_size_in_file
        << " HRIR coefficients. The user has specified to load "
        << hrir_size << ".");
  }

  // fill HRTF container
  for (int n = 0; n < 360; n++)
  {
    Convolver::data_t dummy;

    // initialize hrtfs.first[n] and hrtfs.second[n]
    hrtfs.first.push_back(dummy);
    hrtfs.second.push_back(dummy);

    // transfer hrirs to frequency domain and store them in vector hrtfs
    Convolver::prepare_impulse_response(hrtfs.first[n],
                            container.data() + 2*n*hrir_size_in_file,
                            effective_hrir_size, this->get_buffer_size());
    Convolver::prepare_impulse_response(hrtfs.second[n],
                            container.data() + (2*n+1)*hrir_size_in_file,
                            effective_hrir_size, this->get_buffer_size());
  }

  VERBOSE("HRIRs of size " << effective_hrir_size << " loaded.");

  return true;
}

NewBinauralRenderer::~NewBinauralRenderer()
{
}

// TODO: These are boilerplate, i could create them
//       with templates in the renderer.
//
//       not 100% sure though. the NewRenderer Class must not
//       be a template. so i guess, i need a template Class deriving from it.

NewRenderer::Input *
NewBinauralRenderer::create_input(const parameter_map &params)
{
  BinInput* dumbinput = new BinInput(*this, params);
  return (Input *) dumbinput;
}

NewRenderer::Output *
NewBinauralRenderer::create_output(const parameter_map &params)
{
  BinOutput* dumboutput = new BinOutput(*this, params);
  return (Output *) dumboutput;
}

NewRenderer::RenderSource *
NewBinauralRenderer::create_source(NewRenderer::Input *input
    , const parameter_map &params)
{
  BinSource* dumbsource = new BinSource(*this, input, params);
  return (RenderSource *) dumbsource;
}

void
NewBinauralRenderer::render(RenderSource *s, Output *o)
{
  //BinInput  *input  = dynamic_cast<BinInput *>(s->get_input());
  BinSource *source = dynamic_cast<BinSource *>(s);
  BinOutput *output = dynamic_cast<BinOutput *>(o);

  sample_t* out;
  out = output->get_sample_destination();

  float weighting_factor = 1.0f;
  float interp_factor = 0.0f;
  float ampl_ref_dist = _scene.get_amplitude_reference_distance();

  // no distance attenuation for plane waves 
  if (source->src.model == Source::plane)
  {
    weighting_factor *= 0.5f / ampl_ref_dist; // 1/r
    // weighting_factor *= 0.25f / sqrt(ampl_ref_dist); // 1/sqrt(r)
  }
  else
  {
    // consider distance attenuation
    // TODO: we need the reference position. gonna use zero for now.
    //float source_distance = (source->src.position - this->position).length();

    float source_distance = (source->src.position).length();

    if (source_distance < 0.5f) interp_factor = 1.0f - 2*source_distance;

    // no volume increase for sources closer than 0.5m
    source_distance = std::max(source_distance, 0.5f);

    weighting_factor *= 0.5f / source_distance; // 1/r
    // weighting_factor *= 0.25f / sqrt(source_distance); // 1/sqrt(r)
  }

  // apply the gain factor of the current source
  weighting_factor *= source->src.gain;

  // apply master volume
  weighting_factor *= _scene.get_master_volume();

  // TODO: Save computation power!
  if (source->src.mute == true)
  {
    weighting_factor = 0.0f;
    //continue; // next source
  }

  assert(weighting_factor >= 0.0f);

  //current_input->fetch_new_buffer(nframes);
  //current_input->weight_audio_data(weighting_factor);

  // 

  // calculate relative position of sound source
  // TODO: again... position == 0
  //int hrtf_index = static_cast<int>(mathtools::rad2deg(
  //	    atan2(i->position.y - this->position.y, 
  //		  i->position.x - this->position.x))
  // 				      - 0.5f)%360;
  int hrtf_index = static_cast<int>(mathtools::rad2deg(
        atan2(source->src.position.y, source->src.position.x)) - 0.5f - 90)%360;

  // make sure that index is in proper range
  if (hrtf_index < 0) hrtf_index += 360;

  sample_t *audio_sample_in;
  if (output->id == 1)
  {
    audio_sample_in = source->apply_hrtf_and_mix_into_output(
        _hrtfs.first[hrtf_index], interp_factor, output, weighting_factor);
  }
  else
  {
    audio_sample_in = source->apply_hrtf_and_mix_into_output(
        _hrtfs.second[hrtf_index], interp_factor, output, weighting_factor);
  }

  sample_t* audio_sample_out = output->get_sample_destination();

  if (output->accumulate)
  {
    // add audio data to output buffers
    for (unsigned int n = 0; n<_nframes; n++)
    {
      audio_sample_out[n]  += audio_sample_in[n];
    }
  }
  else
  {
    for (unsigned int n = 0; n<_nframes; n++)
    {
      audio_sample_out[n] = audio_sample_in[n];
    }
    output->accumulate = true;
  }
}

/** _.
 * This will only be called once. Loudspeakers can @b not be added later.
 * %Loudspeaker-IDs always range from 1 to @a number.
 * @param number number of loudspeakers
 * @return @b true on success.
 **/
bool
NewBinauralRenderer::_make_outputs()
{
  parameter_map params_l;
  params_l["id"] = A2S(1);
  params_l["port_name"] = "out_L";

  addOutput(params_l);

  parameter_map params_r;
  params_r["id"] = A2S(2);
  params_r["port_name"] = "out_R";

  addOutput(params_r);

  VERBOSE("WFSRenderer: 2 Output objects registered.");
  return true;
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class WFSRenderer::Output
////////////////////////////////////////////////////////////////////////////////

/** ctor.
 * create a JACK output port and get its full name.
 * @param renderer calling WFSRenderer object
 * @param port_name the desired output port name (the part after the ":")
 *
 * The full port name including the client name can be obtained from the
 * port_name member variable.
 * @throw output_error
 **/
NewBinauralRenderer::BinOutput::BinOutput(NewRenderer& renderer
    , const parameter_map &params)
  : NewRenderer::Output(renderer, params)
{}

NewBinauralRenderer::BinOutput::~BinOutput()
{}

void
NewBinauralRenderer::BinOutput::cycle_start()
{
  NewRenderer::Output::cycle_start();
  accumulate = false;
}

void
NewBinauralRenderer::BinOutput::cycle_end()
{
  if (accumulate == false) write_empty_buffer(_renderer._nframes);

  NewRenderer::Output::cycle_end();
}

void
NewBinauralRenderer::BinOutput::write_empty_buffer(const jack_nframes_t nframes)
{
    sample_t *out;
    out = get_sample_destination();
    for (size_t i=0; i<nframes; i++) out[i]=0.0f;
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class WFSRenderer::Input
////////////////////////////////////////////////////////////////////////////////

/**
 * @param renderer reference to the calling WFSRenderer object. Using this, we
 * can access the public members of the WFS renderer.
 **/
NewBinauralRenderer::BinInput::BinInput(NewRenderer& renderer
    , const parameter_map &params)
  : NewRenderer::Input(renderer, params)
{
  //NewBinauralRenderer &binrenderer = static_cast<NewBinauralRenderer &>(renderer);
}

NewBinauralRenderer::BinInput::~BinInput()
{}

void
NewBinauralRenderer::BinInput::cycle_start()
{
  NewRenderer::Input::cycle_start();
}

NewBinauralRenderer::BinSource::BinSource(NewRenderer &renderer, Input *input
    , const parameter_map &params)
  : NewRenderer::RenderSource(renderer, input, params)
{
  int i;

  for (i=0; i<2; i++)
  {
    _convolver[i] = Convolver::create(_renderer.get_buffer_size());

    if (ptrtools::is_null(_convolver[i]))
    {
      WARNING("Initialization of convolver failed!");
    }

    _convolver_dir[i] = Convolver::create(_renderer.get_buffer_size());

    if (ptrtools::is_null(_convolver_dir[i]))
    {
      WARNING("Initialization of convolver failed!");
    }
  }

  // calculate transfer function of a dirac
  Convolver::data_t dirac_buffer(_renderer.get_buffer_size(), 0.0f);
  dirac_buffer[0] = 1.0f;
  //Convolver::data_t dirac_container;

  //Convolver::prepare_impulse_response(dirac_container, dirac_buffer.data(),
  Convolver::prepare_impulse_response(_dirac, dirac_buffer.data(),
      dirac_buffer.size(), _renderer.get_buffer_size());

  //_dirac = *dirac_container.begin();

  /*  WARNING: not ported... vvv 
   *
  ////////////// prepare directivity //////////////////////////
  std::string directivity_file = "/home/j.ahrens/audio_files/impulse.wav";//"dipole_test.wav";

  // get sample rate of directivity coefficients
  const int sr = audiofileiotools::get_sample_rate(directivity_file);
  
  // if file could not be opened
  if (!sr) return false; 
  
  // if hrirs do not have the correct sample rate
  else if (sr != static_cast<int>(_renderer.get_sample_rate()))
  {
    ERROR("The coefficients of the directivity expansion in file '" << directivity_file
          << "' are sampled at rate " << sr 
          << " Hz. JACK is currently running at " 
          << _renderer.get_sample_rate() << " Hz.");
    
    // return false;
  }
  
  
  _directivity = (audiofileiotools::load_directivity_f(directivity_file, _renderer.get_buffer_size())).release();

  */

  //    if (ptrtools::is_null(*_directivity)) return false;
}

void
NewBinauralRenderer::BinSource::cycle_start()
{
  NewRenderer::RenderSource::cycle_start();
}

void
NewBinauralRenderer::BinSource::cycle_end()
{
  NewRenderer::RenderSource::cycle_end();
}

float*
NewBinauralRenderer::BinSource::apply_hrtf_and_mix_into_output(
    Convolver::data_t& hrtf, const float interp_factor
    , NewBinauralRenderer::BinOutput *output, float weighting_factor)
{
  // if hrtf has to be interpolated with the dirac
  if (interp_factor > 0.0f)
  {
    Convolver::data_t interp_filter = hrtf;

    const unsigned int partition_size = 2u*_renderer.get_buffer_size();
    const unsigned int overall_filter_size = interp_filter.size();

    float* filter_coefficient = interp_filter.data();
    float* dirac_coefficient = _dirac.data();

    // first partition, loop over all samples
    for (unsigned int sample = 0u; sample < partition_size; sample++)
    {
      // interpolate linearly between dirac and hrtfs
      (*filter_coefficient) = (*filter_coefficient) * (1.0f-interp_factor)
        + interp_factor * (*dirac_coefficient);

      filter_coefficient++; dirac_coefficient++;
    } // for

    // loop over all other samples
    for (unsigned int sample = partition_size
        ; sample < overall_filter_size
        ; sample++)
    {
      // interpolate linearly between dirac and hrtfs (Dirac is zero here???)
      (*filter_coefficient) *= (1.0f-interp_factor);

       filter_coefficient++;
    } // for

    _convolver[output->id-1]->set_filter_f(interp_filter);

  } // if
  else _convolver[output->id-1]->set_filter_f(hrtf);

  /*
  Convolver::filter_t tmp;
  float angle = 270.0f - ((*_scene.get_source_position(_source_id))+Position(-0.085f,0.0f)).orientation().azimuth - (*_scene.get_source_orientation(_source_id)).azimuth;
  _directivity->get_spectrum_of_plane_wave(tmp, angle);
  _convolver_dir_left->set_filter_f(tmp);
  */

  //  return _convolver_dir_left->convolve_signal(&_audio_buffer[0]); 
  // return _convolver_left_ear->convolve_signal(_convolver_dir_left->convolve_signal(&_audio_buffer[0])); 
  return _convolver[output->id-1]->convolve_signal(_input->get_sample_data()
      , weighting_factor);
}

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
