/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * new binaural renderer class (definition).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef NEWBINAURALRENDERER_H
#define NEWBINAURALRENDERER_H

#include <map>
#include <set>
#include <vector>

#include "renderer.h"   // base class of NewBinauralRenderer
#include "convolver.h"
#include "directionalpoint.h" // Position and Orientation
#include "scene.h"

class Controller; // forward declaration

using ssr::id_t;

/** Binaural renderer class.
 * This class does the binaural rendering. Audio processing is done with the
 * inherited JACK functionality.
 **/
class NewBinauralRenderer : public NewRenderer
{
  public:
    NewBinauralRenderer(const Scene& scene, Controller& controller
        , const parameter_map &params);
    virtual ~NewBinauralRenderer();

  protected:
    class BinOutput; // nested class, defined later
    class BinInput;  // nested class, defined later
    class BinSource; // nested class, defined later

    virtual Input *create_input(const parameter_map &params);
    virtual Output *create_output(const parameter_map &params);
    virtual RenderSource *create_source(Input *input
        , const parameter_map &params);

    virtual void render(RenderSource *source, Output *output);

  private:
    /// HRTF set for one ear
    typedef std::vector<Convolver::data_t> hrtf_set_t;
    /// HRTFs for both ears 
    //typedef std::vector< hrtf_set_t > hrtfs_t;
    typedef std::pair<hrtf_set_t, hrtf_set_t> hrtfs_t;

    hrtfs_t _hrtfs;

    unsigned int _hrir_size;
    std::string _hrir_filename;

    bool _make_outputs(); ///< initialize Output objects

    bool _load_hrtfs(hrtfs_t& hrtfs, const unsigned int hrir_size
        , const std::string& hrir_file_name);

};

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Output
////////////////////////////////////////////////////////////////////////////////

/** Manages an output buffer and the corresponding JACK calls.
 * @warning z coordinates and elevation are ignored
 **/
class NewBinauralRenderer::BinOutput : public NewRenderer::Output
{
  public:
    BinOutput(NewRenderer &renderer, const parameter_map &params);
    virtual ~BinOutput();
    virtual void cycle_start();
    virtual void cycle_end();


    float get_signal_level(const jack_nframes_t nframes);

    /// clear output buffer
    void write_empty_buffer(jack_nframes_t nframes);
    bool accumulate;

  private:
    //int id; // left = 0; right = 1;
};

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Input
////////////////////////////////////////////////////////////////////////////////

/** Manages an input buffer (including delay line) and the corresponding JACK
 * calls.
 **/
class NewBinauralRenderer::BinInput : public NewRenderer::Input
{
  public:
    BinInput(NewRenderer& renderer, const parameter_map &params);
    virtual ~BinInput();

    virtual void cycle_start();
};


class NewBinauralRenderer::BinSource : public NewRenderer::RenderSource
{
  public:
    BinSource(NewRenderer &renderer, Input *input, const parameter_map &params);
    virtual void cycle_start();
    virtual void cycle_end();

    float* apply_hrtf_and_mix_into_output(
        Convolver::data_t& hrtf, const float interp_factor
        , NewBinauralRenderer::BinOutput *output, float weighting_factor);

  protected:
    Convolver::ptr_t _convolver[2];

    Convolver::ptr_t _convolver_dir[2];

    Convolver::data_t _dirac; ///< in frequency domain

};
#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
