/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * Newrenderer (implementation). 
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include "renderer.h"
#include "dumbrenderer.h"
#include "newvbaprenderer.h"
#include "newwfsrenderer.h"
#include "newbinauralrenderer.h"
#include "stringtools.h"
#include "controller.h"
#include "ptrtools.h" // for is_null()

#include <pthread.h>
#include <semaphore.h>

#include <jack/thread.h>

#ifndef _REENTRANT
#error You need to compile with _REENTRANT defined since this uses threads!
#endif

//-----------------------------------------------------------------------------
// NewRenderer Implementation 
//-----------------------------------------------------------------------------
/** ctor.
 * @param scene Reference to a Scene
 * @param controller Reference to Controller
 * @param params_ Reference to parameter_map 
 * @see create()
 * @throws jack_error, renderer_error
 **/
NewRenderer::NewRenderer(const Scene& scene, Controller& controller,
    const parameter_map &params_) throw (jack_error, renderer_error)
  : JackClient(params_["jackname"], true)
  , params(params_)
  , _engine()
  , _exiting(0)
  , _num_threads(params_.get("threads", 1))
  , _slave_wait_semaphores(_num_threads)
  , _slave_cont_semaphores(_num_threads)
  , _scene(scene)
  , _controller(controller)
  , _input_list(_engine)
  , _source_list(_engine)
  , _output_list()
  , _input_prefix("in")
  , _output_prefix("out")
{
  /// - Initialize semaphore variable for metering thread.  
  if (sem_init(&_meter_semaphore, 0, 0))
  {
    std::cerr << "Cant obtain metering semaphore. (this should never happen)"
      << std::endl;
    exit(20);
  }

  /// - Start metering in own thread. 
  if (pthread_create(&_meter_threadid, 0, NewRenderer::meteringThreadAux, this))
  {
    std::cerr << "Cant create metering thread. (this should never happen)"
      << std::endl;
    exit(20);
  }

  // XXX: creating Outputs is bad, because Binaural is ignoring speakers.
  //      bah. need to think some more.
  //
  //      perhaps i can supply a default function, for all speaker based,
  //      renderers, and call that from the derived constructors.
  //
  //      goal is still, to hide containers from the Renderer implementation.

  /// - Create slave audio threads.
  create_slave_audio_threads(_num_threads);

  /// - Publish sample rate.
  _controller.publish_sample_rate((unsigned int)this->get_sample_rate());
  //_controller.set_processing_state(ssr::ready, 0); // TODO: lock with mutex?
}

/// dtor. 
NewRenderer::~NewRenderer()
{
  this->deactivate(); // deactivate JACK client (TODO: should we do this later?)
  _controller.set_processing_state(ssr::exiting); // TODO: is this necessary?
  _exiting = 1;
  sem_post(&_meter_semaphore);
  pthread_join(_meter_threadid, NULL);
}

/** Create a NewRenderer.
 * @param scene Reference to a Scene
 * @param controller Reference to Controller
 * @param params Reference to parameter_map 
 * @return Ptr to a NewRenderer which can be of type Dumb, VBAP, WFS or Binaural
 **/
NewRenderer::ptr_t
NewRenderer::create(const Scene& scene, Controller& controller,
    const parameter_map &params)
{
  ptr_t temp; // temp = NULL
  try
  {
    if (params["type"] == "dumb")
    {
      temp.reset(new DumbRenderer(scene, controller, params));
    }
    else if (params["type"] == "vbap")
    {
      temp.reset(new NewVBAPRenderer(scene, controller, params));
    }
    else if (params["type"] == "wfs")
    {
      temp.reset(new NewWFSRenderer(scene, controller, params));
    }
    else if (params["type"] == "binaural")
    {
      temp.reset(new NewBinauralRenderer(scene, controller, params));
    }
    else
    {
      ERROR("Unknown Renderer Type specified!");
    }
  }
  catch(jack_error& e)
  {
    ERROR(e.what());
  }
  catch(renderer_error& e)
  {
    ERROR(e.what());
  }
  if (ptrtools::is_null(temp)) return temp; // if an error happened, return NULL

  if (temp->get_buffer_size() == 0)
  {
    ERROR("Buffersize == 0!");
    ptrtools::destroy(temp);
    return temp; // which is NULL now
  }

  // TODO: lock with mutex?
  temp->_controller.set_processing_state(ssr::ready);

  temp->activate(); // activate JACK client

  return temp;
}

/** Auxiliary function to start slave audio threads.  
 * This is the start_routine for pthread_create(). It must be static so it can 
 * be passed to pthread_create(). From this function code will run in its own
 * thread. 
 *
 * Set the scheduling policy and call slave_audio_thread().
 * @param data Ptr to std::pair<NewRenderer *, int> where int is an ID for the 
 * thread used by this instance of NewRenderer.
 **/
void *
NewRenderer::slave_audio_thread_aux(void *data)
{
  std::pair<NewRenderer *, int> *init_data
    = static_cast<std::pair<NewRenderer *, int> *> (data);
  NewRenderer *r = init_data->first;
  int n = init_data->second;
  delete init_data;

  struct sched_param param;
  param.sched_priority = r->get_rt_prio();

  if (param.sched_priority != -1)
  {
    pthread_setschedparam(pthread_self(), SCHED_FIFO, &param);
  }

  r->slave_audio_thread(n);
  return 0;
}

/** Compute in slave audio thread.
 * This function will run in its thread until destructor is called. It is 
 * responsible for the computation of jobs. At startup it will sem_wait for 
 * jack_process() to start computation. 
 *
 * Before each stage of rendering (cycle_start(), render(), cycle_end()) it 
 * will wait for jack_process to signal the start of computation. After it will 
 * signal back (sem_post) that the job has been finished. 
 *
 * @param thread_number ID of thread 
 * @see jack_process()
 * @see wait_slave_threads()
 * @see cont_slave_threads()
 * @see do_in_this_thread()
 **/
void
NewRenderer::slave_audio_thread(int thread_number)
{
  int n;
  while (1)
  {
    /** - sem_wait for jack_process() to signal start of computation. Cycle
     * start input. 
     **/
    sem_wait(& (_slave_cont_semaphores[thread_number]));
    n=0;
    for (input_list_t::iterator i=_input_list.rt_begin()
        ; i != _input_list.rt_end()
        ; ++i, n++)
    {
      if (do_in_this_thread(n, thread_number))
      {
        (*i)->cycle_start();
      }
    }
    /// - sem_post (signal) to jack_process() termination of computation. 
    sem_post(& (_slave_wait_semaphores[thread_number]));
    /// - sem_wait for jack_process(). Cycle start RenderSource.
    sem_wait(& (_slave_cont_semaphores[thread_number]));

    n=0;
    for (source_list_t::iterator i=_source_list.rt_begin()
        ; i != _source_list.rt_end()
        ; ++i, n++)
    {
      if (do_in_this_thread(n, thread_number))
      {
        (*i)->meter();
        (*i)->cycle_start();
      }
    }

    /// - sem_post (signal) to jack_process() termination of computation. 
    sem_post(& (_slave_wait_semaphores[thread_number]));
    /// - sem_wait for jack_process(). Cycle start Output.
    sem_wait(& (_slave_cont_semaphores[thread_number]));
    // the output_list is not an RT list.
    n=0;
    for (output_list_t::iterator i=_output_list.begin()
        ; i != _output_list.end()
        ; ++i, n++)
    {
      if (do_in_this_thread(n, thread_number))
      {
        (*i)->cycle_start();
      }
    }

    /// - sem_post (signal) to jack_process() termination of computation. 
    sem_post(& (_slave_wait_semaphores[thread_number]));
    /** - sem_wait for jack_process(). Call render() in derived class for all 
     * (Source, Output) pairs.
     **/
    sem_wait(& (_slave_cont_semaphores[thread_number]));
    // the Pointer to the Input, must be a field of the Source for now.
    n=0;
    for (output_list_t::iterator out=_output_list.begin()
        ; out != _output_list.end()
        ; ++out, n++)
    {
      if (do_in_this_thread(n, thread_number))
      {
        for (source_list_t::iterator src=_source_list.rt_begin()
            ; src != _source_list.rt_end()
            ; ++src)
        {
          this->render(*src, *out);
          Output *output = dynamic_cast<Output *>(*out);
          output->meter_master();
        }
      }
    }
    /// - sem_post (signal) to jack_process() termination of computation. 
    sem_post(& (_slave_wait_semaphores[thread_number]));
    /// - sem_wait for jack_process(). Cycle end Input.
    sem_wait(& (_slave_cont_semaphores[thread_number]));
    // Call cycle_end for all ports.
    n=0;
    for (input_list_t::iterator i=_input_list.rt_begin()
        ; i != _input_list.rt_end()
        ; ++i, n++)
    {
      if (do_in_this_thread(n, thread_number))
      {
        (*i)->cycle_end();
      }
    }

    /// - sem_post (signal) to jack_process() termination of computation. 
    sem_post(& (_slave_wait_semaphores[thread_number]));
    /// - sem_wait for jack_process(). Cycle end RenderSource.
    sem_wait(& (_slave_cont_semaphores[thread_number]));
    n=0;
    for (source_list_t::iterator i=_source_list.rt_begin()
        ; i != _source_list.rt_end()
        ; ++i, n++)
    {
      if (do_in_this_thread(n, thread_number))
      {
        (*i)->cycle_end();
      }
    }

    /// - sem_post (signal) to jack_process() termination of computation. 
    sem_post(& (_slave_wait_semaphores[thread_number]));
    /// - sem_wait for jack_process(). Cycle end Output.
    sem_wait(& (_slave_cont_semaphores[thread_number]));
    // the output_list is not an RT list.
    n=0;
    for (output_list_t::iterator i=_output_list.begin()
        ; i != _output_list.end()
        ; ++i, n++)
    {
      if (do_in_this_thread(n, thread_number))
      {
        (*i)->cycle_end();
      }
    }

    /// - sem_post (signal) to jack_process() termination of computation. 
    sem_post(& (_slave_wait_semaphores[thread_number]));
  }
}


/** Create slave audio threads.
 * Create a specified number of slave threads. At creation threads will wait for
 * a sem_post to start computation.  
 * @param number slave count
 **/
void
NewRenderer::create_slave_audio_threads(int number)
{
  int i;
  std::vector<pthread_t> thread_ids;

  /// - Initialise the synchronization semaphores and set its value to 0.
  for (i=0; i<number; i++)
  {
    if (sem_init(&(_slave_wait_semaphores[i]), 0, 0))
    {
      std::cerr
        << "Cant obtain audio slave semaphore. (this should never happen)"
        << std::endl;
      exit(20);
    }
    if (sem_init(& (_slave_cont_semaphores[i]), 0, 0))
    {
      std::cerr
        << "Cant obtain audio slave semaphore. (this should never happen)"
        << std::endl;
      exit(20);
    }
  }

  /** - Start the slave threads.
   * Thread number 0 is reserved to jack_process().
   **/
  for (i=1; i<number; i++)
  {
    pthread_t tid;

    // this gets freed from the proxy.
    std::pair<NewRenderer *, int> *init_data
      = new std::pair<NewRenderer *,int>(this, i);

    if (pthread_create(&tid, 0, NewRenderer::slave_audio_thread_aux, init_data))
    {
      std::cerr << "Cant create slave audio thread. (this should never happen)"
        << std::endl;
      exit(20);
    }
    _slave_thread_ids.push_back(tid);
  }

}

/** Map jobs to available threads.  
 * A job is an entity that requires to know if it can be executed in this thread
 * , e.g. execution of render() for an Input.  
 *
 * @see slave_audio_thread
 *
 * @param current_job number of job 
 * @param thread_number which thread; thread ID 
 * @return if current job should be computed in this thread. 
 **/
bool
NewRenderer::do_in_this_thread(int current_job, int thread_number)
{
  return thread_number == current_job % _num_threads;
}

/// Signal to all slave threads to wait. Called by jack_process(). 
void
NewRenderer::wait_slave_threads()
{
  int i;
  for (i=1; i<_num_threads; i++)
  {
    sem_wait(& (_slave_wait_semaphores[i]));
  }
}

/// Signal to all slave threads to continue. Called by jack_process(). 
void
NewRenderer::cont_slave_threads()
{
  int i;
  for (i=1; i<_num_threads; i++)
  {
    sem_post(& (_slave_cont_semaphores[i]));
  }
}

/** JACK callback and trigger of the SSR. 
 * Is called by JACK in each audio cycle. This master audio thread triggers all
 * slave audio threads and is threated as realtime thread.  
 *
 * Renderers should not be concerned about JACK, so the JACK callback is 
 * private.
 * @param nframes JACK sample frame count number 
 * @note NewRenderer::_output_list is not an RT list.
 * @note Thread number 0 is reserved for this function. 
 * @see do_in_this_thread()
 * @see process_rt_commands()
 * @return 0 if ok, 1 if this function should not be called any more. 
 **/
int
NewRenderer::jack_process(jack_nframes_t nframes)
{
  int n;
  this->_nframes = nframes;
  _engine.process_rt_commands();

  /// - Start/continue slave audio threads. 
  cont_slave_threads();
  /// - Cycle start Input for thread number 0. 
  n=0;
  for (input_list_t::iterator i=_input_list.rt_begin()
      ; i != _input_list.rt_end()
      ; ++i, n++)
  {
    if (do_in_this_thread(n, 0))
    {
      (*i)->cycle_start();
    }
  }
  /// - Wait for slave audio threads. 
  wait_slave_threads();
  /// - Continue slave audio threads. 
  cont_slave_threads();
  /// - Cycle start RenderSource for thread number 0. 
  n=0;
  for (source_list_t::iterator i=_source_list.rt_begin()
      ; i != _source_list.rt_end()
      ; ++i, n++)
  {
    if (do_in_this_thread(n, 0))
    {
      (*i)->meter();
      (*i)->cycle_start();
    }
  }
  /// - Wait for slave audio threads. 
  wait_slave_threads();

  /// - Continue slave audio threads. 
  cont_slave_threads();
  /// - Cycle start Output for thread number 0.
  n=0;
  for (output_list_t::iterator i=_output_list.begin()
      ; i != _output_list.end()
      ; ++i, n++)
  {
    if (do_in_this_thread(n, 0))
    {
      (*i)->cycle_start();
    }
  }
  /// - Wait for slave audio threads. 
  wait_slave_threads();

  /// - Continue slave audio threads. 
  cont_slave_threads();
  /// - Call render() for all (Source, Output) pairs for thread number 0.
  // the Pointer to the Input, must be a field of the Source for now.
  n=0;
  for (output_list_t::iterator out=_output_list.begin()
      ; out != _output_list.end()
      ; ++out, n++)
  {
    if (do_in_this_thread(n, 0))
    {
      for (source_list_t::iterator src=_source_list.rt_begin()
          ; src != _source_list.rt_end()
          ; ++src)
      {
        this->render(*src, *out);
        Output *output = dynamic_cast<Output *>(*out);
        output->meter_master();
      }
    }
  }
  /// - Wait for slave audio threads. 
  wait_slave_threads();

  /// - Continue slave audio threads. 
  cont_slave_threads();
  /// - Cycle end Input for thread number 0. 
  n=0;
  for (input_list_t::iterator i=_input_list.rt_begin()
      ; i != _input_list.rt_end()
      ; ++i, n++)
  {
    if (do_in_this_thread(n, 0))
    {
      (*i)->cycle_end();
    }
  }
  /// - Wait for slave audio threads. 
  wait_slave_threads();

  /// - Continue slave audio threads. 
  cont_slave_threads();
  /// - Cycle end for thread number 0.
  n=0;
  for (source_list_t::iterator i=_source_list.rt_begin()
      ; i != _source_list.rt_end()
      ; ++i, n++)
  {
    if (do_in_this_thread(n, 0))
    {
      (*i)->cycle_end();
    }
  }
  /// - Wait for slave audio threads. 
  wait_slave_threads();

  /// - Continue slave audio threads. 
  cont_slave_threads();
  /// - Cycle end Output for thread number 0.  
  // the output_list is not an RT list.
  n=0;
  for (output_list_t::iterator i=_output_list.begin()
      ; i != _output_list.end()
      ; ++i, n++)
  {
    if (do_in_this_thread(n, 0))
    {
      (*i)->cycle_end();
    }
  }
  /// - Wait for slave audio threads. 
  wait_slave_threads();

  /// - Increment the metering semaphore to make the meterThread publish values.
  sem_post(&_meter_semaphore);

  return 0;
}

/** Add a pair of (Input Source) thread safe.
 * - Create Input and add to NewRenderer::_input_list.
 * - Create Source and add to NewRenderer::_source_list.
 * - Add Source to NewRenderer::_shadow_sources.
 * @param params 
 **/
void
NewRenderer::addSourceInput(const parameter_map &params)
{
  ScopedLock lock(this);

  Input *in = create_input(params);
  RenderSource *src = create_source(in, params);
  int id;

  // The order of adds is important here.
  // it might get split, when process_cb is called
  // in between the adds.
  _input_list.add(in);
  _source_list.add(src);

  stringtools::S2A(params["id"], id);
  _shadow_sources[id] = src->src;
}

/** Remove (Source Input) pair thread safe.
 * @param id
 **/
void
NewRenderer::remSourceInput(int id)
{
  ScopedLock lock(this);

  source_list_t::iterator i;
  for (i=_source_list.begin(); i!=_source_list.end(); i++)
  {
    if ((*i)->id == id) break;
  }

  // check if we did not find the right source.
  if (i==_source_list.end())
  {
    std::cout << "not found id " << id << std::endl;
    return;
  }

  // TODO: this looks like race.
  //       i think i need to delete the source first.
  _shadow_sources.erase(id);

  Input *input = (*i)->get_input();
  _source_list.rem((*i));
  _input_list.rem(input);
}

/// Clear all inputs thread safe. 
void
NewRenderer::clearAllSourceInput()
{
  ScopedLock lock(this);

  _shadow_sources.clear();
  _source_list.clear();
  _input_list.clear();
}

/** Get copy of Source from _shadow_sources thread safe.
 * @param id
 * @return Source
 **/
Source
NewRenderer::getSourceCopy(int id)
{
  // TODO: get read-lock?
  ScopedLock lock(this);

  return _shadow_sources[id];
}

/** Applies changes of Source thread safe.
 * e.g. position change.
 * @param id of Source to be changes
 * @param s Source object carrying new changes 
 **/
void
NewRenderer::pushSourceChange(int id, Source s)
{
  ScopedLock lock(this);

  _shadow_sources[id] = s;

  // TODO: create a Command which pushes this through RTEngine.
  //       the current code is bad.

  source_list_t::iterator i;
  for (i=_source_list.begin(); i!=_source_list.end(); i++)
  {
    if ((*i)->id == id) break;
  }

  // check if we did not find the right source.
  if (i==_source_list.end())
  {
    // This might happen due to another thread removing the source
    // while the current thread tries to update values.
    return;
  }
  (*i)->src = s;
}

/** Auxiliary function to start meteringThread().
 * This is the start_routine for pthread_create().
 * @param data void Ptr to NewRenderer 
 **/
void *
NewRenderer::meteringThreadAux(void *data)
{
  NewRenderer *r = static_cast<NewRenderer *> (data);
  r->meteringThread();
  return 0;
}

/// Thread for metering 
void NewRenderer::meteringThread()
{
  while (!_exiting)
  {
    sem_wait(&_meter_semaphore);
    publishMeter();
    _controller.set_transport_state(this->get_transport_state());
    _controller.set_cpu_load(this->get_cpu_load());
  }
}

/** Publish Meter.
 * - Get meter level from Input and set it in Controller.
 * - Check active Sources and set it in Controller.
 **/
void
NewRenderer::publishMeter()
{
  // TODO: get read-lock?
  ScopedLock lock(this);

  for (source_list_t::iterator i=_source_list.begin()
      ; i != _source_list.end()
      ; ++i)
  {
    _controller.set_source_signal_level((*i)->id
        , (*i)->get_meter_level());
  }

  for (output_list_t::iterator o=_output_list.begin()
      ; o != _output_list.end()
      ; ++o)
  {
    std::set<id_t> source_set;
    for (source_list_t::iterator s=_source_list.begin()
        ; s != _source_list.end()
        ; ++s)
    {
      if ((*s)->output_active((*o)->id))
      {
        source_set.insert((*s)->id);
      }
    }
    _controller.set_active_sources((*o)->id, source_set);
    _controller.set_master_signal_level((*o)->get_meter_level());
  }
}

/** Meter samples.
 * Return the maximum amplitude of provided sampels. 
 * @param data Samples to be metered. 
 **/
float
NewRenderer::get_meter(sample_t* data) const {
  jack_nframes_t nframes = _nframes;
  jack_nframes_t n;
  float maximum = 0.0f;
  //float rms = 0.0;
  for (n=0; n<nframes; n++)
  {
    maximum = std::max(data[n], maximum);
    //rms += data[n] * data[n];
  }
  //rms /= (float) nframes;
  //return rms;
  return maximum;
}
/** Create an Output thread safe.
 * The Output obj is added to _output_list.
 * @param params 
 **/
void
NewRenderer::addOutput(const parameter_map &params)
{
  ScopedLock lock(this);

  Output *out = create_output(params);
  _output_list.push_back(out);
}


/** Create an (Output Loudspeaker) pair thread safe and add it to _output_list. 
 * The Loudspeaker is set on the Output obj which is added to _output_list.
 * @param params parameter_map
 * @param speaker a Loudspeaker  
 **/
void
NewRenderer::addOutput(const parameter_map &params, const Loudspeaker &speaker)
{
  ScopedLock lock(this);

  Output *out = create_output(params);
  out->_speaker = speaker;
  _output_list.push_back(out);
}

/** Pushes port names on port list.  
 * @param port_list Reference to port_list
 **/
void
NewRenderer::get_output_port_names(port_list_t& port_list) const
{
  const output_list_t output_list = this->get_output_list();
  for (output_list_t::const_iterator i = output_list.begin()
      ; i != output_list.end()
      ; ++i)
  {
    port_list.push_back((*i)->port_name);
  }
}

//-----------------------------------------------------------------------------
// Output Implementation
//-----------------------------------------------------------------------------

/** ctor.
 **/
NewRenderer::Output::Output(NewRenderer &renderer, const parameter_map &params)
  : _renderer(renderer)
  , _port(_renderer.register_out_port(params["port_name"]))
  // get actual port name and save it to member variable
  , port_name(_port ? jack_port_name(_port) : "")
{
  stringtools::S2A(params["id"], this->id);
}

/// Get a memory area that can be written to.
float *
NewRenderer::Output::get_sample_destination()
{
  return (float *) jack_port_get_buffer(this->_port, this->_renderer._nframes);
}

float NewRenderer::Output::_master_meter_max = 0;

/** Meter the Output.
 * Set the meter for the master output which is the signal level of the output
 * channel with highest amplitude/level. 
 **/
void
NewRenderer::Output::meter_master()
{
  sample_t* out = this->get_sample_destination();
  float channel_meter = _renderer.get_meter(out);
  if(_channel_meter_max < channel_meter){
      NewRenderer::Output::_master_meter_max = channel_meter;
  }
}

/// dtor. Unregister JACK port.  
NewRenderer::Output::~Output()
{
  this->_renderer.unregister_port(this->_port); // Calls JackClient::unregister_port.
}

//-----------------------------------------------------------------------------
// Input Implementation
//-----------------------------------------------------------------------------

/** ctor.
 * Save acutal port name and connect optionally to jack_port.  
 * @param renderer Reference to a NewRenderer
 * @param params Reference to a parameter_map
 **/
NewRenderer::Input::Input(NewRenderer &renderer, const parameter_map &params)
  : _renderer(renderer)
  , _port(_renderer.register_in_port(params["port_name"]))
  // get actual port name and save it to member variable
  , port_name(_port ? jack_port_name(_port) : "")
{
  // optionally connect to jack_port
  if (params.has_key("connect_to") && params["connect_to"] != "")
  {
    _renderer.connect_ports(params["connect_to"], this->port_name);
  }
}

/// Get audio from a jack port. 
float *
NewRenderer::Input::get_sample_data() const
{
  return (float *)jack_port_get_buffer(this->_port, this->_renderer._nframes);
}

/// dtor. Unregister JACK port. 
NewRenderer::Input::~Input()
{
  this->_renderer.unregister_port(this->_port); // Calls JackClient::unregister_port.
}

//-----------------------------------------------------------------------------
// RenderSource Implementation
//-----------------------------------------------------------------------------

/** ctor.
 * - Set all outputs of renderer to false in activity_map.
 * - Set id in parameter_map. 
 * @param renderer Reference to NewRenderer
 * @param input Pointer to Input
 * @param params Reference to parameter_map
 **/
NewRenderer::RenderSource::RenderSource(NewRenderer &renderer, Input *input
    , const parameter_map &params)
  : _renderer(renderer)
  , _input(input)
{
  const output_list_t outs = renderer.get_output_list();
  for (output_list_t::const_iterator i=outs.begin(); i != outs.end(); ++i)
  {
    _activity_map[(*i)->id] = false;
  }

  stringtools::S2A(params["id"], id);
  src.mute = false;
}

/** Meter a source.  
 * Set maximum amplitued value or 0 if source is muted.
 * @note: Leveling of RenderSources/Inputs is not applied yet. 
 **/
void
NewRenderer::RenderSource::meter()
{
  if (src.mute)
  {
    _meter_max = 0.0f;
    return;
  }
  _meter_max = _renderer.get_meter(get_input()->get_sample_data());
}

/** Set if Ouput is active. 
 * @param output ID of Output
 * @param active bool
 * @see render()
 **/
void
NewRenderer::RenderSource::set_output_activity(int output, bool active)
{
  _activity_map[output] = active;
}

/** Return if Output is active.
 * @param output ID of Output
 * @return bool
 **/
bool
NewRenderer::RenderSource::output_active(int output)
{
  return _activity_map[output];
}

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
