/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * %VBAPRenderer class (definition).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef VBAPRENDERER_H
#define VBAPRENDERER_H

#include <map>
#include <set>
#include <vector>

#include "renderer.h"         // base class of VBAPRenderer
#include "delayline.h"        // base class of VBAPRenderer::Output
#include "directionalpoint.h" // Position and Orientation

using ssr::id_t;

/** VBAP renderer class.
 * This class does the VBAP rendering. Audio processing is done with the
 * inherited JACK functionality.
 **/
class NewVBAPRenderer : public NewRenderer
{
  public:
    NewVBAPRenderer(const Scene &scene, Controller& controller
        , const parameter_map &params);
    virtual ~NewVBAPRenderer();

  protected:
    // The NewRenderer Interface.
    // lets see how this goes.
    class VBAPOutput; // nested class, defined later
    class VBAPInput;  // nested class, defined later
    class VBAPSource;

    virtual Input *create_input(const parameter_map &params);
    virtual Output *create_output(const parameter_map &params);
    virtual RenderSource *create_source(Input *input
        , const parameter_map &params);

    virtual void render(RenderSource *source, Output *output);

    struct SpeakerWeights
    {
      SpeakerWeights() : left(0.0f), right(0.0f) {}; ///< ctor

      float left;
      float right;
    };

    bool _is_angle_larger(float angle1, float angle2);
    bool _is_valid_loudspeaker_pair(const Loudspeaker& left,
            const Loudspeaker& right);

    void _render_source_on_output(const RenderSource *source
        , Output* current_output, float weighting_factor
        , const jack_nframes_t nframes);

    void _fade_out_source_on_output(const RenderSource* source_id
        , Output* current_output, const jack_nframes_t nframes);

  public:
    void weight_buffer_and_add(const sample_t* buffer_to_be_added
        , const float prev_weight, const float curr_weight
        , sample_t* target_buffer, const jack_nframes_t nframes);

    void clear_buffer(sample_t* const buffer, jack_nframes_t buffer_size) const;

  private:
    /// initialize renderer
    bool _init(const std::string& client_name);

    /// this will be filled with zeroes
    mutable std::vector<sample_t> _silence;

    std::vector<sample_t> _fade_in;
    std::vector<sample_t> _fade_out;

    const std::string  _input_prefix; ///< prefix for the input  ports
    const std::string _output_prefix; ///< prefix for the output ports

    output_list_t _normal_ls_list;
    output_list_t _subwoofers_list;

    /// initialize Output objects
    bool _make_outputs();

    SpeakerWeights _calculate_speaker_weights(const RenderSource *source,
                const Loudspeaker& left,
                const Loudspeaker& right);
}; // end of VBAPRenderer class

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Output
////////////////////////////////////////////////////////////////////////////////

/** Manages an output buffer and the corresponding JACK calls.
 * @warning z coordinates and elevation are ignored
 **/
class NewVBAPRenderer::VBAPOutput : public NewRenderer::Output
                                  , public DelayLine<sample_t>
{
  public:
    VBAPOutput(NewRenderer &renderer, const parameter_map &params);

    virtual ~VBAPOutput(); ///< Unregister associated JACK output port
    virtual void cycle_start();

    struct Coefficient
    {
      explicit Coefficient(float weight = 1.0) : weight(weight) {}
      float weight;
    };

    void  update_previous_source_weights();
    void  store_current_source_weight(const id_t id, const float weight);
    float get_previous_source_weight(id_t source_id);

    bool unrendered_source_left() const;

    id_t return_id_of_remaining_source();

    void write_output_buffer(const jack_nframes_t nframes);

    /// clear output buffer
    void write_empty_buffer(jack_nframes_t nframes);

    void clear_active_source_ids();

    void clear_output_buffer(const jack_nframes_t nframes);

    float get_signal_level(const jack_nframes_t nframes);

    float* get_output_buffer(const jack_nframes_t nframes);
    /// use the assignment operator of the base class.
    //using DirectionalPoint::operator=;

    /// overloaded assignment operator for Loudspeaker
    //Output& operator=(const Loudspeaker& ls);

    /// operator needed to sort loudspeakers according to their positional angle
    static bool sort_operator(const Output* one, const Output* other)
    {
      return one->_speaker.position.orientation().azimuth
         < other->_speaker.position.orientation().azimuth;
    }

    //bool operator<(const Output* other) const
    //{
    //  return this->position.orientation().azimuth
    //      < other->position.orientation().azimuth;
    //}

    /// sources for which current Output is active
    std::set<id_t> active_source_ids;
    // those are public, because I was too lazy to write get_active_source_ids()

  private:
    bool _init(const std::string& port_name);

    //const VBAPRenderer& _renderer; ///< ref. to the calling VBAPRenderer object

    typedef std::map<id_t, Coefficient> coefficient_map_t;

    /// to accumulate the weights of the current audio cycle for all sources
    coefficient_map_t _current_source_weights;

    /// holds the weights of the previous audio cycle for all sources
    coefficient_map_t _previous_source_weights;

    std::vector<sample_t> _audio_buffer;

    bool _delay_line_required;
};

class NewVBAPRenderer::VBAPSource : public NewRenderer::RenderSource
{
  public:
    VBAPSource(NewRenderer &renderer, Input *input
        , const parameter_map &params);

    virtual void cycle_start();

    Output *left_speaker;
    Output *right_speaker;
    float weight_left;
    float weight_right;

  private:
    void find_speaker_pair();
};

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Input
////////////////////////////////////////////////////////////////////////////////

/** Manages an input buffer and the corresponding JACK
 * calls.
 **/
class NewVBAPRenderer::VBAPInput : public NewRenderer::Input
{
  public:
    VBAPInput(NewRenderer &renderer, const parameter_map &params);

    /// Unregisters associated JACK input port
    ~VBAPInput();
    virtual void cycle_start();

    /// gets a buffer full of audio from JACK and writes it into the delay line
    bool fetch_new_buffer(jack_nframes_t nframes);
};

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
