/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * %WFSRenderer class (definition).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef NEWWFSRENDERER_H
#define NEWWFSRENDERER_H

#include <map>
#include <set>
#include <vector>

#include "renderer.h"   // base class of WFSRenderer
#include "delayline.h"  // base class of WFSRenderer::Input
#include "convolver.h"
#include "directionalpoint.h" // Position and Orientation
#include "scene.h"

class Controller; // forward declaration

/** WFS renderer class.
 * This class does the WFS rendering. Audio processing is done with the
 * inherited JACK functionality.
 **/
class NewWFSRenderer : public NewRenderer
{
  public:
    NewWFSRenderer(const Scene& scene, Controller& controller
        , const parameter_map &params);
    virtual ~NewWFSRenderer();

  protected:
    class WFSOutput; // nested class, defined later
    class WFSInput;  // nested class, defined later
    class WFSSource; // nested class, defined later

    /// Type for specifying length and indizes of DelayLines
    typedef DelayLine<sample_t>::size_type delay_size_t;

    const delay_size_t delayline_size; ///< Size of delay lines (in samples)
    const delay_size_t initial_delay;  ///< Initial delay (in samples, >= 0)

    virtual Input *create_input(const parameter_map &params);
    virtual Output *create_output(const parameter_map &params);
    virtual RenderSource *create_source(Input *input
        , const parameter_map &params);

    virtual void render(RenderSource *source, Output *output);

    void clear_buffer(sample_t* const buffer, jack_nframes_t buffer_size) const;

  private:
    const std::string _prefilter_file_name; /// < dto.

    //std::map<ssr::id_t, Position> _previous_source_positions; ///< source positions during the previous frame

    /// initialize Output objects
    bool _make_outputs();

    //void _store_source_positions(SourceCopy::container_t& sources);

    Convolver::data_t _prefilter; ///< container for impulse response
};

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Output
////////////////////////////////////////////////////////////////////////////////

/** Manages an output buffer and the corresponding JACK calls.
 * @warning z coordinates and elevation are ignored
 **/
class NewWFSRenderer::WFSOutput : public NewRenderer::Output
{
  public:
    WFSOutput(NewRenderer &renderer, const parameter_map &params);
    virtual ~WFSOutput();
    virtual void cycle_start();
    virtual void cycle_end();

    /// Render audio data to the corresponding output object.
    //void render(const SourceCopy::container_t& source_list,
    //    const DirectionalPoint& reference, 
    //    const jack_nframes_t nframes,
    //    const float master_volume = 1.0f,
    //    const float ampl_ref_dist = 3.0f);

    float get_signal_level(const jack_nframes_t nframes);

    /// clear output buffer
    void write_empty_buffer(jack_nframes_t nframes);
    bool accumulate;
};

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Input
////////////////////////////////////////////////////////////////////////////////

/** Manages an input buffer (including delay line) and the corresponding JACK
 * calls.
 **/
class NewWFSRenderer::WFSInput : public NewRenderer::Input
                               , public DelayLine<sample_t>
{
  public:
    WFSInput(NewRenderer& renderer, const parameter_map &params);
    virtual ~WFSInput();

    virtual void cycle_start();

  protected:
    Convolver::ptr_t _convolver;

    unsigned int _prefilter_cycles; ///< number of subframes for prefiltering

    bool _prefilter_signal(sample_t *in, jack_nframes_t nsubframes);
};


class NewWFSRenderer::WFSSource : public NewRenderer::RenderSource
{
  public:
    WFSSource(NewRenderer &renderer, Input *input, const parameter_map &params);
    virtual void cycle_start();
    virtual void cycle_end();

    struct Coefficient
    {
      explicit Coefficient(int int_delay = 0, float weighting_factor = 1.0)
        : delay(int_delay)
        , factor(weighting_factor)
      {}
      int   delay;
      float factor;
    };
    std::map<ssr::id_t,Coefficient> last_coefficients;

    bool focused;
};

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
