/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * A "dictionary" for parameters.
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef PARAMETER_MAP_H
#define PARAMETER_MAP_H

#include <map>
#include "stringtools.h"

using stringtools::S2RV;

/** A "dictionary" for parameters.
 * All values are std::string's.
 *
 * Usage examples:
 * @code
 * parameter_map params;
 * params["one"] = "first value";
 * params["two"] = "2";
 * std::string val1;
 * int val2;
 * val1 = params["one"];
 * val2 = params.get<int>("two");
 * if (params.has_key("three"))
 * {
 *   // this is not done because there is no key named "three":
 *   do_something();
 * }
 * @endcode
 **/
struct parameter_map : std::map<std::string,std::string>
{
  typedef std::map<std::string,std::string> map_t; ///< the map type

  /** "Getter".
   * @param s parameter name which should be retrieved.
   * @return a const reference to the value referenced by @p s.
   * @throw std::runtime_error if the key @p s doesn't exist. You should've
   *   checked beforehand with has_key() ...
   **/
  const std::string& operator[](const std::string& s) const
    throw (std::runtime_error)
  {
    map_t::const_iterator it;
    it = this->find(s);
    if (it == this->end())
    {
      // parameter does not exist. perhaps we should fall back
      // to some defaults. lets throw an exception for now.
      throw std::runtime_error("parameter " + s + " does not exist in map");
    }
    return it->second;
  }

  /** "Setter".
   * @param s parameter name which should be set.
   * @return a reference to the value referenced by @p s.
   **/
  std::string& operator[](const std::string& s)
  {
    return this->map_t::operator[](s);
  }

  /** Get value converted to given type.
   * @tparam T the desired type
   * @throw anything that operator[]() throws (and anything T's ctor may throw).
   **/
  template<typename T>
  T get(const std::string& s, const T& def = T()) const
  {
    return S2RV(this->operator[](s), def);
  }

  /** Check if a given parameter is available.
   * @param s the parameter name
   * @return @b true if @p s is available
   **/
  bool has_key(const std::string& s) const
  {
    return this->count(s) > 0;
  }
};

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
