/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * Realtime engine and realtime safe list.
 * 
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef RTCOMMAND_H
#define RTCOMMAND_H

#include "lockfreefifo.h"
#include <unistd.h>  // for usleep()

/** Define the basic (abstract) RTCommand.
 * These commands are passed through the LockFreeFIFO into the RT thread, e.g. 
 * they either change parameters or add/remove sources. 
 *
 * The function execute() implements the behaviour of the RTCommand and will 
 * be called by RTCommandEngine::process_rt_commands(). 
 * @see RTListAddCommand
 * @see RTListRemCommand
 **/
class RTCommand
{
  public:
    /** Execute the command. 
     * The implementation of this function determines what the command will do. 
     * @see RTListAddCommand::execute() for example. 
     **/
    virtual void execute() = 0;
    /** Cleanup the command. 
     * For example free the memory of this object.   
     **/
    virtual void cleanup() = 0;
    /// dtor
    virtual ~RTCommand() {}
};

/** Scheduler for RTCommands.
 * The scheduler is driven and must be started by an RT thread. The behaviour of
 * a RTCommand is implemented in its execute() function. 
 * @see process_rt_commands 
 *
 * It is based on two LockFreeFIFO lists:
 * - RTCommandEngine::in_fifo: Contains RTCommand objects that are not yet 
 *   executed.
 * - RTCommandEngine::out_fifo: Contains RTCommand objects that have been 
 *   executed and must be deleted. 
 * The RTCommands will be executed in process_rt_comands().
 * 
 * Using this FIFO model to schedule the execution of RTCommands realtime  
 * safety is reached. 
 *
 * @see push_command for adding RT commands.
 * @see wait_command to wait for the execution of an RT command.
 * @see RTList which uses this engine as scheduler for adding and removing
 * elements to a list.  
 **/
class RTCommandEngine
{
  private:
    /// Commands to execute
    LockFreeFIFO<RTCommand*> in_fifo;
    /// Executed commands to delete
    LockFreeFIFO<RTCommand*> out_fifo;

  public:
    RTCommandEngine() {}

    /** Push a command through the RT thread.
     * Must be used to add RTCommands that should be executed. 
     * @note The command will be deleted when it comes out of the RT thread 
     * through the RTCommandEngine::out_fifo again.
     * @param cmd A RTCommand to be executed. 
     **/
    void push_command(RTCommand *cmd)
    {
      /** First pop all commands from the RTCommandEngine::out_fifo, call 
       * cleanup and delete the popped commands. This ensures that the out_fifo 
       * is not going to be full which would block 
       * RTCommandEngine::process_rt_commands and its calling realtime thread.  
       **/
      while (1)
      {
        RTCommand *out_cmd = out_fifo.pop();
        if (out_cmd == 0)
          break;

        out_cmd->cleanup();
        delete out_cmd;
      }

      /** Now push the command on RTCommandEngine::in_fifo. 
       * Wait for the item to be pushed if the fifo is full
       **/
      // TODO: add some usleep, when queue is full.
      //       but this is unlikely to happen.
      while (!in_fifo.push(cmd)) ;
    }

    /** Wait until a RTCommand was executed.  
     * So it will pop out of the RTCommandEngine::out_fifo. 
     * Call cleanup for this command and its followers and delete them. 
     * @param cmd A RTCommand to wait for execution.  
     **/
    void wait_command(RTCommand *cmd)
    {
      bool end_loop = false;
      while (!end_loop)
      {
        RTCommand *out_cmd = out_fifo.pop();
        if (out_cmd == 0)
        {
          usleep(50);
          continue;
        }
        if (out_cmd == cmd)
          end_loop = true;

        out_cmd->cleanup();
        delete out_cmd;
      }
    }
    /// Push a RTCommand and wait for its execution. 
    void push_command_and_wait(RTCommand *cmd)
    {
      push_command(cmd);
      wait_command(cmd);
    }

    /** Start RTCommandEngine.  
     * This function must be called from the RT thread (e.g. jack_process()) to 
     * execute the commands. For each call, in JACK context for each audio 
     * cycle, all commands in RTCommandEngine::in_fifo will be executed and 
     * then pushed on the RTCommandEngine::out_fifo.
     * 
     * - Pop the command from RTCommandEngine::in_fifo list.
     * - execute() the command.
     * - Push the command to RTCommandEngine::out_fifo list. 
     *
     * @see RTListAddCommand::execute()
     * @see RTListRemCommand::execute()
     **/
    void process_rt_commands()
    {
      while (1)
      {
        RTCommand *the_cmd = in_fifo.pop();
        if (the_cmd == 0)
          break;

        the_cmd->execute();
        out_fifo.push(the_cmd);
      }
    }
};

/// RTList add command.
template<typename T>
class RTListAddCommand : public RTCommand
{
  private:
    /// List of elements to be added.  
    std::list<T *> _splice_list;
    /** List of elements that will be processed (via RTCommandEngine). 
     * In this case the elements of this list will be added the RT list passed
     * to the ctor. 
     **/
    std::list<T *> *_dst_list;
  public:
    /** ctor.
     * @param dst_list Ptr to the RT list to which the element will be added.
     * @param element Ptr to the element that will be added.
     **/
    RTListAddCommand(std::list<T *> *dst_list, T *element)
    {
      _dst_list = dst_list;
      _splice_list.push_back(element);
    }

    /** Execute the add RTCommand.
     * Add element(s) to the RT list. 
     *
     * Append elements of RTListAddCommand::_splice_list which contains the
     * elements to be added to the RT list to RTListAddCommand::_dst_list which
     * is the RT list. 
     *
     * Only changes the pointers in the node structures and executes in 
     * constant time. So its RT safe.
     **/
    void execute()
    {
      _dst_list->splice(_dst_list->end(), _splice_list);
    }
    /// Empty function, because no cleanup is necessary. 
    void cleanup()
    {}
};

/// RTList remove command.  
template<typename T>
class RTListRemCommand : public RTCommand
{
  private:
    /// List of elements to be processed
    std::list<T *> _splice_list;
    /// List of elements that will be processed (via RTCommandEngine). 
    std::list<T *> *_dst_list;
    /// self-explanatory
    T *_element_to_remove;

  public:
    /** ctor.
     * @param dst_list List from which the element will be removed.
     * @param element Element that will be removed. 
     **/
    RTListRemCommand(std::list<T *> *dst_list, T *element)
    {
      _dst_list = dst_list;
      _element_to_remove =  element;
    }

    /** Execute the remove RTCommand.  
     * Remove an elements from the RT list.  
     *
     * Find the element and splice it over to RTListRemCommand::_splice_list.
     * Memory will be freed in in cleanup() based on this list.
     **/
    void execute()
    {
      for (typename std::list<T *>::iterator i = _dst_list->begin();
          i != _dst_list->end(); i++)
      {
        if (*i == _element_to_remove)
        {
          _splice_list.splice(_splice_list.begin(), *_dst_list, i);
          break;
        }
      }
    }

    /// Free the memory of element. 
    // this might be dangerous/unexpected.
    // but i would need a synchronized RTCommand, which waited
    // for the operation to complete, otherwise.
    void cleanup()
    {
      for (typename std::list<T *>::iterator i = _splice_list.begin();
          i != _splice_list.end(); i++)
      {
        //std::cout << "delete called" << std::endl;
        delete *i;
      }
      // The List and node are destroyed in the destructor.
    }
};

/** Implementation of a RT safe list.  
 * A List of T * which is owned by the RT thread.
 *
 * Uses the RTCommandEngine and RTCommand to implement a RT safe list and 
 * maintains a non RT shadow list containing the same elements.  
 *
 * The add() and rem() operations shall be called by the normal thread, and the
 * RT safe list will be processed in the RT thread by the RTCommandEngine. 
 *
 * This engine maintains a list of RTCommand objects and calls inherited 
 * RTCommand::execute() where adding or removing from the RT safe list will
 * happen (in this case). The engine will delete all processed commands.  
 *
 * An RTList is composed of two serialized lists named:
 * - RTList::the_actual_list: Maintained by the RTCommandEngine. 
 * @see rt_begin() 
 * @see rt_end() 
 * @see rt_size() 
 *
 * - RTList::shadow_list: Uses std::list. 
 * @see begin() 
 * @see end() 
 * @see size() 
 *
 * So a RTList, composed of an RT list and a serialized non RT list, 
 * can be accessed form RT and non RT contexts. This is important for the SSR, 
 * because a Source will depend on an Input, which has to be added to the 
 * renderer datastructures before the Source.
 **/
template<typename T>
class RTList
{
  private:
    /// The used RTCommandEngine
    RTCommandEngine &_engine;
    /// List of elements that will be processed by RTCommandEngine. 
    std::list<T *> the_actual_list;
    /// A simple list serialized with RTList::the_actual_list 
    std::list<T *> shadow_list;
  public:
    /// self-explanatory
    typedef typename std::list<T *>::iterator iterator;
    /** ctor.
     * @param engine Ref to an RTCommandEngine
     **/
    RTList(RTCommandEngine &engine): _engine(engine)
    {}

    /// Add an element to the lists.
    // Ownership is passed to the list.
    void add(T *to_add)
    {
      RTListAddCommand<T> *cmd = new RTListAddCommand<T>(&the_actual_list
          , to_add);
      _engine.push_command(cmd);
      // The command will be freed, when comes out of the out_fifo.
      // the command will be inside the list soon. At least its valid.
      // so we can add it to the shadow list.
      shadow_list.push_back(to_add);
    }

    /// Remove an element from the lists.
    void rem(T *to_rem)
    {
      // First lets remove the Element from the shadow list.
      // the element will be deleted later, but we cant know when, and it
      // must be treated as if it was deleted now.
      shadow_list.remove(to_rem);

      RTListRemCommand<T> *cmd = new RTListRemCommand<T>(&the_actual_list
          , to_rem);
      _engine.push_command_and_wait(cmd);
    }
    /// Clear RTList. 
    void clear()
    {
      // wait_command fails, when list is empty.
      if (shadow_list.empty())
        return;

      // push Remove Command for every member of the list.
      RTListRemCommand<T> *cmd;
      while (! shadow_list.empty())
      {
        typename std::list<T *>::iterator it = shadow_list.begin();
        cmd = new RTListRemCommand<T>(&the_actual_list, *it);
        shadow_list.erase(it);
        _engine.push_command(cmd);
      }
      // now wait for the last command to come out of the RT Thread.
      _engine.wait_command(cmd);
    }
    /// Get begin of RT list. 
    iterator rt_begin() { return the_actual_list.begin(); }
    /// Get end of RT list. 
    iterator rt_end() { return the_actual_list.end(); }
    /// Get size of RT list. 
    int rt_size() { return the_actual_list.size(); }

    /// Get begin of non-RT list (shadow list). 
    iterator begin() { return shadow_list.begin(); }
    /// Get end of non-RT list (shadow list). 
    iterator end() { return shadow_list.end(); }
    /// Get size of non-RT list (shadow list). 
    int size() { return shadow_list.size(); }
};

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
