/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * %VBAPRenderer class (implementation).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include <cassert>       // for assert()
#include <cmath>         // for cos()
#include "vbaprenderer.h"
#include "controller.h"  // for set_active_sources(), ...
#include "ssr_global.h"
#include "maptools.h"    // for purge()
#include "mathtools.h"    // for purge()
#include "ptrtools.h"    // for is_null(), destroy()
#include "stringtools.h" // for A2S(), S2A()

// for debugging only:
#include <libxml/debugXML.h>

using ptrtools::is_null;
using ptrtools::destroy;
using stringtools::S2A;
using stringtools::A2S;
using namespace ssr; // for id_t, processing, ready, ...

/** ctor.
 * @param scene see create()
 * @param controller see create()
 * @param delayline_size see create()
 **/
VBAPRenderer::VBAPRenderer(
    const Scene&       scene,
          Controller&  controller,
    const std::string& client_name
    ) throw (jack_error, vbap_error) :
  Renderer(scene, client_name, true), // base class ctor.
  // the second argument enables the jack_callback function
  _controller(controller),
  _input_prefix("in"),
  _output_prefix("out")
{
  if (!_make_outputs())
  {
    maptools::purge(_output_map); // delete any leftovers in _output_map
    throw(vbap_error("Output objects could not be created!"));
  }

  // There shouldn't be any sources in the scene at this point.
  // If there are, they are ignored!

  if (this->get_buffer_size() == 0)
  {
    WARNING("Buffersize == 0!");
  }
  else
  {
    // fill silence buffer with zeroes
    _silence.resize(this->get_buffer_size(), 0);
  }

  // init memory for crossfade slopes
  _fade_in.resize (this->get_buffer_size(), static_cast<sample_t>(0.0f));
  _fade_out.resize(this->get_buffer_size(), static_cast<sample_t>(0.0f));

  // raised cosine fade
  for(unsigned int n = 0u; n < this->get_buffer_size(); n++)
  {
    _fade_in[n]  = static_cast<sample_t>(0.5f
        + 0.5f * cos(static_cast<float>(this->get_buffer_size()-n)
		     / static_cast<float>(this->get_buffer_size()) * mathtools::PI));
    _fade_out[n] = static_cast<sample_t>(0.5f
        + 0.5f * cos(static_cast<float>(n)
		     / static_cast<float>(this->get_buffer_size()) * mathtools::PI));
  } // for

  _controller.publish_sample_rate((unsigned int)this->get_sample_rate());
  _controller.set_processing_state(ready); // TODO: lock with mutex?

  this->activate(); // activate JACK client
}

VBAPRenderer::~VBAPRenderer()
{
  _controller.set_processing_state(exiting);
  // cond_signal to any waiting threads?
  maptools::purge(_output_map);
  VERBOSE2("VBAPRenderer: Output map cleared.");
  maptools::purge(_input_map);
  VERBOSE2("VBAPRenderer: Input map cleared.");
}

/** stop rendering.
 * This is only called by JACK. Do not dare to call it yourself!
 * - stops rendering
 * - calls its base class counterpart
 *
 * @return I don't know what
 */
int VBAPRenderer::jack_shutdown()
{
  ERROR("Prematurely disconnected from JACK server! "
      "Trying to stop processing ...");
  _controller.set_processing_state(exiting);
  // cond_signal to any waiting threads?
  maptools::purge(_output_map); WARNING("Output map cleared.");
  maptools::purge(_input_map);  WARNING("Input map cleared.");

  return JackClient::jack_shutdown(); // call original base class function
}

/** _.
 * @param scene Reference to the Scene object of the calling class. This
 * is used to obtain the geometry data in each JACK callback.
 * @param controller Reference to the Controller object. This is used to send
 * back information about active/inactive loudspeakers.
 * @param delayline_size Size of the delay lines in samples (all of them have
 * the same length).
 * @param initial_delay Amount of negative delay that can be used when reading
 * from the delay line
 * @param client_name see JackClient::create()
 * @return auto_ptr to the new VBAPRenderer object.
 * @todo information about active loudspeakers per source should be passed to
 * the Controller.
 **/
VBAPRenderer::ptr_t VBAPRenderer::create(const Scene& scene,
    Controller& controller, const std::string& client_name)
{
  ptr_t temp; // temp = NULL
  try
  {
    temp.reset(new VBAPRenderer(scene, controller, client_name));
  }
  catch(jack_error& e)
  {
    ERROR(e.what());
  }
  catch(vbap_error& e)
  {
    ERROR(e.what());
  }
  return temp;
}

bool VBAPRenderer::new_source(id_t id, const std::string& port_name)
{
  if (_input_map.find(id) != _input_map.end())
  {
    WARNING("Input '" << id << "' already exists!");
    return false;
  }
  Input::ptr_t temp(Input::create(*this,
          _scene,
          _controller,
          id,
          _input_prefix + "_" + A2S(id)));
  if (is_null(temp))
  {
    WARNING("Initialization of input '" << id << "' failed!");
    return false;
  }
  // connect portname with input port
  if (!this->connect_ports(port_name, temp->get_port_name()))
  {
    WARNING("Input " << id << " could not be connected!");
    return false;
  }
  _input_map[id] = temp.release();

  VERBOSE("VBAPRenderer: Input " << id << " registered.");
  return true;
}

bool VBAPRenderer::new_source(id_t id, const std::string& port_name, 
				  const std::string& brir_file)
{
  (void) id;
  (void) port_name;
  (void) brir_file;
  
  ERROR("This function may not be called.");
  
  return false;
}

/** _.
 * This will only be called once. Loudspeakers can @b not be added later.
 * %Loudspeaker-IDs always range from 1 to @a number.
 * @param number number of loudspeakers
 * @return @b true on success.
 **/
bool VBAPRenderer::_make_outputs()
{
  if (!_output_map.empty())
  {
    WARNING("BUG: Output map is not empty!");
    maptools::purge(_output_map);
  }

  // get loudspeakers
  Loudspeaker::container_t loudspeakers;
  _scene.get_loudspeakers(loudspeakers);

  Loudspeaker::container_t::const_iterator i;
  Scene::loudspeakers_t::size_type id = 1;

  /////////// determine loudspeaker delays ////////////////
  for (i = loudspeakers.begin(); i != loudspeakers.end(); ++i)
  {

    DelayLine<sample_t>::size_type delay_samples = 
         static_cast<DelayLine<sample_t>::size_type>
         (i->delay * this->get_sample_rate() + 0.5f);

    // make Output
    Output::ptr_t temp(Output::create(*this, 
				      delay_samples + this->get_buffer_size(),
				      delay_samples,
				      _output_prefix + "_" + A2S(id)));
    if (is_null(temp))
    {
      WARNING("Initialization of output " << id << " failed!");
      maptools::purge(_output_map);
      return false;
    }
    _output_map[id] = temp.release();

    ++id;
  }


  VERBOSE("VBAPRenderer: " << _output_map.size() << " Output objects registered.");

  return true;
}

bool VBAPRenderer::delete_source(const id_t id)
{
  return maptools::delete_item(_input_map, id);
}


void VBAPRenderer::delete_all_sources()
{
  maptools::purge(_input_map);
  // TODO: anything else?
}

/** _.
 * @param id identifier of the input channel
 * @return Pointer to the corresponding @a id. If none was found, @b NULL is
 * returned.
 **/
const VBAPRenderer::Input* VBAPRenderer::get_input(id_t id) const
{
  return maptools::get_item(_input_map, id);
}

/** fill a buffer with zeroes.
 * @param buffer pointer to the memory to be cleared
 * @param buffer_size number of data elements to be cleared
 **/
void VBAPRenderer::clear_buffer(sample_t* const buffer,
    const jack_nframes_t buffer_size) const
{
  if (buffer_size != _silence.size())
  {
    // this should never happen because JACK buffer size will not change!
    WARNING("BUG?: silence resized! (" << _silence.size() << " -> "
        << buffer_size << ")");
    _silence.resize(buffer_size, 0);
  }
  else
  {
    std::copy(_silence.begin(), _silence.end(), buffer);
  }
}

void VBAPRenderer::get_output_port_names(port_list_t& port_list) const
{
  for (output_map_t::const_iterator i = _output_map.begin();
      i != _output_map.end(); ++i)
  {
    assert(i->second != NULL);
    port_list.push_back(i->second->get_port_name());
  }
}
/** This function calculates the speaker weight according the the vector base 
 * formulation proposed by Ville Pulkki in "Virtual Sound Source Positioning Using 
 * Vector Base Amplitude Panning", Journal of the Audio Engineering Society (JAES), 
 * Vol.45(6), June 1997.\par
 * The speaker weights are calculated as follows:
 * \f$ \displaystyle g_{left} = \frac{\cos \phi \sin \phi_0 + 
 * \sin \phi \cos \phi_0}{2 \cos \phi_0 \sin \phi_0} \f$
 * \f$ \displaystyle g_{right} = \frac{\cos \phi \sin \phi_0 - 
 * \sin \phi \cos \phi_0}{2 \cos \phi_0 \sin \phi_0} \f$
 * \par 
 * whereby \f$ \displaystyle \phi\f$ denotes blah, blah
 */
VBAPRenderer::SpeakerWeights VBAPRenderer::_calculate_speaker_weights(
    const SourceCopy& source, const Loudspeaker& left, const Loudspeaker& right)
{
  float phi_0 = ((left.position - _scene.get_reference().position).orientation().azimuth -
     (right.position - _scene.get_reference().position).orientation().azimuth);

  phi_0 = mathtools::wrap_angle(phi_0);

  phi_0 /= 2.0f;

  float phi = (source.position - _scene.get_reference().position).orientation().azimuth -
    (right.position - _scene.get_reference().position).orientation().azimuth;

  phi = mathtools::wrap_angle(phi);

  phi -= phi_0;

  // TODO: This is an emergency brake.
  if (phi > phi_0) phi = phi_0;
  else if (phi < -phi_0) phi = -phi_0;

  // make sure that sound source lies between two speakers
  assert(phi_0 >= std::abs(phi));

  SpeakerWeights weights;
  if (phi_0 >= 180.f) return weights; // TODO: reject rendering???

  float num1 = cos(mathtools::deg2rad(phi)) * sin(mathtools::deg2rad(phi_0));
  float num2 = sin(mathtools::deg2rad(phi)) * cos(mathtools::deg2rad(phi_0));
  float den  = 2 * cos(mathtools::deg2rad(phi_0)) * sin(mathtools::deg2rad(phi_0));

  weights.left  = (num1 + num2) / den;
  weights.right = (num1 - num2) / den;

  // WARNING("g_l: " << weights.left << " g_r: " << weights.right);

  return weights;

}

/** Determines if an angle is larger than another, i.e. if an angle
 * is farther rotated counterclockwise than another
 * @param angle1 angle to be checked against 
 * @param agnle2 reference angle
 * @return @b true if @ angle1 is larger
 */
bool VBAPRenderer::_is_angle_larger(float angle1, float angle2)
{
  return mathtools::wrap_angle(angle1) > mathtools::wrap_angle(angle2);
}

/** Determines if a loudspeaker pair is suitable for rendering a source
 * via VBAP, i.e. if the angle between the loudspeakers is 
 * smaller than 180 degrees
 * @param left loudspeaker left of source
 * @param right loudspeaker right of source
 * @return @b true if loudspeaker pair is suitable
 */
bool VBAPRenderer::_is_valid_loudspeaker_pair(const Loudspeaker& left,
                const Loudspeaker& right)
{
  float phi_0 = (left.position - _scene.get_reference().position).orientation().azimuth -
    (right.position - _scene.get_reference().position).orientation().azimuth;

  phi_0 = mathtools::wrap_angle(phi_0);

  return phi_0 < 180.0f;
}

/** JACK process callback function.
 * This function is called every time the JACK server has new audio data.
 * @param nframes number of frames
 * @return I don't know what
 */
int VBAPRenderer::jack_process(const jack_nframes_t nframes)
{
// code must be suitable for real-time execution.
// That means following function are not allowed:
// malloc, free, printf, pthread_mutex_lock, sleep,
// wait, poll, select, pthread_join, pthread_cond_wait,
// etc, etc.

  // clear all information about active/inactive loudspeakers
  _controller.clear_all_active_sources();
  // send some information to the controller (and then to the subscribers)
  _controller.set_transport_state(this->get_transport_state());
  _controller.set_cpu_load(this->get_cpu_load());

  switch (_scene.get_processing_state())
  {
    case processing:
      {
        Loudspeaker::container_t all_loudspeakers;
        // necessary memory is reserved inside the function
        _scene.get_loudspeakers(all_loudspeakers);
        // the loudspeakers contain already absolute positions

        // TODO: remove this if clause?
        if (all_loudspeakers.size() != _output_map.size())
        {
          WARNING("Different number of loudspeakers in Scene and VBAPRenderer!");
          // TODO: how to react on that?
        }

        output_list_t normal_ls_list;
        output_list_t subwoofers_list;

	/////////// prepare loudspeakers to render ///////////////////

        for (Loudspeaker::container_t::size_type i = 0;
            i < all_loudspeakers.size(); ++i)
        {
          // WARNING: vector indices start with 0, loudspeaker IDs start with 1!
          if (_output_map.find(i + 1) == _output_map.end())
          {
            // This should never happen!
            ERROR("Loudspeaker #" << i + 1 << " not found in output map!");
          }
          else
          {
	    _output_map.find(i + 1)->second->update_previous_source_weights();

            *_output_map[i + 1] = all_loudspeakers[i]; // assign new position,...

            _output_map.find(i + 1)->second->clear_active_source_ids();

            // clear output buffer (to accumulate data)
	    _output_map.find(i + 1)->second->clear_output_buffer(nframes);

            // cull normal loudspeakers 
            if (_output_map.find(i + 1)->second->model != Loudspeaker::subwoofer)
            {
              normal_ls_list.push_back(_output_map.find(i + 1)->second);
            }
            // cull subwoofers
            else //if (_output_map.find(i + 1)->second->model == Loudspeaker::subwoofer)
            {
              subwoofers_list.push_back(_output_map.find(i + 1)->second);
            }
          }

        } // for

        // sort loudspeakers according to azimuth
        normal_ls_list.sort(Output::sort_operator);

        SourceCopy::container_t sources;
        _scene.get_sources(sources);

        for (SourceCopy::container_t::iterator i = sources.begin();
             i != sources.end(); ++i)
        {
          if (_input_map.find(i->id) == _input_map.end())
          {
            // This should never happen!
            ERROR("Source #" << i->id << " not found in input map!");
          }
          else
          {
            *_input_map[i->id] = *i; // assign new position,...            
          }
        }

        const DirectionalPoint reference = _scene.get_reference();
	// float master_volume = _scene.get_corrected_master_volume();

	// iterate over all sources
        for (input_map_t::const_iterator in_iter = _input_map.begin();
             in_iter != _input_map.end(); ++in_iter)
        {
          assert(in_iter->second != NULL);

          // write new audio data into delayline
          in_iter->second->fetch_new_buffer(nframes);

          // determine active loudspeakers for each source
          output_list_t::iterator iter_to_loudspeaker = normal_ls_list.begin();
          output_list_t::iterator iter_to_next_loudspeaker = ++normal_ls_list.begin();

          // if only one loudspeaker present
          if (iter_to_next_loudspeaker == normal_ls_list.end())
          {

	    _render_source_on_output(*in_iter->second, *iter_to_loudspeaker, 
				     1.0f, nframes);

            continue; // next source
          }

          bool ls_pair_found = false;

          // iterate to loudspeaker left of source
          while(!ls_pair_found && iter_to_next_loudspeaker != normal_ls_list.end())
          {
            // check if we have surpassed the source
            if (_is_angle_larger((((*iter_to_loudspeaker)->position
                 - _scene.get_reference().position).orientation()
            - (*_scene.get_source_position(in_iter->first)
               - _scene.get_reference().position).orientation()).azimuth,
               (((*iter_to_next_loudspeaker)->position
                 - _scene.get_reference().position).orientation()
            - (*_scene.get_source_position(in_iter->first)
               - _scene.get_reference().position).orientation()).azimuth))
            {
              ls_pair_found = true;
            }
            else
            {
              iter_to_loudspeaker++;
              iter_to_next_loudspeaker++;
            }

          } // while

          // if between first and last
          if (!ls_pair_found && iter_to_next_loudspeaker == normal_ls_list.end())
          {
            iter_to_loudspeaker = --normal_ls_list.end();
            iter_to_next_loudspeaker = normal_ls_list.begin();
          }

          if (!_is_valid_loudspeaker_pair(**iter_to_next_loudspeaker,
                  **iter_to_loudspeaker))
          {
            // find loudspeaker which is closer
            float angle_to_left_loudspeaker = (((*iter_to_next_loudspeaker)->position
                  - _scene.get_reference().position).orientation()
                - (*_scene.get_source_position(in_iter->first) 
                  - _scene.get_reference().position).orientation()).azimuth;

            if (angle_to_left_loudspeaker > 180.0f) 
            {
              angle_to_left_loudspeaker = std::abs(angle_to_left_loudspeaker - 360);
            }

            float angle_to_right_loudspeaker = (((*iter_to_loudspeaker)->position
                  - _scene.get_reference().position).orientation()
                - (*_scene.get_source_position(in_iter->first)
                  - _scene.get_reference().position).orientation()).azimuth;

            if (angle_to_right_loudspeaker > 180.0f)
            {
              angle_to_right_loudspeaker = std::abs(angle_to_right_loudspeaker - 360);
            }

            Output* active_loudspeaker = NULL;

            float weight = 1.0f;

            if (_is_angle_larger(angle_to_right_loudspeaker, angle_to_left_loudspeaker))
            {
              if (angle_to_left_loudspeaker < 30.0f)
              {
                active_loudspeaker = *iter_to_next_loudspeaker;
                weight = 0.5f * (1 + cos(mathtools::PI/30.0f * angle_to_left_loudspeaker));
              }
            }
            else
            {
              if (angle_to_right_loudspeaker < 30.0f)
              {
                active_loudspeaker = *iter_to_loudspeaker;
                weight = 0.5f * (1 + cos(mathtools::PI/30.0f * angle_to_right_loudspeaker));
              }
            }

            if (active_loudspeaker != NULL)
            {
              VERBOSE2("Couldn't find valid loudspeaker pair for source '"
                  << _scene.get_source_name(in_iter->first) << "'. Using closest loudspeaker only.");

              // render source
	      _render_source_on_output(*in_iter->second, active_loudspeaker, 
				       weight, nframes);
              // render subwoofers
              for (output_list_t::const_iterator i = subwoofers_list.begin();
                  i != subwoofers_list.end(); ++i)
              {
		_render_source_on_output(*in_iter->second, *i, weight, nframes);
              }
            }
            else
            {
              VERBOSE2("Couldn't find valid loudspeaker pair for source '"
                  << _scene.get_source_name(in_iter->first) << "'.");
            }

            continue; // next source
          }
          else // if a valid loudspeaker pair was found
          {
            SpeakerWeights weights = _calculate_speaker_weights(*in_iter->second,
                      **iter_to_next_loudspeaker,
                      **iter_to_loudspeaker);

            // render source
	    _render_source_on_output(*in_iter->second, *iter_to_next_loudspeaker, 
				     weights.left, nframes);
	    
	    _render_source_on_output(*in_iter->second, *iter_to_loudspeaker, 
				     weights.right, nframes);

            // render subwoofers
            for (output_list_t::const_iterator i = subwoofers_list.begin();
                i != subwoofers_list.end(); ++i)
            {
	      _render_source_on_output(*in_iter->second, *i, 1.0f, nframes);
            }
          }
        } // for all sources

        float master_level = 0.0f;

	// finish rendering on all outputs
        for (output_map_t::const_iterator i = _output_map.begin();
             i != _output_map.end(); ++i)
        {
          assert(i->second != NULL);
	  
	  // fade-out sources on loudspeakers which are not used anymore
	  while (i->second->unrendered_source_left())
	  {
	    _fade_out_source_on_output(
			    i->second->return_id_of_remaining_source(), 
			    &(*i->second), nframes);
	  }

	  // finally release audio data
	  i->second->write_output_buffer(nframes);

          master_level = std::max(master_level, i->second->get_signal_level(nframes));

          if (i->second->active_source_ids.size() != 0)
          {
            _controller.set_active_sources(i->first,
                i->second->active_source_ids);
          }
        }

        _controller.set_master_signal_level(master_level);

        // return something?
        break;
      }

    case ready: // audio processing is OFF
      for (output_map_t::const_iterator i = _output_map.begin();
          i != _output_map.end(); ++i)
      {
        assert(i->second != NULL); // assert that object really exists
        i->second->write_empty_buffer(nframes);
      }
      break;

    case initializing:
      WARNING("JACK process callback called while in \"initializing\" state!");
      // return something bad?
      break;

    case exiting:
      WARNING("JACK process callback called while in \"exiting\" state!");
      // return something bad?
      break;

    default:
      ERROR("BUG: Unknown client state!");
      // return something bad?
  }
  // TODO: find out what the return value means
  return 0;
}

void VBAPRenderer::_render_source_on_output(const SourceCopy& source, 
					    Output* current_output, 
					    float weighting_factor,
					    const jack_nframes_t nframes)
{
  //const Source &source = s->src; // this was put by Torben. 
  //                                  s was declared VBAPSource

  VERBOSE3("\n======================================"
      "==========================================");
  VERBOSE3("== output position: x: " << current_output->position.x << ", y: "
      << current_output->position.y << ", azimuth: " 
      << current_output->orientation.azimuth);
  VERBOSE3("== JACK port: " << current_output->get_port_name());

  // no distance attenuation for plane waves 
  if (source.model == Source::plane)
  {
    weighting_factor *= 0.5f / _scene.get_amplitude_reference_distance(); // 1/r
    // weighting_factor *= 
    //        0.25f / sqrt(_scene.get_amplitude_reference_distance()); // 1/sqrt(r)
  }
  else
  {
    // consider distance attenuation
    float source_distance = (source.position - 
			 _scene.get_reference().position).length();

    // no volume increase for sources closer than 0.5m
    source_distance = std::max(source_distance, 0.5f);
    weighting_factor *= 0.5f / source_distance; // 1/r
    //weighting_factor *= 0.25f / sqrt(source_distance); // 1/sqrt(r)
  }

  // apply the gain factor of the current source
  weighting_factor *= source.gain;

  // apply master volume
  weighting_factor *= _scene.get_corrected_master_volume();

  // apply loudspeaker weight
  weighting_factor *= current_output->weight;

  if (source.mute == true) weighting_factor = 0.0f;

  assert(weighting_factor >= 0.0f);

  // input object must not be deleted! It's deleted in ~VBAPRenderer()
  const Input* current_input = this->get_input(source.id);
  if (current_input == NULL)
  { // this happens regularly when a source is added and the renderer is
    // already processing because it takes some time until the JACK ports are
    // ready. Therefore the following warning is only printed if verbose > 1.
    VERBOSE2("VBAPRenderer::Output: Input for source " << source.id
       << " doesn't exist in input map!");
    return; //continue; // continue with next source
  }

  float previous_source_weight = 
         current_output->get_previous_source_weight(source.id);

  sample_t* in = current_input->get_current_audio_frame();
  sample_t* out = current_output->get_output_buffer(nframes);

  // this is the actual audio processing
  weight_buffer_and_add(in, previous_source_weight, weighting_factor, 
			out, nframes);
 
  current_output->store_current_source_weight(source.id, weighting_factor);
  
  // The current loudspeaker is active for the current source:
  current_output->active_source_ids.insert(source.id);

}

void VBAPRenderer::_fade_out_source_on_output(const id_t source_id, 
					      Output* current_output,
					      const jack_nframes_t nframes)
{
  const Input* current_input = this->get_input(source_id);
  
  if (current_input == NULL)
  {
    VERBOSE2("VBAPRenderer::Output: Input for source " << source_id
	     << " doesn't exist in input map!");
    return;
  }
  
  sample_t* in = current_input->get_current_audio_frame();
  sample_t* out = current_output->get_output_buffer(nframes);
  
  float previous_source_weight = 
    current_output->get_previous_source_weight(source_id);

  // this is the actual audio processing
  weight_buffer_and_add(in, previous_source_weight, 0.0f, 
			out, nframes);

}

void VBAPRenderer::weight_buffer_and_add(const sample_t* buffer_to_be_added,
					 const float prev_weight,
					 const float curr_weight,
					 sample_t* target_buffer,
					 const jack_nframes_t nframes)
{
  // weight audio data and add to output buffer  
  for (unsigned int sample = 0; sample < nframes; sample++)
  {
    // TODO: Make crossfade more elegant?
    *(target_buffer+sample) += *(buffer_to_be_added+sample) *
                                  (curr_weight * _fade_in[sample] +
                                   prev_weight * _fade_out[sample]);
  }
}

unsigned int VBAPRenderer::get_number_of_output_channels() const
{
  return _output_map.size();
}

Renderer::renderer_t VBAPRenderer::get_renderer_type() const
{
  return Renderer::VBAP;
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class VBAPRenderer::SourceCopy
////////////////////////////////////////////////////////////////////////////////

VBAPRenderer::SourceCopy::SourceCopy(const std::pair<id_t, Source>& other) :
  DirectionalPoint(other.second), // copy ctor of base class
  id(other.first),
  model(other.second.model),
  gain(other.second.gain),
  mute(other.second.mute)
{}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class VBAPRenderer::Output
////////////////////////////////////////////////////////////////////////////////

/** ctor.
 * @param renderer calling VBAPRenderer object
 **/
VBAPRenderer::Output::Output(const VBAPRenderer& renderer,
    const DelayLine<sample_t>::size_type delay_line_size, 
    const DelayLine<sample_t>::size_type delay) :
  DelayLine<sample_t>(delay_line_size, delay), 
  _renderer(renderer),
  _delay_line_required(delay)
{}

VBAPRenderer::Output::~Output()
{
  if (_port) _renderer.unregister_port(_port);
}

VBAPRenderer::Output::ptr_t VBAPRenderer::Output::create(
    const VBAPRenderer& renderer, 
    const DelayLine<sample_t>::size_type delay_line_size, 
    const DelayLine<sample_t>::size_type delay, 
    const std::string& port_name)
{
  ptr_t temp(new Output(renderer, delay_line_size, delay));
  if (!temp->_init(port_name))
  {
    ERROR("Unable to initialize output port '" + port_name + "'!");
    destroy(temp);
  }
  return temp;
}

VBAPRenderer::Output& VBAPRenderer::Output::operator=(const Loudspeaker& ls)
{
  this->DirectionalPoint::operator=(ls);
  this->weight = ls.weight;
  this->model  = ls.model;
  this->mute   = ls.mute;
  this->active = ls.active; // even if we are not interested in it
  return *this;
}

/** create a JACK output port and get its full name.
 * @param port_name the desired output port name (the part after the ":")
 *
 * The full port name including the client name can be obtained by calling
 * get_port_name() afterwards.
 **/
bool VBAPRenderer::Output::_init(const std::string& port_name)
{
  _port = _renderer.register_out_port(port_name);
  if (_port == NULL)
  {
    WARNING("Output port '" + port_name
        + "' could not be registered.");
    return false;
  }
  // get actual port name and save it to member variable
  _port_name = jack_port_name(_port);

  _audio_buffer.resize(_renderer.get_buffer_size(), 0.0f);

  return true;
}

Renderer::sample_t* VBAPRenderer::Output::get_output_buffer(const jack_nframes_t nframes)
{
  if (!_delay_line_required)
  {
    // return JACK output buffer
    return 
      static_cast<sample_t*>(jack_port_get_buffer(this->get_port(), nframes));
  }
  // return internal audio buffer
  else return &_audio_buffer[0];
}

void VBAPRenderer::Output::write_output_buffer(const jack_nframes_t nframes)
{
  if (!_delay_line_required) return; // nothing to do

  // else: write to delay line, then read from it

  this->prepare_write();
  this->write(&_audio_buffer[0], nframes);

  // get JACK output buffer
  sample_t* out = static_cast<sample_t*>
       (jack_port_get_buffer(this->get_port(), nframes));

  if (!read(out, nframes))
  {
    ERROR("Can not write output buffer.");
  }
}

void VBAPRenderer::Output::update_previous_source_weights()
{
  _previous_source_weights.clear();
  _previous_source_weights = _current_source_weights;
  _current_source_weights.clear();
}

float VBAPRenderer::Output::get_previous_source_weight(id_t source_id) 
{

  coefficient_map_t::iterator iter = 
                            _previous_source_weights.find(source_id);
  float weight = 0.0f;

  if (iter != _previous_source_weights.end())
  {
    weight = iter->second.weight;
    _previous_source_weights.erase(iter);
  }
  
  return weight;
}

void VBAPRenderer::Output::store_current_source_weight(const id_t id, const float weight)
{
  _current_source_weights[id] = Coefficient(weight);
}

bool VBAPRenderer::Output::unrendered_source_left() const
{
  return !_previous_source_weights.empty();
}

id_t VBAPRenderer::Output::return_id_of_remaining_source()
{
  id_t id = _previous_source_weights.begin()->first;
  
  // erase source
  _previous_source_weights.erase(_previous_source_weights.begin());
  
  return id;
}

std::string VBAPRenderer::Output::get_port_name()
{
  return _port_name;
}

jack_port_t* VBAPRenderer::Output::get_port() const
{
  return _port;
}

void VBAPRenderer::Output::write_empty_buffer(const jack_nframes_t nframes) const
{
  if (_port)
  {
    sample_t *out = static_cast<sample_t*>
                         (jack_port_get_buffer(_port, nframes));
    _renderer.clear_buffer(out, nframes);
  }
}

void VBAPRenderer::Output::clear_active_source_ids()
{
  this->active_source_ids.clear(); // reset active sources to an empty set
}

void VBAPRenderer::Output::clear_output_buffer(const jack_nframes_t nframes)
{

  sample_t* out;

  if (!_delay_line_required)
  {
    out = static_cast<sample_t*>(jack_port_get_buffer(_port, nframes));
  }
  else
  {
    out = &_audio_buffer[0]; 
  }

  _renderer.clear_buffer(out, nframes);
}

float VBAPRenderer::Output::get_signal_level(const jack_nframes_t nframes)
{
  // get JACK output buffer
  sample_t* out;
  out = static_cast<sample_t*>(jack_port_get_buffer(_port, nframes));

  return _renderer.get_max_signal_amplitude(out, nframes);
}


////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class VBAPRenderer::Input
////////////////////////////////////////////////////////////////////////////////

/** private ctor. Object can only be created with create().
 * @param renderer reference to the calling VBAPRenderer object. Using this, we
 * can access the public members of the VBAP renderer.
 **/
VBAPRenderer::Input::Input(const VBAPRenderer& renderer,
        const Scene& scene,
        Controller& controller,
        const id_t source_id) :
  SourceCopy(source_id),
  _renderer(renderer),
  _scene(scene),
  _controller(controller),
  _port(NULL),
  _port_name("")
{}

VBAPRenderer::Input::~Input()
{
  if (_port) _renderer.unregister_port(_port);
}

VBAPRenderer::Input::ptr_t VBAPRenderer::Input::create(
        const VBAPRenderer& renderer,
        const Scene& scene,
        Controller& controller,
        const id_t source_id,
        const std::string& name)
{
  ptr_t temp(new Input(renderer, scene, controller, source_id));
  if (!temp->_init(name))
  {
    ERROR("Unable to initialize input '" + name + "'!");
    destroy(temp);
  }
  return temp;
}

bool VBAPRenderer::Input::_init(const std::string& name)
{
  _port = _renderer.register_port(name, JackPortIsInput);
  if (_port == NULL)
  {
    WARNING("Input port '" + name + "' could not be registered.");
    return false;
  }
  // get actual port name and save it to member variable
  _port_name = jack_port_name(_port);
  return true;
}

VBAPRenderer::Input& VBAPRenderer::Input::operator=(const SourceCopy& source)
{
  this->DirectionalPoint::operator=(source);
  this->model  = source.model;
  this->mute   = source.mute;
  this->gain   = source.gain;

  return *this;
}

bool VBAPRenderer::Input::fetch_new_buffer(jack_nframes_t nframes)
{
  sample_t *in;
  in = static_cast<sample_t*>(jack_port_get_buffer(_port, nframes));
  if (in == NULL)
  {
    WARNING("Got no input buffer from JACK!");
    return false;
  }
  assert(in != NULL);

  _current_audio_frame = in;

  if (_scene.get_source_mute_state(this->id))
  {
    // update source signal level
    _controller.set_source_signal_level(this->id, 0.0f);
  }
  else
  {
    // update source signal level (max amplitude)
    _controller.set_source_signal_level(this->id,
        _renderer.get_max_signal_amplitude(_current_audio_frame, nframes)
	     * _scene.get_source_gain(this->id));

    // update source signal level (RMS)
    //_controller.set_source_signal_level(_source_id, 
    //      _renderer.get_signal_rms(in, nframes) * 
    //      std::pow(_scene.get_source_gain(_source_id), int(nframes))
    //      , 0);
  }
  return true;
}

Renderer::sample_t* VBAPRenderer::Input::get_current_audio_frame() const
{
  return _current_audio_frame;
}

std::string VBAPRenderer::Input::get_port_name() const
{
  return _port_name;
}

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
