/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * %VBAPRenderer class (definition).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef VBAPRENDERER_H
#define VBAPRENDERER_H

#include <map>
#include <set>
#include <vector>

#include "renderer.h"   // base class of VBAPRenderer
#include "delayline.h"  // base class of VBAPRenderer::Output
#include "directionalpoint.h" // Position and Orientation

class Controller; // forward declaration

using ssr::id_t;

/** VBAP renderer class.
 * This class does the VBAP rendering. Audio processing is done with the
 * inherited JACK functionality.
 **/
class VBAPRenderer : public Renderer
{
  private:
    class Output;          // nested class, defined later
    class Input;           // nested class, defined later
    //  class SourceCopy;

    /** Class to temporarily store selected source information.
     * By use of a class like this, the desired information for every source can
     * be obtained using the function Scene::get_source_list().
     **/
    struct SourceCopy : DirectionalPoint
    {
      /// ctor
      SourceCopy() {};
      SourceCopy(id_t id) : id(id) {};

      typedef std::vector<SourceCopy> container_t; ///< list of SourceCopys
      /// type conversion constructor
      SourceCopy(const std::pair<id_t, Source>& other);

      id_t id;               ///< unique ID, see Scene::source_map_t
      Source::model_t model; ///< model, see Source
      float gain;            ///< gain, see Source
      bool mute;             ///< mute/unmute, see Source
    };

    struct SpeakerWeights
    {
      /// ctor
      SpeakerWeights() : left(0.0f),right(0.0f){};

      float left;
      float right;
    };

    bool _is_angle_larger(float angle1, float angle2);
    bool _is_valid_loudspeaker_pair(const Loudspeaker& left,
            const Loudspeaker& right);

    void _render_source_on_output(const SourceCopy& source, 
				  Output* current_output, 
				  float weighting_factor,
				  const jack_nframes_t nframes);

    void _fade_out_source_on_output(const id_t source_id, 
				    Output* current_output,
				    const jack_nframes_t nframes);

  public:
    typedef std::auto_ptr<VBAPRenderer> ptr_t;

    /// a map which holds all output objects.
    typedef std::map<Loudspeaker::container_t::size_type, Output*> output_map_t;
    /// a map which holds all input objects.
    typedef std::map<id_t, Input*> input_map_t;
    /// temporary list of outputs
    typedef std::list<Output*> output_list_t;

    /// exception to be thrown by ctor.
    struct vbap_error : public std::runtime_error
    {
      vbap_error(const std::string& s): std::runtime_error(s) {}
    };

    VBAPRenderer(const Scene& scene, Controller& controller,
		 const std::string& client_name)
      throw (jack_error, vbap_error);

    /// static factory function for %VBAPRenderer objects
    static ptr_t create(const Scene& scene, Controller& controller,
			const std::string& client_name = "VBAP-Renderer");

    /// delete all Input/Output objects and clear input and output map.
    virtual ~VBAPRenderer();

    // from JackClient:
    virtual int jack_shutdown();
    // from Renderer:
    virtual bool new_source(id_t id, const std::string& port_name);
    virtual bool new_source(id_t id, const std::string& port_name,
			    const std::string& brir_file);
    virtual bool delete_source(id_t id);
    virtual void delete_all_sources();
    virtual void get_output_port_names(port_list_t& port_list) const;
    virtual unsigned int get_number_of_output_channels() const;
    virtual renderer_t get_renderer_type() const;

    void weight_buffer_and_add(const sample_t* buffer_to_be_added,
			       const float prev_weight,
			       const float curr_weight,
			       sample_t* target_buffer,
			       const jack_nframes_t nframes);
      
    /// Get pointer to input object corresponding to @a id.
    const Input* get_input(id_t id) const;

    void clear_buffer(sample_t* const buffer, jack_nframes_t buffer_size) const;

  private:
    /// initialize renderer
    bool _init(const std::string& client_name);

    /// overwrites virtual member function declared in JackClient class
    virtual int jack_process(jack_nframes_t nframes);

    /// this will be filled with zeroes
    mutable std::vector<sample_t> _silence;

    std::vector<sample_t> _fade_in;
    std::vector<sample_t> _fade_out;

    input_map_t   _input_map;    ///< all input  objects are stored here
    output_map_t _output_map;    ///< all output objects are stored here

    // deprecated:
    /// relative angles of the array loudspeakers
    //std::map<float, Scene::loudspeakers_t::size_type> _loudspeaker_angles;

    Controller& _controller;          ///< Reference to the Controller
    const std::string  _input_prefix; ///< prefix for the input  ports
    const std::string _output_prefix; ///< prefix for the output ports

    /// initialize Output objects
    bool _make_outputs();

    SpeakerWeights _calculate_speaker_weights(const SourceCopy& source,
                const Loudspeaker& left,
                const Loudspeaker& right);

}; // end of VBAPRenderer class

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Output
////////////////////////////////////////////////////////////////////////////////

/** Manages an output buffer and the corresponding JACK calls.
 * @warning z coordinates and elevation are ignored
 **/
class VBAPRenderer::Output : public Loudspeaker, public DelayLine<sample_t>
{
  public:
    typedef std::auto_ptr<Output> ptr_t;
    /// static factory function
    static ptr_t create(const VBAPRenderer& renderer,
			const DelayLine<sample_t>::size_type delay_line_size, 
			const DelayLine<sample_t>::size_type delay, 
			const std::string& port_name);

    ~Output(); ///< Unregister associated JACK output port

    struct Coefficient
    {
      explicit Coefficient(float weight = 1.0) :
        weight(weight)
      {}
      float weight;
    };
    void  update_previous_source_weights();
    void  store_current_source_weight(const id_t id, const float weight);
    float get_previous_source_weight(id_t source_id);
    
    bool unrendered_source_left() const;

    id_t return_id_of_remaining_source();

    std::string get_port_name(); ///< get JACK port name of the output object

    sample_t* get_output_buffer(const jack_nframes_t nframes);
    
    void write_output_buffer(const jack_nframes_t nframes);

    /// clear output buffer
    void write_empty_buffer(jack_nframes_t nframes) const;

    void clear_active_source_ids();

    void clear_output_buffer(const jack_nframes_t nframes);
    
    float get_signal_level(const jack_nframes_t nframes);

    jack_port_t* get_port() const;

    /// use the assignment operator of the base class.
    using DirectionalPoint::operator=;

    /// overloaded assignment operator for Loudspeaker
    Output& operator=(const Loudspeaker& ls);

    /// operator needed to sort loudspeakers according to their positional angle
    static bool sort_operator(const Output* one, const Output* other)
      {return one->position.orientation().azimuth < other->position.orientation().azimuth;};
    bool operator<(const Output* other) const
    {return this->position.orientation().azimuth < other->position.orientation().azimuth;};

    /// sources for which current Output is active
    std::set<id_t> active_source_ids;
    // those are public, because I was too lazy to write get_active_source_ids()

  private:
    Output(const VBAPRenderer& renderer,    
	   const DelayLine<sample_t>::size_type delay_line_size, 
	   const DelayLine<sample_t>::size_type delay);

    bool _init(const std::string& port_name);

    const VBAPRenderer& _renderer; /// ref. to the calling VBAPRenderer object

    jack_port_t* _port;     ///< JACK port associated with this output
    std::string _port_name; ///< full name of the JACK port

    typedef std::map<id_t, Coefficient> coefficient_map_t;

    /// is used to accumulate the weights of the current audio cycle for all sources
    coefficient_map_t _current_source_weights;
      
    /// holds the weights of the previous audio cycle for all sources
    coefficient_map_t _previous_source_weights;

    std::vector<sample_t> _audio_buffer;

    bool _delay_line_required;

};

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Input
////////////////////////////////////////////////////////////////////////////////

/** Manages an input buffer (including delay line) and the corresponding JACK
 * calls.
 **/
class VBAPRenderer::Input : public VBAPRenderer::SourceCopy
{
  public:
    typedef std::auto_ptr<Input> ptr_t;
    /// static factory function
    static ptr_t create(const VBAPRenderer& renderer,
            const Scene& scene,
            Controller& controller,
            const id_t source_id,
            const std::string& name);

    /// Unregisters associated JACK input port
    ~Input();

    /// gets a buffer full of audio from JACK and writes it into the delay line
    bool fetch_new_buffer(jack_nframes_t nframes);

    sample_t* get_current_audio_frame() const;

    std::string get_port_name() const; ///< get JACK port name of the input object

    /// use the assignment operator of the base class.
    using DirectionalPoint::operator=;
    /// overloaded assignment operator for Loudspeaker
    Input& operator=(const SourceCopy& source);

  private:
    // if it were not private, I'd make it "explicit"
    Input(const VBAPRenderer& renderer,
          const Scene& scene,
          Controller& controller,
          const id_t source_id);

    /// creates a JACK port and maybe does some other things, too.
    bool _init(const std::string& port_name);

    /// holds a reference to the calling VBAPRenderer object
    const VBAPRenderer& _renderer;
    /// holds a reference to the scene

    const Scene& _scene;
    /// Reference to the Controller

    Controller& _controller;

    sample_t* _current_audio_frame;

    /// the JACK port corresponding to this input object
    jack_port_t* _port;

    /// full name of the JACK port
    std::string _port_name;

    Input(const Input&);            ///< deactivated
    Input& operator=(const Input&); ///< deactivated

};

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
